#ifndef __OtrAlign__H__
#define __OtrAlign__H__

#ifdef __cplusplus
extern "C" {
#endif

/** @name OtrAlign

    The OtrAlign interface provides access to the alignment rotation of
    3-vectors. 
 */
/*@{*/

/*------------------- Functions -------------------------- */
/** Switch between different solutions of the task of the package.

  @param   i  integer number (currently any number except OTRIInvalid)
  */
void  OtrAlignSwitch(int i);

/** Declare all database tables properly.

  */
void  OtrAlignDefine();

/** Declare all database tables by yourself.

  */

void OtrAlignUserDefine(char *server,int major,int minor,int match,int version);


/** Initialize OTrAlign package.

  All function of the package (beside OtrAlignSetAlgorithm) check
  whether the initialization was done properly.

  @return 0 if state ok, else error code
  */
  
int   OtrAlignInitialize();

/** Update memory structures, arrays ... from database or ascii files.

  Input information is taken from file referenced by
  environment variable OtrAlignInput.

  Example file for run 12333 can be found in
  /afs/desy.de/group/hera-b/OTR/db/align/run12333.dat

  @return 0 if state ok, else error code
  */
  
int   OtrAlignUpdate();

/** Save memory structures, arrays ... to database or ascii files.

  @return 0 if state ok, else error code
  */

int   OtrAlignWriteToDb();

/** Close OtrAlign package
  (free allocate memory,
    clean up to be in a state that initialization is necessary).

  @return 0 if state ok, else error code
  */

int   OtrAlignClose();

/** Read input file and fill structure OtrAlignStruct with data.

  @param  FILE *fp   Input file with alignment information
  @return 0 if state ok, else error code

    */
int   OtrAlignScan(FILE *fp);

/** Print memory structure to ascii files.

  @param  FILE *fp   Output file with alignment information
  @return 0 if state ok, else error code
  */
  
int   OtrAlignPrint(FILE *fp);


/** Test bench to be sucessfully executed before any release of a new version.

  @return 0 if state ok, else error code
  */

int   OtrAlignTestMe();

/**
Transform direction vector from non-aligned to aligned direction.
Coordinates are given in laboratory system.

  @param gi   GeoIndex specification of NEDE
  @param n[3] 3-vector to be transformed (both input and output)
  @return 0 if state ok, else error code
*/


/* Test bench 2 - the same but without proper initialization
   - should give an error.

  @return 0 if state ok, else error code
  */

int   OtrAlignTestMe2();
  
int OtrAlignRotate(int gi,float n[3]);

/**
Transform direction vector from aligned to non-aligned direction.
Coordinates are given in laboratory system.

  @param gi   GeoIndex specification of NEDE
  @param n[3] 3-vector to be transformed (both input and output)
  @return 0 if state ok, else error code
*/

int OtrAlignRotateBack(int gi,float n[3]);

/**
Transform X,Y,Z-vector from non-aligned to aligned.
Coordinates are given in laboratory system.

  @param gi   GeoIndex specification of NEDE
  @param x[3] 3-vector to be transformed (both input and output)
  @return 0 if state ok, else error code
*/

int OtrAlignShift(int gi,float x[3]);

/**
Transform X,Y,Z-vector from aligned to non-aligned.
Coordinates are given in laboratory system.

  @param gi   GeoIndex specification of NEDE
  @param x[3] 3-vector to be transformed (both input and output)
  @return 0 if state ok, else error code
*/

int OtrAlignShiftBack(int gi,float x[3]);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif
