#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include "database/table.h"
#include "database/keytable.h"
#include "Otr/OtrDb.h"
#include "Otr/OtrI.h"

#define MAXREVISION 10

  
static OtrDbTableStruct *tables=0;
static int OtrIsInit=0;
static int OtrDbAlgorithm=OtrDbAutomaticKey|OtrDbTest;
static int numtables=0;

#ifdef __cplusplus
extern "C" {
#endif


#define DEBUG

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

int  OtrDbWritePermission(char *server){
  passwd* p=getpwuid(geteuid());
  char *l=p->pw_name;
  if (!l) return 0;
  printf("OtrDbWritePermission: user=%s server=%s\n",l,server);
  if (!strcmp(l,"skar")) return 1;
  if (!strcmp(l,"root")&&!strcmp(server,"/OTR_STATUS")) return 1;
  if (!strcmp(l,"hbonline")&&!strcmp(server,"/OTR_STATUS")) return 1;

#ifdef DEBUG
  if (!strcmp(l,"eckmann")) return 1;
#endif

  return 0;
  }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

int  OtrDbKeyPermission(){
  passwd* p=getpwuid(geteuid());
  char *l=p->pw_name;
  if (!l) return 0;
  if (!strcmp(l,"root")) return 1;
#ifdef DEBUG
  if (!strcmp(l,"eckmann")) return 1;
#endif
  return 0;
  }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

void  OtrDbSwitch(int i){
  char *db=getenv("OtrDbServerMode");
  OtrDbAlgorithm=i;
  if (!db) OtrDbAlgorithm|=OtrDbTest;
  else if (strcmp(db,"notest")) OtrDbAlgorithm|=OtrDbTest;
  }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

int OtrDbDefine(char *generic_servername,char *generic_tablename,int (*updatefcn)())
  {
  int i;
  int fcnfl=OTRIInvalid;
  OtrDbTableStruct* t;
  for (i=0;i<numtables;i++) {
    t=&tables[i];
    if (t->update==updatefcn) fcnfl=i;
    if (!strcmp(t->generic_servername,generic_servername)&&
        !strcmp(t->generic_tablename,generic_tablename)) break;
    }
  if (i!=numtables) {
    printf("Table: %s %s already defined\n",generic_servername,generic_tablename);
    return OtrIInvalid();
    }
  tables=(OtrDbTableStruct*)realloc(tables,sizeof(OtrDbTableStruct)*(numtables+=1));
  tables[numtables-1].generic_servername=strdup(generic_servername);
  tables[numtables-1].generic_tablename=strdup(generic_tablename);
  tables[numtables-1].update=updatefcn;
  tables[numtables-1].major=0x7fffffff;
  tables[numtables-1].minor=0x7fffffff;
  tables[numtables-1].match=4;
  tables[numtables-1].version=OTRIInvalid;
  tables[numtables-1].updatebyothertable=fcnfl;
  return numtables-1;

  }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  
int OtrDbUserDefine(char *generic_servername,char *generic_tablename,
	      int (*updatefcn)(),int major,int minor, int match,int version)
  {
  int fcnfl=OTRIInvalid;
  int i;
  OtrDbTableStruct* t;
  for (i=0;i<numtables;i++) {
    t=&tables[i];
    if (t->update==updatefcn) fcnfl=i;
    if (!strcmp(t->generic_servername,generic_servername)&&
        !strcmp(t->generic_tablename,generic_tablename)) break;
    }

  if (i==numtables) {
    tables=(OtrDbTableStruct*)realloc(tables,sizeof(OtrDbTableStruct)*(numtables+=1));
    }
  tables[i].generic_servername=strdup(generic_servername);
  tables[i].generic_tablename=strdup(generic_tablename);
  tables[i].update=updatefcn;
  tables[i].major=major;
  tables[i].minor=minor;
  tables[i].match=match;
  tables[i].version=version;
  if (i==numtables) tables[i].updatebyothertable=fcnfl;
  
  printf("OtrDbUserDefine[debug]: tableindex:%d %s/%s_v%d %d %d %d\n",
      i,generic_servername,generic_tablename,version,
      major,minor,match);

  return numtables-1;
  }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

int OtrDbChangeServer(char *server,char *table, char *newserver) {
  int i;
  OtrDbTableStruct* t;
  for (i=0;i<numtables;i++) {
    t=&tables[i];
    if (!strcmp(t->generic_servername,server)&&
        !strcmp(t->generic_tablename,table)) break;
    }
  if (i<numtables) {
    free(t->generic_servername);
    tables[i].generic_servername=strdup(newserver);
    return OtrIValid();
    }
  else {
    printf("OtrDbChangeServer: Table not defined: %s %s\n",server,table);
    return OtrIInvalid();
    }
  }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

int OtrDbSetKey( char *name, int major, int minor) {
  if (major!=0) OtrDbAlgorithm &= ~OtrDbAutomaticKey;
  else OtrDbAlgorithm |= OtrDbAutomaticKey;
  if (!name) {
    int i;
    for (i=0;i<numtables;i++) {
      tables[i].major=major;
      tables[i].minor=minor;
      }
    }
  else {
    OtrDbTableStruct* t=0;
    int i;
    for (i=0;i<numtables;i++) {
      t=&tables[i];
      if (!strcmp(t->generic_tablename,name)) break;
      }
    if (i==numtables) return OtrIInvalid();
    t->major=major;
    t->minor=minor;
    }
  return OtrIValid();
  }
  
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

int OtrDbGetCurrentMajor() {
  return 1;
  }
  
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  
int OtrDbOpen(){
#ifndef USE_ARTE
    return KeyBookTables("/key_table_TEST","keytable",0);
#endif 
#ifdef USE_ARTE
    return 1;
#endif 
      }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

OtrDbTableStruct* OtrDbTableGet(OtrDbTable *tt) {
  OtrDbTableStruct* t=0;
  int i;
  for (i=0;i<numtables;i++) {
    t=&tables[i];
    if (!strcmp(t->generic_tablename,tt->name)){
      break;
      }
    }
  return t;
  }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

void OtrDbPrint(FILE* fp,OtrDbTable* tt){
  if (tt!=0) {
    OtrDbTableStruct* t = OtrDbTableGet(tt);
    fprintf(fp,"Table: server: %s  generic name: %s version: v%d"
               "  major: %d minor: %d match:%d\n",
      t->generic_servername,t->generic_tablename,t->version,t->major,t->minor,t->match);
    }
  }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

int OtrDbGetTableVersion(char *name)  {
  char tname[256];
  
  OtrDbTableStruct* t=0;
  int i;

  OtrIValid();
  printf("OtrDbGetTableVersion[debug]: for table %s\n",name);

  for (i=0;i<numtables;i++) {
    t=&tables[i];
    if (!strcmp(t->generic_tablename,name))break;
    }
  if (i==numtables) {
    OtrIInvalid();
    return -1;
    }

  if (t->updatebyothertable!=OTRIInvalid) {
    printf("OtrDbGetTableVersion[debug]:" 
             " take version %d from mother %d for table %s\n",
	        tables[t->updatebyothertable].version,t->updatebyothertable,name);
    t->version=tables[t->updatebyothertable].version;
    if (t->version!=OTRIInvalid) return t->version;
    }

  printf("OtrDbGetTableVersion[debug]: continue\n");

  if (t&&t->version==OTRIInvalid) {
    int bt;
    Tables tid=OpenTables(t->generic_servername,0);
    if(tid==0) {
      fprintf(stderr,
        "OtrDbGetTableVersion: Can't open database server: %s for table: %s\n",
              t->generic_servername,name);
      return OtrIInvalid();
      }
    for (t->version=1;;t->version++) {
      if (t->version!=0) {
        sprintf(tname,"%s_v%d",t->generic_tablename,t->version);
        }
      else strcpy(tname,t->generic_tablename);
      bt=BookTable(tid,tname,0,0);
      if(bt==0) {
        if (t->version==0) break;
        else if (t->version==1) {
	  t->version=0;
	  continue;
	  }
	else break;
	}
      }
    t->version=t->version-1;
    }
    
  printf("OtrDbGetTableVersion[debug]: version=%d",t->version);
  return (t) ? t->version : -1;
  }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  
// made to be able to update the key tables

int OtrDbGetTableList(int runno,int time,const OtrDbTableStruct **t,int *n) {
  int i;
  
  for (i=0;i<numtables;i++) {
    OtrDbTableStruct* t=&tables[i];
    char tname[256];
    int  bt;
    Tables tid=OpenTables(t->generic_servername,0);

    t->version=OtrDbGetTableVersion(t->generic_tablename);
    if (t->version!=0) {
      sprintf(tname,"%s_v%d",t->generic_tablename,t->version);
      }
    else strcpy(tname,t->generic_tablename);
    t->tablename=strdup(tname);
    
    bt=BookTable(tid,tname,0,0);
    if(bt==0) {
      printf("Can't book table: server: %s table %s\n",
          t->generic_servername,tname);
      return OtrIInvalid();
      }
      
    TableStamp(tid,time,t->minor);
    if (!MatchTable(tid,bt,t->match)) {
      if (!MatchTable(tid,bt,TABLEge)) {
        if (!MatchTable(tid,bt,TABLElt)) {
	  printf("OtrDbGetTableList[warning]: No Match for table %s/%s\n",
	    t->generic_tablename,tname);
	  continue;
	  }
	}
      }
    t->major=TableMajor(tid);
    t->minor=TableMinor(tid);
    }
    
  *n=numtables;
  *t=tables;
  return OtrIValid();
  }


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

int   OtrDbClose(){
#ifndef USE_ARTE
    return KeyCloseTables();
#endif 
#ifdef USE_ARTE
    return 1;
#endif 
  }
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

int OtrDbUpdateDefault(OtrDbTableStruct* t) {
  int version=OtrDbGetTableVersion(t->generic_tablename);
  char tname[256];
  if (version!=0) sprintf(tname,"%s_v%d",t->generic_tablename,version);
  else strcpy(tname,t->generic_tablename);
  
  printf("OtrDbUpdate[warning]: Key table not available. "
          "Update with OtrDbDefine settings for %s/%s\n",
	       t->generic_servername,tname);

  if (!t->minor && !t->major && !t->match) {
    printf("OtrDbUpdate[error]: Table's minor and major not defined.\n");
    printf("OtrDbUpdate[info]: Use OtrDbUserDefine(...) function or use correct key table number\n");
    return OtrIInvalid();
    }
  else {
    printf("OtrDbUpdate[debug]: Get table %s from server %s\n",
	                  t->generic_tablename,t->generic_servername);
    if (t->updatebyothertable==OTRIInvalid) {
      t->update();
      printf("OtrDbUpdate[debug]: update done\n");
      }
    else printf("OtrDbUpdate[debug]: Update together table %d requested\n",
               t->updatebyothertable);
	       
    if (OtrIError()) return OtrIInvalid();
    }
  return OtrIValid();
  }
  
  
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

int OtrDbUpdate(int keynumber){
    static int keytableworks=1;
#ifndef USE_ARTE
    int rc;
#endif
    int oldmajor,oldminor,servernum,tablenum,i;
    char *servname,*tablename;
    OtrDbTableStruct* t;
    
    printf("OtrDbUpdate: Enterning function. key=%d\n",keynumber);
    if (!keytableworks) return OtrIValid();
    
    if (keynumber==-1) {
      int rc;
      for (i=0;i<numtables;i++) {
        if ((rc=OtrDbUpdateDefault(&tables[i]))) return rc;
	}
      } 
    else if (!OtrIsInit) {
#ifndef USE_ARTE
      rc=KeyLoadTables(keynumber);
      if (rc==0) return keynumber ; 
#endif 
      for (i=0;i<numtables;i++) { 
        int v=0;
        char tname[256];
        t=&tables[i];
	printf("OtrDbUpdate[debug]: Call KeyFindServer: %s %s\n",
	                  t->generic_servername,t->generic_tablename);
        servernum=KeyFindServer(t->generic_servername);
	printf("OtrDbUpdate: servernum: %d\n",servernum);
        if (servernum==-1){
	  int rc;
	  printf("OtrDbUpdate[FATAL]: Server:%s not found in the key table\n",t->generic_servername);
	  if ((rc=OtrDbUpdateDefault(t))) return rc;
	  keynumber=-1;
	  continue;
	  }
        else  servname=KeyServerName(servernum);
	
	//printf("servname=%s,servernum=%d\n",servname,servernum);
	tablenum=KeyFindTable(servernum,t->generic_tablename);
	while (tablenum==-1&&v<MAXREVISION) {
	  v++;
          sprintf(tname,"%s_v%d",t->generic_tablename,v);
	  tablenum=KeyFindTable(KeyFindServer(t->generic_servername),tname);
	     //printf("tablename=%s,tablenum=%d\n",tablename,tablenum);
	  }
	if (v==MAXREVISION) {
	  int rc;
	  printf("OtrDbUpdate[FATAL]: Table:%s not found in the key table\n",t->generic_tablename);
	  if ((rc=OtrDbUpdateDefault(t))) return rc;
	  keynumber=-1;
	  continue;
	  }
	tablename=KeyTableName(tablenum);
	   //printf("tablename=%s,tablenum=%d\n",tablename,tablenum);
        t->major=KeyTableMajor(tablenum);
        t->minor=KeyTableMinor(tablenum);
        t->match=0;
	t->version=v;
        if (t->updatebyothertable==OTRIInvalid) {
	  t->update();
	  }
	if (OtrIError()) return OtrIInvalid();
 	}
      OtrIsInit=keynumber;
      if (keynumber==-1) {
        printf("OtrDbUpdate[warning]: DATABASE UPDATES ARE SWITCHED OFF."
	        "Key table mechanism is not working properly.\n");
        keytableworks=0;
	}
      OtrIValid();
      return OtrIsInit;
      }
    else {
      for (i=0;i<numtables;i++) {
        KeyLoadTables(OtrIsInit);
        t=&tables[i];
        servname=KeyServerName(KeyFindServer(t->generic_servername));
        tablename=KeyTableName(KeyFindTable(KeyFindServer(t->generic_servername),t->generic_tablename));
	tablenum=KeyFindTable(KeyFindServer(t->generic_servername),t->generic_tablename);
	oldmajor=KeyTableMajor(tablenum);
        oldminor=KeyTableMinor(tablenum);
        KeyLoadTables(keynumber);
	tablenum=KeyFindTable(KeyFindServer(t->generic_servername),t->generic_tablename);
   	t->major=KeyTableMajor(tablenum);
	t->minor=KeyTableMinor(tablenum);
        t->match=0;
        if(t->major!=oldmajor||t->minor!=oldminor){
          printf("OtrDbUpdate: The following tables were modified:\n");      
          printf("OtrDbUpdate: Server:%s Table:%s Major=%d Minor=%d\n",servname,tablename,t->major,t->minor);      
	  if (t->updatebyothertable==OTRIInvalid) t->update();
	  if (OtrIError()) return OtrIInvalid();
 	  }
        }
      OtrIsInit=keynumber;
      OtrIValid();
      return OtrIsInit;
      }
    return OtrIInvalid();
  }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  
int OtrDbFieldType(char* type){
  int t=0;    
    if (!strcmp(type,"char") )              { t=TABLEbyte; }
    else if (!strcmp(type,"short"))         { t=TABLEshort;}
    else if (!strcmp(type,"int"))           { t=TABLEint;  }
    else if (!strcmp(type,"unsigned char")) { t=TABLEubyte;}
    else if (!strcmp(type,"unsigned short")){ t=TABLEushort;}
    else if (!strcmp(type,"unsigned int"))  { t=TABLEuint;  }
    else if (!strcmp(type,"float"))         { t=TABLEfloat; }
    else if (!strcmp(type,"double"))        { t=TABLEdouble;}
    else if (!strcmp(type,"ascii"))         { t=TABLEascii; }
   // else if (!strcmp(type,"char*"))         { t=TABLEstring;}
    else if (!strcmp(type,"char[]"))        { t=TABLEstring;}
    else {fprintf(stderr,"Invalid type of field %s\n",type);
          exit(0);}
   return t;
   }
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++  

void* OtrDbGet(OtrDbTable *table,void* data,int *nitem) {
   
  OtrDbField* f=table->f;
  int size=table->size;
  int field_size,i,bt;
  Tables tid;
  OtrDbTableStruct* t;

  for (i=0;i<numtables;i++) {
    t=&tables[i];
    if (!strcmp(t->generic_tablename,table->name)) break;
    } 
    
    
  tid=OpenTables(t->generic_servername,0);
  if(tid==0) {
    fprintf(stderr,"Can't open database server: %s\n",t->generic_servername);
    OtrIInvalid();
    return data;
    }
  
  char tname[256];
  if (t->version!=0) {
    sprintf(tname,"%s_v%d",t->generic_tablename,t->version);
    }
  else strcpy(tname,t->generic_tablename);

    //printf("tname=%s t->version=%d\n",tname,t->version);

  bt=BookTable(tid,tname,0,0);
  if(bt==0) {
    fprintf(stderr,"Can't book table: server: %s table %s\n",
        t->generic_servername,tname);
    OtrIInvalid();
    return data;
    }
    
  TableStamp(tid,t->major,t->minor);
  //printf("t->major=%d t->minor%d\n",t->major,t->minor);
  FetchTable(tid,bt,t->match);
 
  field_size=0;

  SkipTableField(tid,1);
  if (/*!strcmp(f[0].fmt,"char*")||*/ !strcmp(f[0].fmt,"char[]")){
    char *ss=0;
    field_size=GetTableField(tid,TABLEstring,0,&ss);
    SkipTableField(tid,1);
    }
  else {
    field_size=TableFieldSize(tid);  
    }

  printf("OtrDbGet[info]: get table %s from server %s, "
           "number of item: %d  major:%d minor:%d match:%d\n",
     tname,t->generic_servername,field_size,t->major,t->minor,t->match);
    
  char *d;
  if (!data) d=(char*) malloc(field_size*size);
  else if (field_size!=*nitem) d=(char*) realloc(data,field_size*size);
  else d=(char*)data;
  for (i=0;i<TableLabels(tid,bt);i++) {
    int   field_type=0;
    char *ff=f[i].fmt;
    field_type=OtrDbFieldType(ff);     
    if (field_type!=TABLEstring) {
      GetTableStruct(tid,field_type,size,field_size,d+f[i].ofs);
      }
    else {	   
      char **s;
      int k;
      s=(char **) malloc (field_size * sizeof(char *));
      if (s==0){
        fprintf(stderr,"Memory was not allocated\n");
        exit(0);
	}
      s[0]=0;
      GetTableField(tid,TABLEstring,field_size,s);
      if (!strcmp(ff,"char*")) {
        for (k=0; k<field_size; k++) {
	  *((char**)(d+f[i].ofs+k*size))=strdup(s[k]);
	  }  
	}
      else { 
        int nsize;
	char *c=d+f[i].ofs-size;
        if (i<table->nfield-1)  nsize=f[i+1].ofs-f[i].ofs;
        else nsize=size-f[i].ofs;
        for (k=0; k<field_size; k++){
	  strncpy((c+=size),s[k],nsize);
	  }  
	}
//      free(s[0]);
      free(s);
      }
    } 
  *nitem=field_size;
  return d;    
  }  

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
int OtrDbPut(const OtrDbTable* table,void *data,int nitem) {   
  char name[256];
  const OtrDbField* f=table->f;
  int nfield=table->nfield;
  int size=table->size;
  char* comment=(char*)table->comment;
  int field_size=nitem,i,bt;
  Tables tid;
  OtrDbTableStruct* t;
  char *format=(char*)malloc(sizeof(char)); 
  strcpy (format,"");
  char *d=(char*)data;

  
  if (!data) {
    printf("OtrDbPut[warning]: Data invalid, address=0x%p\n",data);
    return OtrIInvalid();
    }
    
  for (i=0;i<numtables;i++) {
    t=&tables[i];
    if (!strcmp(t->generic_tablename,table->name)) break;
    } 
  for (i=0;i<nfield;i++) {
    if (!strcmp(f[i].fmt,"char") ) { 
	   format=(char*)realloc(format,strlen(format)+4*sizeof(char)+strlen(f[i].field_name)); 
	   strcat (format,"b:");strcat (format,f[i].field_name);
	   }
    else if (!strcmp(f[i].fmt,"char[]")){ 
	   format=(char*)realloc(format,strlen(format)+4*sizeof(char)+strlen(f[i].field_name)); 
	   strcat (format,"a:");strcat (format,f[i].field_name);
	   }
    else if (!strcmp(f[i].fmt,"short")){ 
	   format=(char*)realloc(format,strlen(format)+4*sizeof(char)+strlen(f[i].field_name)); 
	   strcat (format,"s:");strcat (format,f[i].field_name);
	   }
    else if (!strcmp(f[i].fmt,"int")){ 
	   format=(char*)realloc(format,strlen(format)+4*sizeof(char)+strlen(f[i].field_name)); 
	   strcat (format,"i:");strcat (format,f[i].field_name);
	   }
    else if (!strcmp(f[i].fmt,"unsigned char")){ 
	   format=(char*)realloc(format,strlen(format)+5*sizeof(char)+strlen(f[i].field_name)); 
	   strcat (format,"ub:");strcat (format,f[i].field_name);
	   }
    else if (!strcmp(f[i].fmt,"unsigned short")){ 
	   format=(char*)realloc(format,strlen(format)+5*sizeof(char)+strlen(f[i].field_name)); 
	   strcat (format,"us:");strcat (format,f[i].field_name);
	   }
    else if (!strcmp(f[i].fmt,"unsigned int")){ 
	   format=(char*)realloc(format,strlen(format)+5*sizeof(char)+strlen(f[i].field_name)); 
	   strcat (format,"ui:");strcat (format,f[i].field_name);
	   }
    else if (!strcmp(f[i].fmt,"float")){ 
	   format=(char*)realloc(format,strlen(format)+4*sizeof(char)+strlen(f[i].field_name)); 
	   strcat (format,"f:");strcat (format,f[i].field_name);
	   }
    else if (!strcmp(f[i].fmt,"double")){ 
	   format=(char*)realloc(format,strlen(format)+4*sizeof(char)+strlen(f[i].field_name)); 
	   strcat (format,"d:");strcat (format,f[i].field_name);
	   }
    else if (!strcmp(f[i].fmt,"ascii")){ 
	   format=(char*)realloc(format,strlen(format)+4*sizeof(char)+strlen(f[i].field_name)); 
	   strcat (format,"a:");strcat (format,f[i].field_name);
	   }
      strcat (format," ");
    } 
  
  if ((OtrDbAlgorithm&OtrDbTest)||!OtrDbWritePermission(t->generic_servername)) {
    if ((OtrDbAlgorithm&OtrDbTest)) {
      strcpy(name,"/OTR_TEST");
      printf("OtrDbPut[info]: YOU HAVE NO WRITE PERMISSION TO THE SERVER %s !.\n",
             t->generic_servername);
      printf("OtrDbPut[info]: Running in test mode: change server %s to %s\n\n",
        t->generic_servername,name);
      }
    else {
      printf("\nOtrDbPut[info]: YOU HAVE NO WRITE PERMISSION TO THE SERVER %s !.\n",
             t->generic_servername);
      return OtrIInvalid();
      }
    }
  else {
    strcpy(name,t->generic_servername);
    }

  tid=OpenTables(name,0);
  if(tid==0) {
    printf("Can't open database\n");
    return OtrIInvalid();
    }

  if (OtrDbKeyPermission()) {
    DistributeTables(tid);
    }
  char tname[256];
  if (table->write_version!=0) {
    sprintf(tname,"%s_v%d",t->generic_tablename,table->write_version);
    }
  else strcpy(tname,t->generic_tablename);
   //printf("tname=%s t->version=%d\n",tname,table->write_version);
   
  printf("OtrDbPut[info]: Update %s/%s\n",name,tname);
  bt=BookTable(tid,tname,format,comment);
  if(bt==0) {
    printf("OtrDbPut: Can't book table: server: %s table: %s\n",name,tname);
    return OtrIInvalid();
    }
  if (OtrDbAlgorithm&OtrDbAutomaticKey) TableStamp(tid,TABLEsec,0);
  else {
    TableStamp(tid,t->major,t->minor);
    }
    /* If SetKey mechanism is used, then disable user key now */
  OtrDbAlgorithm|= OtrDbAutomaticKey; 
  
  SkipTableField(tid,1);

  for (i=0;i<nfield;i++) {
    int   field_type=0;
    char *ff=f[i].fmt;
    field_type=OtrDbFieldType(ff);     
    if (field_type!=TABLEstring) {
      PutTableStruct(tid,field_type,size,field_size,d+f[i].ofs);
      }
    else {	   
      int k,m,nsize,j=0;
      char *c=d+f[i].ofs;
      
/* check if all strings are zero detminated */      

      if (i<nfield-1)  nsize=f[i+1].ofs-f[i].ofs;
      else nsize=size-f[i].ofs;
      for (k=0; k<field_size; k++) {
        for(m=0;m<nsize;m++){
          if (c[m]=='0') {
	    j++;
	    break;
	    }
	  }
	if (m==nsize) break;
	c+=size;
	}
	
      char **s;
      s= (char **) malloc (field_size * sizeof(char *));
      if (j==field_size){	   
	if (!strcmp(ff,"char*")) {
          for (k=0; k<field_size; k++) {
	    s[k]=*((char**)(d+f[i].ofs+k*size));
		//printf("((char*)(d+f[i].ofs+k*size))=%s\n",*((char**)(d+f[i].ofs+k*size)));
            }
	  }
	else {
          for (k=0; k<field_size; k++){ 
	    s[k]=((char*)(d+f[i].ofs+k*size)); 
		//printf("s[%d]=%s offset=%d d=%d f[i].ofs=%d k*size=%d\n",k, s[k],d+f[i].ofs+k*size,d,f[i].ofs,k*size);
	    }
	  }
        PutTableField(tid,TABLEstring,field_size,s);
	}		 
      else {
        int  nmax=1024;
        char *data_string=(char *)malloc(nmax*sizeof(char));
	char *c=d+f[i].ofs-size;
	int  fsize=0;
        for(k=0;k<field_size;k++){
          if(fsize +nsize+1 >= nmax) {
            nmax=nmax+1024;
            data_string=(char *)realloc(data_string,nmax*sizeof(char));
            }    
          memcpy(data_string+fsize,(c+=size),nsize);
	  fsize+=nsize;
          *(data_string+fsize)=0;
	  fsize+=1;
          }
        for(k=0;k<field_size;k++){
	  s[k]=data_string+k*(nsize+1);
	  }
        PutTableField(tid,TABLEstring,field_size,s);
	free(data_string);
        }
      free(s);
      }  
    }
  if (!StoreTable(tid,bt, 0)) {
    printf("Table:%s didn't store\n",t->generic_tablename);
    }
  free(format);
  CloseTables(tid);
  return i;    
  }  
#ifdef __cplusplus
}
#endif
