#include "OtrDecode.h"
#include "Otr/OtrI.h"
#include "Otr/OtrDb.h"

#define OTRIInvalidInt -1

#ifdef __cplusplus
extern "C" {
  void  OtrDecodeSetAlgorithm_(int* i)  { OtrDecodeSetAlgorithm(*i); }
  int   OtrDecodeInitialize_()          { return OtrDecodeInitialize(); }
  int   OtrDecodeReinitialize_()        { return OtrDecodeReinitialize(); }
  int   OtrDecodeWriteToDb_()           { return OtrDecodeWriteToDb(); }
  int   OtrDecodeClose_()               { return OtrDecodeClose(); }
  int   OtrDecodeTestMe_()              { return OtrDecodeTestMe(); }
}
#endif
/*==========================================================================*/ OtrDbField structcabfmt[]={
  {"char[]","gencha1",OtrDbOffset(OtrDecodeCabStruct,gencha1)},
  {"char[]","gencha2",OtrDbOffset(OtrDecodeCabStruct,gencha2)},
  {"char[]","modcha1",OtrDbOffset(OtrDecodeCabStruct,modcha1)},
  {"char[]","modcha",OtrDbOffset(OtrDecodeCabStruct,modcha2)},
  {"char[]","modhv",OtrDbOffset(OtrDecodeCabStruct,modhv)},
  {"char[]","ftbpos",OtrDbOffset(OtrDecodeCabStruct,ftbpos)},
  {"char[]","ftbasd",OtrDbOffset(OtrDecodeCabStruct,ftbasd)},
  {"char[]","ftbhv",OtrDbOffset(OtrDecodeCabStruct,ftbhv)},
  {"char[]","ftbcab",OtrDbOffset(OtrDecodeCabStruct,ftbcab)},
  {"char[]","lvpos",OtrDbOffset(OtrDecodeCabStruct,lvpos)},
  {"char[]","lvcab",OtrDbOffset(OtrDecodeCabStruct,lvcab)},
  {"char[]","tdccab",OtrDbOffset(OtrDecodeCabStruct,tdccab)},
  {"char[]","tdccon",OtrDbOffset(OtrDecodeCabStruct,tdccon)},
  {"char[]","previous1",OtrDbOffset(OtrDecodeCabStruct,previous1)},
  {"char[]","previous2",OtrDbOffset(OtrDecodeCabStruct,previous2)},
  {"char[]","next1",OtrDbOffset(OtrDecodeCabStruct,next1)},
  {"char[]","next2",OtrDbOffset(OtrDecodeCabStruct,next2)},
  {"char[]","status",OtrDbOffset(OtrDecodeCabStruct,status)},
  {"char[]","hvgroup",OtrDbOffset(OtrDecodeCabStruct,hvgroup)},
  {"char[]","hv1",OtrDbOffset(OtrDecodeCabStruct,hv1)},
  {"char[]","hv2",OtrDbOffset(OtrDecodeCabStruct,hv2)},
  {"char[]","hvcab1",OtrDbOffset(OtrDecodeCabStruct,hvcab1)},
  {"char[]","hvcab2",OtrDbOffset(OtrDecodeCabStruct,hvcab2)},
};
OtrDbTable CableTableDef={"OTR_Cabling",1,
			  "cabling table",sizeof(OtrDecodeCabStruct),
			  structcabfmt,
			  sizeof(structcabfmt)/sizeof(OtrDbField)};

OtrDbField structsharcfmt[]={
  {"char[]","TDCboard",OtrDbOffset(OtrDecodeSharcStruct,tdccon)},
  {"unsigned short","FEDid",OtrDbOffset(OtrDecodeSharcStruct,FedID)},
  {"unsigned char","VME",OtrDbOffset(OtrDecodeSharcStruct,VME)},
  {"unsigned short","Boardid",OtrDbOffset(OtrDecodeSharcStruct,SharcBoardID)},
  {"unsigned char","Port",OtrDbOffset(OtrDecodeSharcStruct,Port)},
  {"unsigned char","SlbId",OtrDbOffset(OtrDecodeSharcStruct,SLB)},
};
OtrDbTable SharcTableDef={"OTR_Sharc",1,"Sharc cable table",
			  sizeof(OtrDecodeSharcStruct),structsharcfmt,
			  sizeof(structsharcfmt)/sizeof(OtrDbField)};


/*==========================================================================*/ 

void OtrDecodeDefine() {
  OtrDbDefine("/OTR_CONFIG",CableTableDef.name,OtrDecodeUpdate);
  OtrDbDefine("/OTR_CONFIG",SharcTableDef.name,OtrDecodeUpdate);
  }


/*==========================================================================*/ 
OtrDecodeSharcStruct *sharcdata;
OtrDecodeCabStruct *cabdata;
/*==========================================================================*/ 
static int OtrDecodeAlgorithm=OTRIInvalidInt;
static OtrDecodeStruct *OTR;
const OtrDecodeStruct* OtrDecodeGet(){ return OTR; }
int DecodeDebug=0;
static int IsOtrDecodeInitialize=0;
/*==========================================================================*/
void OtrDecodeSetAlgorithm(int i){
  OtrDecodeAlgorithm=i;
}
  
/*==========================================================================*/
int OtrDecodeReinitialize(){
int i;
  printf("OtrDecodeReinitialize[debug]: enter func\n");
  if(OtrDecodeAlgorithm==OTRIInvalidInt) return OTRIInvalidInt;
  printf("OtrDecodeReinitialize[debug]: continue\n");
  if (IsOtrDecodeInitialize) {
#ifdef __cplusplus
    delete OTR->ASD8s;
    delete OTR->TDCs;  
    delete OTR->FEDs;
    for(i=0;i<32;i++) delete OTR->Slayers[i].Line;
    delete OTR->Slayers;
    delete OTR->Geo.FirstCellInMonolayer;
    delete OTR->Geo.CiArray;
    delete OTR;
#endif
#ifndef __cplusplus
    free(OTR->ASD8s);
    free(OTR->TDCs);  
    free(OTR->FEDs);
    for(i=0;i<32;i++) free(OTR->Slayers[i].Line);
    free(OTR->Slayers);
    free(OTR->Geo.FirstCellInMonolayer);
    free(OTR->Geo.CiArray);
    free(OTR);
#endif
    }
  IsOtrDecodeInitialize=0;
  printf("OtrDecodeReinitialize[debug]: call init\n");
  return OtrDecodeInitialize();
}

/*==========================================================================*/

void check_env() {
  char *t=getenv("OtrDecodeSharc");
  char s[1024];
  OtrDecodeAlgorithm=0;
  if (!t) t=getenv("OTRDECODEUseSharc");
  else {
    sprintf(s,"OTRDECODEUseSharc=%s",t);
    putenv(s);
    }
  if (!t||t[0]=='\0') OtrDecodeAlgorithm|=SHARCTAB;
  else OtrDecodeAlgorithm|=SHARCTABFILE;
  t=getenv("OtrDecodeCables");
  if (!t) t=getenv("OTRDECODEUseCables");
  else {
    sprintf(s,"OTRDECODEUseCables=%s",t);
    putenv(s);
    }
  if (!t||t[0]=='\0') OtrDecodeAlgorithm|=CABTAB;
  else OtrDecodeAlgorithm|=CABTABFILE;  
  }
/*==========================================================================*/
int OtrDecodeUpdate()  { 
  printf("OtrDecodeUpdate[debug]: check environement\n");
  if (OtrDecodeAlgorithm==OTRIInvalidInt) check_env();
  printf("OtrDecodeUpdate[debug]: call reinit\n");
  return OtrDecodeReinitialize();
  };
  
int OtrDecodeClose(){ 
int i;
  if (OtrDecodeAlgorithm==OTRIInvalidInt) return OTRIInvalidInt;
  free(OTR->ASD8s);
  free(OTR->TDCs);  
  free(OTR->FEDs);
  for(i=0;i<32;i++) free(OTR->Slayers[i].Line);
  free(OTR->Slayers);
  free(OTR->Geo.FirstCellInMonolayer);
  free(OTR->Geo.CiArray);
  free(OTR);
  IsOtrDecodeInitialize=0;
  OtrDecodeAlgorithm=OTRIInvalidInt;
  return 1;
}
  
/*==========================================================================*/
int OtrDecodeWriteToDb(){ 
printf("Decode: %d\n",OtrDecodeAlgorithm);
  if (OtrDecodeAlgorithm==OTRIInvalidInt) return OTRIInvalidInt;
  if((OtrDecodeAlgorithm&0x0003)==0x0003){
    if(DecodeDebug)
      printf("OtrDecodeWriteToDb: Writing Cabling data to database...\n");
    OtrDbPut(&CableTableDef,cabdata,OtrDecodeGet()->NASDs);
  }else{
    printf("OtrDecodeWriteToDb: Cabling data are not defined...\n");
  }
  if((OtrDecodeAlgorithm&0x000C)==0x000C){
    if(DecodeDebug)
      printf("OtrDecodeWriteToDb: Writing Sharc data to database...\n");
    OtrDbPut(&SharcTableDef,sharcdata,OtrDecodeGet()->NTDCs);
  }else{
    printf("OtrDecodeWriteToDb: Sharc data are not defined...\n");
  }
  return OTRIInvalidInt;
}
  
  
/*==========================================================================*/
OtrDecodeCabStruct *OtrDecodeReadCab(int *n){
char *list,tab[500],str[500];
FILE *f,*df;
OtrDecodeCabStruct *data=0;
int ind,i,j;

  if((OtrDecodeAlgorithm&0x0001)==0x0001){
    if((OtrDecodeAlgorithm&0x0002)==0x0002){ /*---> data source - file <----*/
      if(DecodeDebug)printf("Reading cabling data from ASCII files...\n"); 
      list=getenv("OTRDECODEUseCables"); 
#ifndef __cplusplus
      data=malloc(sizeof(OtrDecodeCabStruct)*OTRMAXASD);
#endif
#ifdef __cplusplus
      data=new OtrDecodeCabStruct[OTRMAXASD];
#endif
      memset(data,0,sizeof(OtrDecodeCabStruct)*OTRMAXASD);
      if(data==0){
	if(DecodeDebug)printf("OtrDecodeReadCab:Memory allocation error.\n");
	return 0; 
      }
      for(ind=0;ind<OTRMAXASD;ind++) data[ind].gencha1[0]=0;
      f=fopen(list,"r");
      if(f==NULL){
	if(DecodeDebug) printf("OtrDecodeReadCab:Can't open file: %s\n",list);
	return 0;
      }
      ind=0;
      do{
	if(fscanf(f,"%s",tab)==EOF) break;
	df=fopen(tab,"r");
	if(df==NULL){
	  if(DecodeDebug) printf("OtrDecodeReadCab:Can't open file: %s\n",tab);
	  return 0;
	}
	if(DecodeDebug) printf("OtrDecodeReadCab:Reading data from: %s\n",tab);
	fgets(str,499,df);
	do{
	  if(fgets(str,499,df)==NULL) break;
	  if(str[0]!='=' && str[0]!='-' && (strlen(str)>100)){
	    sscanf(str,"%i %s %s %s %s %s %s %s %s %s %s %s %s %s %s "
                   "%s %s %s %s %s %s %s %s %s",&i,
		   data[ind].gencha1,data[ind].gencha2,data[ind].modcha1,
		   data[ind].modcha2,data[ind].modhv,data[ind].ftbpos,
		   data[ind].ftbasd,data[ind].ftbhv,data[ind].ftbcab,
		   data[ind].lvpos,data[ind].lvcab,data[ind].tdccab,
		   data[ind].tdccon,data[ind].previous1,data[ind].previous2,
		   data[ind].next1,data[ind].next2,data[ind].status,
		   data[ind].hvgroup,data[ind].hv1,data[ind].hv2,
		   data[ind].hvcab1,data[ind].hvcab2);
	    if(data[ind].previous1[4]=='0'){ 
	      for(i=5;i<11;i++) data[ind].previous1[i]=' ';
	      data[ind].previous1[11]=0;
	    }
	    if(data[ind].previous2[4]=='0'){ 
	      for(i=5;i<11;i++) data[ind].previous2[i]=' ';
	      data[ind].previous2[11]=0;
	    }
	    if(data[ind].next1[4]=='0'){ 
	      for(i=5;i<11;i++) data[ind].next1[i]=' ';
	      data[ind].next1[11]=0;
	    }
	    if(data[ind].next2[4]=='0'){ 
	      for(i=5;i<11;i++) data[ind].next2[i]=' ';
	      data[ind].next2[11]=0;
	    }
	    for(i=0;i<12;i++)if(data[ind].hvgroup[i]==0) break;
	    if(i!=12) for(j=i;j<12;j++) data[ind].hvgroup[j]=' ';
	    data[ind].hvgroup[12]=0;
	    for(i=0;i<11;i++)if(data[ind].hvcab1[i]==0) break;
	    if(i!=11) for(j=i;j<11;j++) data[ind].hvcab1[j]=' ';
	    data[ind].hvcab1[11]=0;
	    for(i=0;i<11;i++)if(data[ind].hvcab2[i]==0) break;
	    if(i!=11) for(j=i;j<11;j++) data[ind].hvcab2[j]=' ';
	    data[ind].hvcab2[11]=0;
	    ind++;
	    if(ind==OTRMAXASD){
	      if(DecodeDebug)printf("OtrDecodeReadCab:Too many ASDs ...\n");
	      return 0;
	    }
	  }
	}while(1);
	fclose(df);
      }while(1);
      fclose(f);
      *n=ind;
      return data;
    }else{ /*----> data source - database <----*/
      int version;
      if(DecodeDebug)printf("Reading cabling data from database...\n");
      if ((version=OtrDbGetTableVersion(CableTableDef.name))!=-1) {
        printf("OtrDecodeReadCab[debug]: Get cable table from db\n");
        data=(OtrDecodeCabStruct*)OtrDbGet(&CableTableDef,data,&ind);
        printf("OtrDecodeReadCab[debug]: Got cable table from db\n");
        *n=ind;
      }
      return data;
    }
  }else{
    *n=0;
    return 0;
  }  
  return 0;
}

/*==========================================================================*/
OtrDecodeSharcStruct *OtrDecodeReadSharc(int *n){
char *tab;
char str[256],tmp[4],slb[8];
FILE *f;
int ind,fed,vme,sboard;
OtrDecodeSharcStruct *data=0;

  if((OtrDecodeAlgorithm&0x0004)==0x0004){
    if((OtrDecodeAlgorithm&0x0008)==0x0008){ /*---> data source - file <----*/
      tab=getenv("OTRDECODEUseSharc"); 
      f=fopen(tab,"r");
      if(f==NULL){
	if(DecodeDebug) printf("OtrDecodeReadSharc:Can't open file: %s\n",tab);
	return 0;
      }
      if(DecodeDebug) printf("OtrDecodeReadSharc:Reading data from: %s\n",tab);
#ifndef __cplusplus
      data=malloc(sizeof(OtrDecodeSharcStruct)*OTRMAXTDC);
#endif
#ifdef __cplusplus
      data=new OtrDecodeSharcStruct[OTRMAXTDC];
#endif
      memset(data,0,sizeof(OtrDecodeSharcStruct)*OTRMAXTDC);
      if(data==0){
	if(DecodeDebug)printf("OtrDecodeReadSharc:Memory allocation error.\n");
	return 0;
      }
      for(ind=0;ind<OTRMAXTDC;ind++) data[ind].tdccon[0]=0;
      ind=0;
      for(;;){
	if(fgets(str,255,f)==NULL) break;
	if(str[0]=='+' || str[0]=='-'){
	  sscanf(str,"%s %i %i %i %s %s",
		 data[ind].tdccon,&fed,&vme,&sboard,tmp,slb);
	  data[ind].tdccon[8]=0;
	  data[ind].FedID=fed;
	  data[ind].VME=vme;
	  data[ind].SharcBoardID=sboard;
	  data[ind].Port=(atoi(tmp)-1)*6+atoi(tmp+2);
	  data[ind].SLB=atoi(slb+4);
	  ind++;
	  if(ind==OTRMAXTDC){
	    if(DecodeDebug)printf("OtrDecodeReadSharc:Too many boards ...\n");
	    return 0;
	  }
	}
      }
      fclose(f);
      *n=ind;
      return data;
    }else{ /*----> data source - database <----*/
      int version;
      if(DecodeDebug)printf("Reading Sharc data from database...\n"); 
      if ((version=OtrDbGetTableVersion(SharcTableDef.name))!=-1) {
        data=(OtrDecodeSharcStruct*)OtrDbGet(&SharcTableDef,data,&ind);
        *n=ind;
      }
      return data;
    }
  }else{
    *n=0;
    return 0;
  }  
  return 0;
}

/*==========================================================================*/
int OtrDecodeInitialize(){
int Nasd,Ntdc; 

  if(IsOtrDecodeInitialize){
    if(DecodeDebug) 
      printf("OtrDecodeInitialize: OtrDecode allready initialized.\n");
    return OTRIInvalidInt;
  }
  if (OtrDecodeAlgorithm==OTRIInvalidInt) check_env();
  if(OtrDecodeAlgorithm==OTRIInvalidInt) return OTRIInvalidInt;
#ifndef __cplusplus
  OTR=malloc(sizeof(OtrDecodeStruct));
#endif
#ifdef __cplusplus
  OTR=new OtrDecodeStruct;
#endif
  memset(OTR,0,sizeof(OtrDecodeStruct));
  if(OTR==0){
    if(DecodeDebug) printf("OtrDecodeInitialize:Memory allocation error.\n");
    return OTRIInvalidInt;
  }
  OTR->Algorithm=OtrDecodeAlgorithm;

  cabdata=OtrDecodeReadCab(&Nasd);
  if(cabdata==0){
    if(DecodeDebug)
      printf("OtrDecodeInitialize:OtrDecodeReadCab returned error.\n");
    return OTRIInvalidInt;
  }
  else printf("OtrDecode[debug]: read cable data. Now fill asd\n"); 
  OTR->NASDs=Nasd;
  if(!OtrDecodeFillAsds(cabdata)){
    printf("OtrDecodeInitialize:OtrDecodeFillAsds returned error.\n");
    return OTRIInvalidInt;
  }

  sharcdata=OtrDecodeReadSharc(&Ntdc);
  if(sharcdata==0){
    if(DecodeDebug)
      printf("OtrDecodeInitialize:OtrDecodeReadSharc returned error.\n");
    return OTRIInvalidInt;
  }
  else printf("OtrDecode[debug]: read cable data. now fill tdc's\n"); 
  OTR->NTDCs=Ntdc;
  if(!OtrDecodeFillTdcs(cabdata,sharcdata)){
    if(DecodeDebug)
      printf("OtrDecodeInitialize:OtrDecodeFillTdcs returned error.\n");
    return OTRIInvalidInt;
  }

  if(!OtrDecodeFillWire(cabdata)){
    if(DecodeDebug)
      printf("OtrDecodeInitialize:OtrDecodeFillWire returned error.\n");
    return OTRIInvalidInt;
  }
  if(!IsOtrDecodeInitialize) IsOtrDecodeInitialize=1;
  return OtrDecodeAlgorithm;
}  

/*==========================================================================*/
int OtrDecodeFillAsds(OtrDecodeCabStruct *data){
int i,j,k,pos,NLay,sl;
char *names[32]={"   ","-m1","-m2","-m3","-m4","-m5","-m6","   ",
		 "-m8","-p1","-p2","-p3","-p4","-t1","-t2","   ",
		 "   ","+m1","+m2","+m3","+m4","+m5","+m6","   ",
		 "+m8","+p1","+p2","+p3","+p4","+t1","+t2","   "};
#ifndef __cplusplus
  OTR->ASD8s=malloc(sizeof(OtrASD8Struct)*OTR->NASDs);
#endif
#ifdef __cplusplus
  OTR->ASD8s=new OtrASD8Struct[OTR->NASDs];
#endif
  memset(OTR->ASD8s,0,sizeof(OtrASD8Struct)*OTR->NASDs);
  if(OTR->ASD8s==0){
    if(DecodeDebug) printf("OtrDecodeFillAsds:Memory allocation error.\n");
    return 0;
  }
  for(i=0;i<32;i++) strncpy(OTR->SLname[i],names[i],4);
  for(i=0;i<OTR->NASDs;i++){

    for(k=0;k<16;k++) OTR->ASD8s[i].wire[k].ChIndex=OtrIFromAsd(i,k);

    sl=0; 
    if(data[i].gencha1[1]=='p') sl+=8;
    if(data[i].gencha1[1]=='t') sl+=12;
    if(data[i].gencha1[0]=='+') sl+=16;
    sl+=atoi(data[i].gencha1+2);
    OTR->ASD8s[i].SuperLayer=sl;
    OTR->ASD8s[i].Layer=atoi(data[i].modcha1+4);
    OTR->ASD8s[i].Sector=atoi(data[i].modcha1+9);
    OTR->ASD8s[i].Module=atoi(data[i].modcha1+6);
    OTR->ASD8s[i].ftb_ASD=atoi(data[i].ftbasd+9);
    OTR->ASD8s[i].TDC_Connector=atoi(data[i].tdccon+9);
    strncpy(OTR->ASD8s[i].hwc,data[i].modcha1,11);
    OTR->ASD8s[i].hwc[11]=0;
    strncpy(OTR->ASD8s[i].lvpos,data[i].lvpos,10);
    strncpy(OTR->ASD8s[i].HVcab1,data[i].hvcab1,12); 
    strncpy(OTR->ASD8s[i].HVcab2,data[i].hvcab2,12); 
    strncpy(OTR->ASD8s[i].HVgroup1,data[i].hv1,14);
    strncpy(OTR->ASD8s[i].HVgroup2,data[i].hv1,14);
  }
#ifndef __cplusplus
  OTR->Slayers=malloc(sizeof(OtrSuperLayerStruct)*32);
#endif
#ifdef __cplusplus
  OTR->Slayers=new OtrSuperLayerStruct[32];
#endif
  memset(OTR->Slayers,0,sizeof(OtrSuperLayerStruct)*32);
  if(OTR->Slayers==0){
    if(DecodeDebug) printf("OtrDecodeFillAsds:Memory allocation error.\n");
    return 0;
  }

  for(i=0;i<32;i++){
    OTR->Slayers[i].NLayer=0; 
    for(k=0;k<13;k++) for(j=0;j<7;j++){
      OTR->Slayers[i].Nasd[j][k]=0; 
      OTR->Slayers[i].FirstASD[j][k]=255;
    }
  }
  for(j=0;j<OTR->NASDs;j++){
    OTR->Slayers[OTR->ASD8s[j].SuperLayer].Nasd[OTR->ASD8s[j].Layer]
      [OTR->ASD8s[j].Sector]++;
    if(OTR->Slayers[OTR->ASD8s[j].SuperLayer].FirstASD[OTR->ASD8s[j].Layer]
       [OTR->ASD8s[j].Sector] > OTR->ASD8s[j].ftb_ASD)
      OTR->Slayers[OTR->ASD8s[j].SuperLayer].FirstASD[OTR->ASD8s[j].Layer]
        [OTR->ASD8s[j].Sector]=OTR->ASD8s[j].ftb_ASD;
    OTR->SLs[OTR->ASD8s[j].SuperLayer]=1;
  }
  for(i=0;i<32;i++){
    NLay=0;
    pos=1;
    for(k=0;k<13;k++) for(j=0;j<7;j++){
      if(OTR->Slayers[i].FirstASD[j][k]<255){
        OTR->Slayers[i].pos[j][k]=pos;
        pos+=OTR->Slayers[i].Nasd[j][k];
        if(NLay<j) NLay=j;
      }    
    }
#ifndef __cplusplus
    OTR->Slayers[i].Line=malloc(pos*sizeof(unsigned short));
#endif
#ifdef __cplusplus
    OTR->Slayers[i].Line=new unsigned short[pos];
#endif
    OTR->Slayers[i].NLayer=NLay;
  }
  for(i=0;i<OTR->NASDs;i++){
    OTR->Slayers[OTR->ASD8s[i].SuperLayer].Line[
      OTR->Slayers[OTR->ASD8s[i].SuperLayer].
      pos[OTR->ASD8s[i].Layer][OTR->ASD8s[i].Sector]-
      OTR->Slayers[OTR->ASD8s[i].SuperLayer].
      FirstASD[OTR->ASD8s[i].Layer][OTR->ASD8s[i].Sector]+
      OTR->ASD8s[i].ftb_ASD]=i;
  }
  return 1;
}

/*==========================================================================*/
int OtrDecodeFillTdcs(OtrDecodeCabStruct *cab,OtrDecodeSharcStruct *sharc){
int i,j;

#ifndef __cplusplus
  OTR->TDCs=malloc(sizeof(OtrTDCStruct)*OTR->NTDCs);
#endif
#ifdef __cplusplus
  OTR->TDCs=new OtrTDCStruct[OTR->NTDCs];
#endif
  memset(OTR->TDCs,0,sizeof(OtrTDCStruct)*OTR->NTDCs);
  if(OTR->TDCs==0){
    if(DecodeDebug) printf("OtrDecodeFillTdcs:Memory allocation error.\n");
    return 0;
  }
#ifndef __cplusplus
  OTR->FEDs=malloc(sizeof(unsigned short)*0xFFFF);
#endif
#ifdef __cplusplus
  OTR->FEDs=new unsigned short[0xFFFF];
#endif
  memset(OTR->FEDs,0,sizeof(unsigned short)*0xFFFF);
  if(OTR->FEDs==0){
    if(DecodeDebug) printf("OtrDecodeFillTdcs:Memory allocation error.\n");
    return 0;
  }
  for(i=0;i<0xFFFF;i++) OTR->FEDs[i]=0xFFFF;
  for(i=0;i<OTR->NTDCs;i++) for(j=0;j<8;j++) OTR->TDCs[i].ASD[j]=0xFFFF;

  for(i=0;i<OTR->NTDCs;i++){
    for(j=0;j<OTR->NASDs;j++){
      if(strncmp(cab[j].tdccon,sharc[i].tdccon,8)==0){
	OTR->TDCs[i].VME_num=sharc[i].VME;
	OTR->TDCs[i].Sharc_Board=sharc[i].SharcBoardID;
	OTR->TDCs[i].Sharc_Connector=sharc[i].Port;
	OTR->TDCs[i].FED=sharc[i].FedID;
	OTR->TDCs[i].crate=atoi(sharc[i].tdccon+4);
	OTR->TDCs[i].slot=atoi(sharc[i].tdccon+6);
	OTR->TDCs[i].SLB=sharc[i].SLB;
	OTR->ASD8s[j].TDC_FED=sharc[i].FedID;
	OTR->FEDs[sharc[i].FedID]=i;
	OTR->TDCs[i].ASD[OTR->ASD8s[j].TDC_Connector-1]=j;
      }
    }
  }
  return 1;
}


int OtrDecodeFillWire(OtrDecodeCabStruct *cab){
int Nnede,i,k,igede,module,modid,wire,mono,mod_first_wire,asd_first_wire,nwi[4];
unsigned short Line;
int NumOfCell=0;
char hwc[12];
int spmod=0,fw=0;
  Nnede=OtrGeoNumberOfVolumes();
  if(Nnede==0){
    if(DecodeDebug) printf("OtrDecodeFillWire:OtrGeo is not initialized...\n");
    return 0;
  }
  OTR->Geo.NumOfVol=Nnede;
#ifndef __cplusplus
  OTR->Geo.FirstCellInMonolayer=malloc(sizeof(int)*Nnede*4);
#endif
#ifdef __cplusplus
  OTR->Geo.FirstCellInMonolayer=new int[Nnede*4];
#endif
  memset(OTR->Geo.FirstCellInMonolayer,0,sizeof(int)*Nnede*4);
  for(igede=0;igede<Nnede;igede++) for(k=0;k<4;k++){
    OtrIGeoIdFromIndex(igede,hwc); 
    if(hwc[0]!='n'){ 
      if(k<OtrIMonoLayerNumberOf(igede)){
	OTR->Geo.FirstCellInMonolayer[igede<<2|k]=NumOfCell;
      }else OTR->Geo.FirstCellInMonolayer[igede<<2|k]=-1;
      NumOfCell+=OtrIMonoLayerNumberOfWires(igede,k);
    }
  }

#ifndef __cplusplus
  OTR->Geo.CiArray=malloc(sizeof(int)*NumOfCell);
#endif
#ifdef __cplusplus
  OTR->Geo.CiArray=new int[NumOfCell];
#endif
  memset(OTR->Geo.CiArray,0,sizeof(int)*NumOfCell);
 
  for(igede=0;igede<Nnede;igede++){
    for(k=0;k<4;k++) nwi[k]=OtrIMonoLayerNumberOfWires(igede,k);
    OtrIGeoIdFromIndex(igede,hwc);

    for(i=0;i<OTR->NASDs;i++){
        if(!strncmp(hwc,OTR->ASD8s[i].hwc,12)){
	  mod_first_wire=atoi(cab[i].modcha1+12);
	  asd_first_wire=atoi(cab[i].gencha1+9)%16-1;
          for(k=asd_first_wire;k<16;k++){
            if(OtrIMonoLayerNumberOf(igede)==2 && hwc[0]=='-'){ 
	      modid=OTR->ASD8s[i].Module;
              module=OtrDecodeAsd2Nede(0,mod_first_wire,
                                       k+1-asd_first_wire,0,nwi,&mono,&wire);
              if(wire<nwi[mono]){
                OTR->ASD8s[i].wire[k].NEDE=igede;
                OTR->ASD8s[i].wire[k].ChID=
		  OtrIChannelId(OTR->ASD8s[i].SuperLayer,
				OTR->ASD8s[i].Layer,OTR->ASD8s[i].Sector,
				modid,mono,wire,OTR->ASD8s[i].ftb_ASD);
		OTR->Geo.CiArray[OTR->Geo.FirstCellInMonolayer[igede<<2|mono]+wire]=
		  (0x80000000|(k<<16)|i);
              }
	      continue;
            }
            if(OtrIMonoLayerNumberOf(igede)==2 && hwc[0]=='+'){ 
	      modid=OTR->ASD8s[i].Module;
              module=OtrDecodeAsd2Nede(0,mod_first_wire,
                                       k+1-asd_first_wire,1,nwi,&mono,&wire);
              if(wire<nwi[mono]){
                OTR->ASD8s[i].wire[k].NEDE=igede;
                OTR->ASD8s[i].wire[k].ChID=
		  OtrIChannelId(OTR->ASD8s[i].SuperLayer,
				OTR->ASD8s[i].Layer,OTR->ASD8s[i].Sector,
				modid,mono,wire,OTR->ASD8s[i].ftb_ASD);
		OTR->Geo.CiArray[OTR->Geo.FirstCellInMonolayer[igede<<2|mono]+wire]=
		  (0x80000000|(k<<16)|i);
              }
	      continue;
            }



            if(OtrIMonoLayerNumberOf(igede)==4 && hwc[0]=='-'){
	      modid=OTR->ASD8s[i].Module;
              module=OtrDecodeAsd2Nede(1,mod_first_wire,
                                       k+1-asd_first_wire,0,nwi,&mono,&wire);
	      if(OTR->ASD8s[i].wire[k].ChID==0){ 
		if(wire<nwi[mono] && (module==0 || module==1)){
		  OTR->ASD8s[i].wire[k].NEDE=igede;
		  OTR->ASD8s[i].wire[k].ChID=
		    OtrIChannelId(OTR->ASD8s[i].SuperLayer,
				  OTR->ASD8s[i].Layer,OTR->ASD8s[i].Sector,
				  modid,mono,wire,OTR->ASD8s[i].ftb_ASD);
		  OTR->Geo.CiArray[OTR->Geo.FirstCellInMonolayer[igede<<2|mono]+wire]=
		    (0x80000000|(k<<16)|i);
		}
	      }
              if((mod_first_wire+k)==1){
                if(OtrDecodeAsd2Index(OTR->ASD8s[i].SuperLayer,
                                              OTR->ASD8s[i].Layer,
				              atoi(cab[i].previous1+6),
				              atoi(cab[i].previous1+9),
                                              &Line)){
		  modid=OTR->ASD8s[i].Module;
		  mono=0; wire=0; 
		  if(OTR->ASD8s[Line].wire[12].ChID==0){
		    OTR->ASD8s[Line].wire[12].NEDE=igede;
		    OTR->ASD8s[Line].wire[12].ChID=
		      OtrIChannelId(OTR->ASD8s[i].SuperLayer,
				    OTR->ASD8s[i].Layer,OTR->ASD8s[Line].Sector,
				    modid,mono,wire,OTR->ASD8s[Line].ftb_ASD);
		    OTR->Geo.CiArray[OTR->Geo.FirstCellInMonolayer[igede<<2|mono]+wire]=
		      (0x80000000|(12<<16)|Line);

		    mono=1; wire=0; 
		    OTR->ASD8s[Line].wire[14].NEDE=igede;
		    OTR->ASD8s[Line].wire[14].ChID=
		      OtrIChannelId(OTR->ASD8s[i].SuperLayer,
				    OTR->ASD8s[i].Layer,OTR->ASD8s[Line].Sector,
				    modid,mono,wire,OTR->ASD8s[Line].ftb_ASD);
		    OTR->Geo.CiArray[OTR->Geo.FirstCellInMonolayer[igede<<2|mono]+wire]=
		      (0x80000000|(14<<16)|Line);
		  }
                }
              }
	      continue;
            }
            if(OtrIMonoLayerNumberOf(igede)==4 && hwc[0]=='+'){
	      if((nwi[1]==nwi[2] && nwi[1]==(nwi[0]-1) && nwi[2]==(nwi[3]+1)) || 
		 (nwi[1]==nwi[2] && nwi[1]==(nwi[0]+1) && nwi[2]==(nwi[3]-1))) 
		spmod=1;
	      if(OTR->ASD8s[i].wire[k].ChID==0){ 
		modid=OTR->ASD8s[i].Module;
		if(!(spmod && OTR->ASD8s[i].Sector==4)){
		  module=OtrDecodeAsd2Nede(1,mod_first_wire,
				   k+1-asd_first_wire,1,nwi,&mono,&wire);
		}else{
		  if(mod_first_wire<16 && fw==0) fw=mod_first_wire-1;
		  module=OtrDecodeAsd2Nede(1,mod_first_wire-fw,
				   k+1-asd_first_wire+6,1,nwi,&mono,&wire);
		  if(mono==2)wire--;
		  if(mono==3)wire-=2;
		  if(mono==1)wire--;
		  module=0;
		}
		if(wire<nwi[mono] && module==0){
		  OTR->ASD8s[i].wire[k].NEDE=igede;
		  OTR->ASD8s[i].wire[k].ChID=
		    OtrIChannelId(OTR->ASD8s[i].SuperLayer,
				  OTR->ASD8s[i].Layer,OTR->ASD8s[i].Sector,
				  modid,mono,wire,OTR->ASD8s[i].ftb_ASD);
		  OTR->Geo.CiArray[OTR->Geo.FirstCellInMonolayer[igede<<2|mono]+wire]=
		    (0x80000000|(k<<16)|i);
		}
	      }
              if(((wire+1)==nwi[2] && mono==2)){
                if(OtrDecodeAsd2Index(OTR->ASD8s[i].SuperLayer,
                                              OTR->ASD8s[i].Layer,
				              atoi(cab[i].next1+6),
				              atoi(cab[i].next1+9),
                                              &Line)){
		  modid=OTR->ASD8s[i].Module;
		  if(OTR->ASD8s[Line].wire[1].ChID==0){
		    if(!(spmod && OTR->ASD8s[Line].Sector==1)){
		      mono=1; wire=nwi[1]-1;
		      OTR->ASD8s[Line].wire[1].NEDE=igede;
		      OTR->ASD8s[Line].wire[1].ChID=
			OtrIChannelId(OTR->ASD8s[i].SuperLayer,
				      OTR->ASD8s[i].Layer,OTR->ASD8s[Line].Sector,
				      modid,mono,wire,OTR->ASD8s[Line].ftb_ASD);
		      OTR->Geo.CiArray[OTR->Geo.FirstCellInMonolayer[igede<<2|mono]+wire]=
			(0x80000000|(1<<16)|Line);

		      OTR->ASD8s[Line].wire[3].NEDE=igede;
		      mono=0; wire=nwi[0]-1;
		      OTR->ASD8s[Line].wire[3].ChID=
			OtrIChannelId(OTR->ASD8s[i].SuperLayer,
				      OTR->ASD8s[i].Layer,OTR->ASD8s[Line].Sector,
				      modid,mono,wire,OTR->ASD8s[Line].ftb_ASD);
		      OTR->Geo.CiArray[OTR->Geo.FirstCellInMonolayer[igede<<2|mono]+wire]=
			(0x80000000|(3<<16)|Line);
		    }
		  }
		}
              }
	      if(spmod) spmod=0;
            }
          }
        }
      }
  }
  return 1;
}

/*===========================================================================*/
int OtrDecodeAsd2Index(unsigned char SL,unsigned char Layer,
		       unsigned char Sector,unsigned char ASD,
		       unsigned short *Line){
  if(Sector<1 || Sector>12 || SL<1 || SL>31 || Layer<1 
     || Layer>OTR->Slayers[SL].NLayer) return 0;
  if(ASD>(OTR->Slayers[SL].Nasd[Layer][Sector]+
          OTR->Slayers[SL].FirstASD[Layer][Sector]-1) ||
     ASD<OTR->Slayers[SL].FirstASD[Layer][Sector]) return 0; 
  *Line=OTR->Slayers[SL].Line[OTR->Slayers[SL].pos[Layer][Sector]-
                            OTR->Slayers[SL].FirstASD[Layer][Sector]+ASD];
  return 1;
}

/*===========================================================================*/
int OtrDecodeAsd2Nede(int dbl, int w1, int ch, int plus, int *wires, 
             int *mono, int *wire){
int w,total_wires; 
    w=w1+ch-1;
    total_wires=wires[0]+wires[1]+wires[2]+wires[3];
    if(!dbl){
      // ------ Single layer ------
      *mono=(plus ? w%2 : !(w%2));
      *wire =(w-1)/2;
      return 0;
    }
    // ------ Double layer ------
    if(plus){
      // ------ +X ------
      if(w==2 || w==4){
        // those ASD channels are connected to 
        // previous chamber
        return -1;
      }
      if(w%2==0){
        // mono#0 or mono#1
        *mono=((w%4==0) ? 0 : 1);
      }else{
        // mono#2 or mono#3
        *mono=((((w-1)%4==0 && w>3) || w==1) ? 3 : 2);
      }
      switch (*mono){
        case 0 : *wire=w/4-2; break;  
        case 1 : *wire=(w-2)/4-1; break;  
        case 2 : *wire=(w-3)/4; break;  
        case 3 : *wire=(w-1)/4; break;  
      }
      return 0;
    }
    // ------ -X ------
    if(w%2==0){
      // mono#0 or mono#1
      *mono=((w%4==0) ? 3 : 2);
    }else{
      // mono#0 or mono#1
      *mono =((((w-1)%4==0 && w>3) || w==1) ? 0 : 1);
    }
    switch (*mono){
      case 3 : *wire =w/4-1; break;  
      case 2 : *wire =(w-2)/4; break;  
      case 1 : *wire =(w-3)/4+1; break;  
      case 0 : *wire =(w-1)/4+1; break;  
    }
    if(w==total_wires-3){
      return 1;
    }
    if(w==total_wires-1){
      return 1;
    }
    return 0;
}


/*==========================================================================*/
typedef struct{
  unsigned int val,cmp,min,max;
  char *name;
}OtrDecodeTestId;

void OtrDecodeCheckId(OtrDecodeTestId* id) {
  if(id->val!=id->cmp){ 
    printf("%s decoding of ChannelId is wrong %d %d\n",id->name,id->val,id->cmp);
  }
  else {   
    if (id->val>id->max) id->max=id->val;
    if (id->val<id->min) id->min=id->val;
  }
}

int OtrDecodeTestMe(){
  int tdcid,tdcch,numchannels=0;
  int ret=OTRIInvalidInt;
  char W[3000][4][32];
  char C[3000][4][32];
  int i,j,k,nwi[4];
  char hwc[12];
  int chid,ci,c1;
  int gi,mono,wire;

  OtrDecodeTestId sl ={0,0,0xFFFFFFFF,0,"SuperLayer"};
  OtrDecodeTestId l  ={0,0,0xFFFFFFFF,0,"Layer"};
  OtrDecodeTestId sec={0,0,0xFFFFFFFF,0,"Sector"};
  OtrDecodeTestId mod={0,0,0xFFFFFFFF,0,"Module"};
  OtrDecodeTestId ml ={0,0,0xFFFFFFFF,0,"MonoLayer"};
  OtrDecodeTestId w  ={0,0,0xFFFFFFFF,0,"Wire"};
  OtrDecodeTestId asd={0,0,0xFFFFFFFF,0,"Asd"};
  
  if(OtrDecodeAlgorithm==OTRIInvalidInt) return ret;

  if(OtrGeoNumberOfVolumes()>3000){
    printf("OtrDecodeTestMe(): Number of NEDE is to big...\n");
    return ret;
  }
  for (i=0;i<OtrITdcNumberOf();i++) {
    for (j=0;j<OtrITdcNumberOfChannel();j++) {
      ci=OtrIFromTdcIndex(i,j);
      if(ci==OTRIInvalidInt){
	continue;
      }
      tdcid=OtrITdcId(ci);
      tdcch=OtrITdcChannel(ci);
      c1=OtrIFromTdc(tdcid,tdcch);
      if (c1!=ci){
        printf("Channel inconsistent i=%d j=%d  f1 %04X c0=%08X c1=%08X\n",
	       i,j,tdcid,ci,c1);
      }
      sl.val=OtrISuperLayer(ci);
      l.val=OtrILayer(ci);
      sec.val=OtrISector(ci);
      ml.val=OtrIMonoLayer(ci);
      w.val=OtrIWire(ci);
      mod.val=OtrIModule(ci);
      asd.val=OtrIAsd(ci);
      
      chid=OtrIdFromIndex(ci);
      sl.cmp=OtrISuperLayerFromId(chid);
      l.cmp=OtrILayerFromId(chid);
      sec.cmp=OtrISectorFromId(chid);
      ml.cmp=OtrIMonoLayerFromId(chid);
      w.cmp=OtrIWireFromId(chid);
      asd.cmp=OtrIAsdFromId(chid);
      mod.cmp=OtrIModuleFromId(chid);

      OtrDecodeCheckId(&sl);
      OtrDecodeCheckId(&l);
      OtrDecodeCheckId(&sec);
      OtrDecodeCheckId(&mod);
      OtrDecodeCheckId(&ml);
      OtrDecodeCheckId(&w);
      OtrDecodeCheckId(&asd);
      if(ci!=c1){
	printf("Error: sl %d %d l %d %d s %d %d m %d %d w %d %d a %d %d  mod:%i %i\n",
	       sl.val,sl.cmp,l.val,l.cmp,sec.val,sec.cmp,ml.val,ml.cmp,
	       w.val,w.cmp,asd.val,asd.cmp,mod.val,mod.cmp);


      }
      numchannels++;
    }
  }
  printf("range sl  %d %d\n",sl.min,sl.max);
  printf("range lay %d %d\n",l.min,l.max);
  printf("range sec %d %d\n",sec.min,sec.max);
  printf("range mod %d %d\n",mod.min,mod.max);
  printf("range asd %d %d\n",asd.min,asd.max);
  printf("range ml  %d %d\n",ml.min,ml.max);
  printf("range w   %d %d\n",w.min,w.max);
  printf("number of channels: %d\n",numchannels);

  printf("Checking correspondence (fed,channel)<->(gi,monolayer,wire)\n");
  printf("========================================================="
	 "=========================================\n");
  numchannels=0;
  for (i=0;i<OtrITdcNumberOf();i++){
    for (j=0;j<OtrITdcNumberOfChannel();j++){
      ci=OtrIFromTdcIndex(i,j);
      if(ci==OTRIInvalidInt){
	continue;
      }
      tdcid=OtrIFedIdFromIndex(i);
      tdcch=j;
      gi=OTR->ASD8s[ci&0xffff].wire[ci>>16].NEDE;
      mono=OtrIMonoLayer(ci);
      wire=OtrIWire(ci);
      ci=OTR->Geo.CiArray[OTR->Geo.FirstCellInMonolayer[(gi<<2)|mono]+wire];        
      if(((ci&0x80000000)!=0) && OTR->ASD8s[ci&0xffff].wire[(ci>>16)&0xf].ChID!=0){
	ci=ci&0xfffff;
	if(tdcid==OtrIFedId(ci) && tdcch==OtrIFedChannel(ci)){
	  numchannels++;
	}else{
	  OtrIGeoIdFromIndex(gi,hwc);
	  printf("Error for FED: %04X  Channel: %i\n",tdcid,tdcch);
	  printf("hwc:%s mono:%i wire:%i\n",hwc,mono,wire);
	}
      }
    }
  }
  printf("Total number of channels: %i\n\n\n",numchannels);
  
  memset(W,0,3000*4*32);
  memset(C,0xff,3000*4*32);
  for (i=0;i<OtrITdcNumberOf();i++) {
    for (j=0;j<OtrITdcNumberOfChannel();j++) {
      int ci=OtrIFromTdcIndex(i,j);
     if((ci&0xFFFF)!=0xFFFF){
       W[OtrIGeoIndex(ci)][OtrIMonoLayer(ci)][OtrIWire(ci)]++;
       C[OtrIGeoIndex(ci)][OtrIMonoLayer(ci)][OtrIWire(ci)]=ci>>16;
     }
    }
  }

  printf("Printing ASD channel -> (monolayer,cell) correspondence:\n");
  printf("========================================================="
	 "=========================================\n");
  ci=0;
  for(i=0;i<OtrGeoNumberOfVolumes();i++){
    OtrIGeoIdFromIndex(i,hwc);
    if(hwc[0]!='n'){
      for(k=0;k<4;k++)nwi[k]=OtrIMonoLayerNumberOfWires(i,k);
      
      if((nwi[1]==nwi[2] && nwi[1]==(nwi[0]-1) &&  nwi[2]==(nwi[3]+1)) || 
         (nwi[1]==nwi[2] && nwi[1]==(nwi[0]+1) &&  nwi[2]==(nwi[3]-1))) 
	ci=1;
      if(hwc[0]=='+' && !ci){
	printf("%s\n%02i ",hwc,nwi[3]);
	for(j=0;j<nwi[3];j++) printf("%02i ",C[i][3][j]);
	printf("\n%02i   ",nwi[2]);
	for(j=0;j<nwi[2];j++) printf("%02i ",C[i][2][j]);
	printf("\n%02i    ",nwi[1]);
	for(j=0;j<nwi[1];j++) printf("%02i ",C[i][1][j]);
	printf("\n%02i      ",nwi[0]);
	for(j=0;j<nwi[0];j++) printf("%02i ",C[i][0][j]);
	printf("\n\n");
      }
      if(hwc[0]=='+' && hwc[10]=='4' && ci){
	printf("------>Special module:\n");
	ci=0;
	printf("%s\n%02i  ",hwc,nwi[3]);
	for(j=0;j<nwi[3];j++) printf("%02i ",C[i][3][j]);
	printf("\n%02i ",nwi[2]);
	for(j=0;j<nwi[2];j++) printf("%02i ",C[i][2][j]);
	printf("\n%02i  ",nwi[1]);
	for(j=0;j<nwi[1];j++) printf("%02i ",C[i][1][j]);
	printf("\n%02i ",nwi[0]);
	for(j=0;j<nwi[0];j++) printf("%02i ",C[i][0][j]);
	printf("\n\n");
      }
      if(hwc[0]=='+' && hwc[10]=='1' && ci){
	printf("------>Special module:\n");
	ci=0;
	printf("%s\n%02i ",hwc,nwi[3]);
	for(j=0;j<nwi[3];j++) printf("%02i ",C[i][3][j]);
	printf("\n%02i   ",nwi[2]);
	for(j=0;j<nwi[2];j++) printf("%02i ",C[i][2][j]);
	printf("\n%02i    ",nwi[1]);
	for(j=0;j<nwi[1];j++) printf("%02i ",C[i][1][j]);
	printf("\n%02i      ",nwi[0]);
	for(j=0;j<nwi[0];j++) printf("%02i ",C[i][0][j]);
	printf("\n\n");
      }
      if(hwc[0]=='-'){
	printf("%s\n%02i      ",hwc,nwi[3]);
	for(j=0;j<nwi[3];j++) printf("%02i ",C[i][3][j]);
	printf("\n%02i    ",nwi[2]);
	for(j=0;j<nwi[2];j++) printf("%02i ",C[i][2][j]);
	printf("\n%02i   ",nwi[1]);
	for(j=0;j<nwi[1];j++) printf("%02i ",C[i][1][j]);
	printf("\n%02i ",nwi[0]);
	for(j=0;j<nwi[0];j++) printf("%02i ",C[i][0][j]);
	printf("\n\n");
      }
    }
  }
  printf("Printing now many TDC channels pointed to (monolayer,cell):\n");
  printf("========================================================="
	 "=========================================\n");
  for(i=0;i<OtrGeoNumberOfVolumes();i++){
    OtrIGeoIdFromIndex(i,hwc);
    if(hwc[0]!='n'){
      for(k=0;k<4;k++)nwi[k]=OtrIMonoLayerNumberOfWires(i,k);
      
      if((nwi[1]==nwi[2] && nwi[1]==(nwi[0]-1) &&  nwi[2]==(nwi[3]+1)) || 
         (nwi[1]==nwi[2] && nwi[1]==(nwi[0]+1) &&  nwi[2]==(nwi[3]-1))) 
	printf("------>Special module:\n");

      if(hwc[0]=='+'){
	printf("%s\n%02i ",hwc,nwi[3]);
	for(j=0;j<nwi[3];j++) printf("%02i ",W[i][3][j]);
	printf("\n%02i   ",nwi[2]);
	for(j=0;j<nwi[2];j++) printf("%02i ",W[i][2][j]);
	printf("\n%02i    ",nwi[1]);
	for(j=0;j<nwi[1];j++) printf("%02i ",W[i][1][j]);
	printf("\n%02i      ",nwi[0]);
	for(j=0;j<nwi[0];j++) printf("%02i ",W[i][0][j]);
	printf("\n\n");
      }
      if(hwc[0]=='-'){
	printf("%s\n%02i      ",hwc,nwi[3]);
	for(j=0;j<nwi[3];j++) printf("%02i ",W[i][3][j]);
	printf("\n%02i    ",nwi[2]);
	for(j=0;j<nwi[2];j++) printf("%02i ",W[i][2][j]);
	printf("\n%02i   ",nwi[1]);
	for(j=0;j<nwi[1];j++) printf("%02i ",W[i][1][j]);
	printf("\n%02i ",nwi[0]);
	for(j=0;j<nwi[0];j++) printf("%02i ",W[i][0][j]);
	printf("\n\n");
      }
    }
  }
  return ret;
}




