#ifndef __OtrDecode__H__
#define __OtrDecode__H__
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "OtrGeo.h"
#include "Otr/OtrDb.h"

#ifdef __cplusplus
extern "C" {
#endif

/*=== definition for Channel Index: =====
   bit  0-15:  asd index.
   bit 16-19:  channel in asd.
   bit 20-31:  reserved.
=======================================*/
#define CABTAB        0x0001 /* 0000000000000001 */
#define CABTABFILE    0x0003 /* 0000000000000011 */
#define SHARCTAB      0x0004 /* 0000000000000100 */
#define SHARCTABFILE  0x000C /* 0000000000001100 */
#define GEOTAB        0x0010 /* 0000000000010000 */

#define OTRMAXTDC 956
#define OTRMAXASD 7300
/*================ Structures of cabling tables: ===================*/
typedef struct{
  char           tdccon[11];
  unsigned short FedID;
  unsigned char  VME;
  unsigned short SharcBoardID;
  unsigned char  Port;    /* 1 ... 30 */
  unsigned char  SLB;
}OtrDecodeSharcStruct;


typedef struct{
  char gencha1[14];
  char gencha2[14];
  char modcha1[16];
  char modcha2[16]; 
  char modhv[14];  
  char ftbpos[13]; 
  char ftbasd[12]; 
  char ftbhv[12];  
  char ftbcab[11]; 
  char lvpos[10];  
  char lvcab[12];  
  char tdccab[12]; 
  char tdccon[11]; 
  char previous1[12];
  char previous2[12];
  char next1[12];   
  char next2[12];
  char status[4];        
  char hvgroup[13];  
  char hv1[14];    
  char hv2[14]; 
  char hvcab1[12];
  char hvcab2[12];   
} OtrDecodeCabStruct;


/*==================================================================*/

typedef struct{
  unsigned char  status;        /* Channel status                   */
  int            NEDE;          /* NEDE ID                          */
  int            ChID;          /* Channel ID                       */
  int            ChIndex;       /* Channel Index                    */
}OtrWireStruct;
  
typedef struct{
  unsigned char   SuperLayer;
  unsigned char   Layer;
  unsigned char   Sector;
  unsigned char   Module;
  unsigned char   ftb_ASD;
  unsigned char   TDC_Connector;
  unsigned short  TDC_FED;
  char            hwc[12];
  char            lvpos[10];
  char            HVcab1[12];
  char            HVcab2[12];
  char            HVgroup1[14];
  char            HVgroup2[14];
  OtrWireStruct   wire[16];
} OtrASD8Struct;

typedef struct{
  unsigned short ASD[8];        /* ASD8 index                       */
  unsigned char  VME_num;
  unsigned char  Sharc_Board;
  unsigned char  Sharc_Connector;
  unsigned short FED;
  unsigned char  crate;
  unsigned char  slot;
  unsigned char  status;
  unsigned char  SLB;
}OtrTDCStruct;
  
typedef struct{
  unsigned char  NLayer;
  unsigned char  Nasd[7][13];
  unsigned char  FirstASD[7][13];
  unsigned short pos[7][13];        /*pointer to 1-st ASD in sector  */
  unsigned short *Line;             /*buffer with pointers to ASD    */
}OtrSuperLayerStruct;

typedef struct{
  int NumOfVol;
  int *FirstCellInMonolayer;
  int *CiArray;
}OtrGeoMapStruct;

typedef struct{
  int NASDs;
  int NTDCs;
  int SLs[32];
  int Algorithm;
  char SLname[32][4];
  OtrASD8Struct       *ASD8s;             /*ASD8 information                */
  OtrTDCStruct        *TDCs;              /*TDC information                 */ 
  unsigned short      *FEDs;              /*Pointer FED->TDCs               */
  OtrSuperLayerStruct *Slayers;           /*Pointer SL,Lay,Sect,ASD->FED,ch */
                                          /*1-14 '-',17-30 '+'side          */ 
  OtrGeoMapStruct      Geo;     
}OtrDecodeStruct;

/*=========================== User functions ===============================*/
void  OtrDecodeSetAlgorithm(int i);
inline void OtrDecodeSwitch(int i) { OtrDecodeSetAlgorithm(i); }
void  OtrDecodeDefine();
int   OtrDecodeInitialize();
int   OtrDecodeReinitialize();
int   OtrDecodeUpdate();
  
int   OtrDecodeWriteToDb();
int   OtrDecodeClose();
int   OtrDecodeTestMe();
const OtrDecodeStruct* OtrDecodeGet();

/*========================== Internal functions ============================*/ 
OtrDecodeCabStruct *OtrDecodeReadCab(int *n);
OtrDecodeSharcStruct *OtrDecodeReadSharc(int *n);
int OtrDecodeFillAsds(OtrDecodeCabStruct *data);
int OtrDecodeFillTdcs(OtrDecodeCabStruct *cab,OtrDecodeSharcStruct *sharc);
int OtrDecodeFillWire(OtrDecodeCabStruct *cab);
int OtrDecodeAsd2Nede(int dbl,int w1,int ch,int plus,int *wires,int *mono,
		      int *wire);
int OtrDecodeAsd2Index(unsigned char SL,unsigned char Layer,
		       unsigned char Sector,unsigned char ASD,
		       unsigned short *Line);

#ifdef __cplusplus
};
#endif

#endif





