#include "Otr/OtrI.h"
#include "Otr/OtrDb.h"
#include "Otr/OtrDq.h"
#include "Otr/OtrDqAlgo.h"


/* Fortran Interface */
#ifdef __cplusplus
extern "C" {
#endif
  void  OtrDqSwitch_(int* i)        { OtrDqSwitch(*i); }
  int   OtrDqInitialize_()          { return OtrDqInitialize(); }
  int   OtrDqReinitialize_()        { return OtrDqUpdate(); }
  int   OtrDqWriteToDb_()           { return OtrDqWriteToDb(); }
  int   OtrDqClose_()               { return OtrDqClose(); }
  int   OtrDqScan_()                { return OtrDqScan(0); }
  int   OtrDqPrint_()               { return OtrDqPrint(0); }
  int   OtrDqTestMe_()              { return OtrDqTestMe(); }
#ifdef __cplusplus
  }
#endif


/* ------------------------------------------------------------------- */

typedef struct{
  int chid;
  int status;
}OtrDqDbaseStruct;

OtrDbField OtrDqDbaseFormat[]={
  {"int","ChannelId",OtrDbOffset(OtrDqDbaseStruct,chid)},
  {"int","Status",OtrDbOffset(OtrDqDbaseStruct,status)},
};

OtrDbTable OtrDqDbaseTableDef={"OTR_Dq", 1,"Status Table",
			       sizeof(OtrDqDbaseStruct),OtrDqDbaseFormat,
			       sizeof(OtrDqDbaseFormat)/sizeof(OtrDbField)};

/* ------------------------------------------------------------------- */

void OtrDqDefine() {
  OtrDbDefine("/OTR_STATUS",OtrDqDbaseTableDef.name,OtrDqUpdate);
  }

/* ------------------------------------------------------------------- */

void OtrDqUserDefine(int major,int minor,int match,int version) {
  OtrDbUserDefine("/OTR_STATUS",OtrDqDbaseTableDef.name,OtrDqUpdate,
     major,minor,match,version);
  }

/* ------------------------------------------------------------------- */

static int OtrDqAlgorithm = 0;

static unsigned short  *OtrDqStat=0;
static OtrDqAlgoStruct *OtrDqData=0;
static int numasd=0;
static int numcha=0;

#define ID(ci) (((ci&0xffff)<<4) | (ci>>16))

int OtrDqEvents = 0;


/* ------------------------------------------------------------------- */

int OtrDqAlgorithmGet() { return OtrDqAlgorithm; }

const OtrDqAlgoStruct *OtrDqDataGet(int ci) { return &OtrDqData[ID(ci)]; }

unsigned short OtrDqStatus(int ci) { 
  return OtrDqStat[ID(ci)]; 
}
  
void OtrDqStatusSet(int ci,unsigned short s) { 
  OtrDqStat[ID(ci)]=s; 
}


/* -------------  Explaination of the status word:
   convention: flag=OTRDQGood/Bad/Unknown/Dead (0/1/2/3)
   --> Bit i    = 0  /   1  /  0  / 1
   --> Bit i+4  = 1  /   1  /  0  / 0                
   ---------------------------------------------------------------*/     

int OtrDqGood(int ci) {
  int algoOr  = (OtrDqAlgorithm&0xff0000)>>16;
  int algoAnd = (OtrDqAlgorithm&0xff000000)>>24;
  int NotalgoAnd = (~algoAnd)&0xff;
   
/* if one of the algorithms is "bad" -> channel is bad */
  if(algoOr) return  (OtrDqStat[ID(ci)]&algoOr) ? 0:1;

/* all selected algorithms say "good" -> channel is good */
  else if (algoAnd) {
    return (((OtrDqStat[ID(ci)]&0xf) | NotalgoAnd) == NotalgoAnd) ? 1:0;
  }

  return 0;
}


/* OccupBit = 1 (bad) OccupAlgoBit = 0 (unknown) -> dead */
int OtrDqDead(int ci) {
  return ((OtrDqStat[ID(ci)]&OTRDQOccup)
         &&!(OtrDqStat[ID(ci)]&(OTRDQOccup<<4))) ? 1 : 0;
}


/* ------------------------------------------------------------------- */
/** 

  void OtrDqSwitch(int i)

  Tasks:
    - switches between different algorithems of the package:
      + to evaluate the channel flag using the 16-bit word    
      + to enable different algorithems for the data evaluation
 
                                                                        */  
/* ---------------------------------------------------------------------*/

void  OtrDqSwitch(int i) {
  OtrDqAlgorithm=i;
  }


/*--------------------------------------------------------------------*/

void OtrDqAsciiDef(int flag, char *name){
  char s[1024];
  if (flag==OTRDQAsciiRead||flag==OTRDQDbaseRead) {
    sprintf(s,"OtrDqInput=%s",name);
    putenv(s);
  }
  if (flag==OTRDQAsciiWrite || flag==OTRDQDbaseUpdate) {
    sprintf(s,"OtrDqOutput=%s",name);
    putenv(s);
  }
  if (flag==OTRDQAsciiWrite) OtrDqAlgorithm|=OTRDQAsciiWrite;
  if (flag==OTRDQDbaseUpdate) {
    OtrDqAlgorithm &= ~OTRDQAsciiWrite;
  }
}

 
/** -------------------------------------------------------------------- */
/**
  Tasks: 
   - debug handling 
   */  

static int OtrDqDebugLevel=OTRDQNodebug;
 
void OtrDqSwitchDebug(int i) {
  OtrDqDebugLevel=i;
  }

int OtrDqDebug(int i) {
  return (i<=OtrDqDebugLevel) ? 1:0;
  }



/** --------------------------------------------------------------------

   OtrDqInitialize()

  Tasks:
    - Memory allocation 
    - prepare for reinitialization
    - after this call reinitalization
    
  All function of the package (beside OtrDqSetAlgorithm) have to check, 
  if the initialization was done properly.
  -------------------------------------------------------------------- */
  
int   OtrDqInitialize() {

  if (OtrDqAlgorithm==0) {
    printf("OtrDqInitialize> Nothing selected->No Init done\n");
    return OtrIInvalid();
  }

  return OtrDqUpdate();
}


/*----------------------------------------------------------------*/
/** 

   OtrDqUpdate()

   Tasks:
       
   Attention !!
 
    - update the channel status word; in general the update  
      should happen with information form the database. If the 
      request comes while calculating the channel status 
      this request forces to overwrite the database value
      whenever values from the calculation are available.
    
      If the package OtDq has not seen any events yet the 
      update always happens from the database (to allow dbase reading 
      at the beginning).
      */
/*----------------------------------------------------------------- */
  
int OtrDqUpdate() {
  int version;
  
  if(OtrDqDebug(OTRDQDebug)) printf("OtrDqUpdate> Updating Status words\n");
/*
  if (OtrDqAlgorithm==OTRIInvalid) {
    printf("OtrDqUpdate> ERROR: OtrDq not properly initialized !\n");
    return OtrIInvalid();
  }
  */
  /* -- check first if OtrDecode was initialized -- */
  if (!OtrDecodeGet()) {
    printf("OtrDqInitilization> no decoding pointer defined\n");
    OtrDqAlgorithm=0;
    return OtrIInvalid();
  }



  if (numasd!=OtrIAsdNumberOf()||numcha!=OtrIAsdNumberOfChannel()) {
    numasd=OtrIAsdNumberOf();
    numcha=OtrIAsdNumberOfChannel();
    
    if (numasd>10000&&numcha>16) {
      printf("OtrDqUpdate: Problems in OtrDecode with "
             "numasd=%d or numasdcha=%d\n",numasd,numcha);
      }

    OtrDqStat=(unsigned short*)
      realloc(OtrDqStat,sizeof(*OtrDqStat)*numasd*numcha);

    memset(OtrDqStat,0,sizeof(*OtrDqStat)*numasd*numcha);

    if (OtrDqAlgorithm&OTRDQCalculate) {
      int i,n;

      OtrDqData=(OtrDqAlgoStruct*)
        realloc(OtrDqData,sizeof(OtrDqAlgoStruct)*numasd*numcha);
      memset(OtrDqData,0,sizeof(OtrDqAlgoStruct)*numasd*numcha);

      for(n=0; n<numasd; n++) {
        for(i=0; i<numcha; i++) {
          int ci=OtrIFromAsd(n,i);
	  if (!OtrIError()) OtrDqData[ID(ci)].Chid=OtrIdFromIndex(ci);	  
	}
      }
    }
  }

  if (!(OtrDqAlgorithm&OTRDQCalculate)||OtrDqEvents==0) {
    char *fname=getenv("OtrDqInput");
    if (fname&&!fname[0]=='\0') {
      FILE *fp=fopen(fname,"r");
      if (!fp) {
        printf("OtrDqUpdate: Cannot open input file: %s\n",fname);
        return OtrIInvalid();
        }
      printf("OtrDqUpdate: update from file: %s\n",fname);fflush(stdout);
      OtrDqScan(fp);    
      }
    else if ((version=OtrDbGetTableVersion(OtrDqDbaseTableDef.name))!=-1) {
      OtrDqDbaseStruct *OtrDqDbaseTable=0;
      int nitems,i;
    
      OtrDqDbaseTable=(OtrDqDbaseStruct*)
         OtrDbGet(&OtrDqDbaseTableDef, OtrDqDbaseTable, &nitems);
      printf("OtrDqUpdate[debug]: got table %d\n",nitems);
      for (i=0;i<OtrIAsdNumberOf()*OtrIAsdNumberOfChannel();i++){
//        OtrDqStat[i]= 0xfff0;
        OtrDqStat[i]= 0x0000;
        }
      for (i=0; i<nitems; i++) {
        int ci=OtrIFromId(OtrDqDbaseTable[i].chid);
	if (ci==-1||OtrIError()) {
	  printf("OtrDqUpdate[warning]: Decoding problem for channel %d."
	    "Was valid, is invalid in this run\n",OtrDqDbaseTable[i].chid);
	  continue;
	  }
        OtrDqStat[ID(ci)]= OtrDqDbaseTable[i].status;
      }
      free(OtrDqDbaseTable);
      printf("OtrDqUpdate[debug]: done\n");
    }
    else return OtrIInvalid();
  } 
  else {  
    printf("OtrDqUpdate: update of status flags using results of calculations\n");
    fflush(stdout);
    for (int i=0;i<numasd*numcha;i++) {
      OtrDqStat[i]=OtrDqData[i].Status;
    }      
  }
  
  return OtrIValid();
}

/** ------------------------------------------------------------------
  Tasks:
    - scan ascii files into memory structures.
    - if fp==0 assume (Fortran case), try to take filename from environment
 
  --------------------------------------------------------------------*/

int   OtrDqScan(FILE *fp) { 
  char line[512],tag[512];
  char *s;
  int fmt=0;
  int chid;
  int ci, i,n=0;
  int err=0;
 
  while (fgets(line,512,fp)) {
    s=line;
    if (!fmt) {
      sscanf(s,"%s",tag);
      if (!strcmp(tag,"Index")) {
        fmt=1;
	OtrDqData=(OtrDqAlgoStruct*)
           realloc(OtrDqData,sizeof(OtrDqAlgoStruct)*numasd*numcha);
        memset(OtrDqData,0,sizeof(OtrDqAlgoStruct)*numasd*numcha);
      }
      else fmt=2;
    }
    if (fmt==1) sscanf(s,"%*s %d %*s %d %n",&i,&chid,&n);
    else if (fmt==2) sscanf(s,"%d %d %n",&i,&chid,&n);
    else continue;
    
    s+=n;
    if (chid==0) continue;
    ci=OtrIFromId(chid);
    if (OtrIError()) {
      printf("OtrDqScan: Channel not existing: id=%d\n",chid);
      continue;
    }
    ci=ID(ci);
    if (i!=ci) {
      int iasd=ci>>4; 
      if (err++<10) printf("OtrDqScan: AsdIndex changed in OtrDecode: %d %x %x\n",i,i,ci);
      if (iasd<0||iasd>OtrIAsdNumberOf()) {
        printf("OtrDqScan: AsdIndex=%d out of range for channel-id %d\n",iasd,chid);
	continue;
        }
    }

    if (fmt==1) {
      OtrDqData[ci].Chid=chid;
      sscanf(s,"%*s %hu %d %d %d %f %hd %hd %hd",
	    &OtrDqData[ci].Status,  
	    &OtrDqData[ci].Nevents, &OtrDqData[ci].Rms, 
	    &OtrDqData[ci].Tmean, &OtrDqData[ci].Lhood, 
	    &OtrDqData[ci].Nboxa, &OtrDqData[ci].Nboxb, 
	    &OtrDqData[ci].Nboxc );
      OtrDqStat[ci]=OtrDqData[ci].Status;
    } 
    else {
      sscanf(s,"%hd",&OtrDqStat[ci]);
    } 
  }
  fclose(fp);
  return OtrIValid(); 
}

  
/** ----------------------------------------------------------------------
  Tasks:
    - provide output to db or ascii files
    
  -----------------------------------------------------------------------*/

int   OtrDqWriteToDb() { 
  char *fname=getenv("OtrDqOutput");
  if (fname&&!fname[0]=='\0') {
    FILE *fp=fopen(fname,"w");
    if (!fp) {
      printf("OtrDqWriteToDb: Cannout open output file: %s\n",fname);
      return OtrIInvalid();
    }
    OtrDqPrint(fp);
    fclose(fp);
  }
  else {
    OtrDqDbaseStruct *OtrDqDbaseTable;
    int asd, nch;
    int nitems=0;
    for (nch=0; nch<numasd*numcha; nch++) {
      if (OtrDqStat[nch]!=0) nitems++;
    }
    if (nitems>0) {
      OtrDqDbaseTable=(OtrDqDbaseStruct*) 
                           malloc(nitems*sizeof(OtrDqDbaseStruct)); 
      nitems=0;
      for (asd=0; asd<OtrIAsdNumberOf(); asd++) {
        for (nch=0; nch<OtrIAsdNumberOfChannel(); nch++) {
          int ci=OtrIFromAsd(asd,nch);
	  int chid=OtrIdFromIndex(ci);
  	  int k=ID(ci);
	  if (OtrDqStat[k]!=0x00f0&&!OtrIError()){
	    OtrDqDbaseTable[nitems].chid=chid;
	    OtrDqDbaseTable[nitems].status=OtrDqStat[k];
	    nitems++;
	  }
	}
      }
      OtrDbPut(&OtrDqDbaseTableDef,OtrDqDbaseTable,nitems);
      free(OtrDqDbaseTable);
    } 
    else {
      printf("Status is good for all channels. No database update\n");
    }
  }
  return OTRIValid;
  }


/** ---------------------------------------------------------------------
  Tasks:
    - print memory structure to ascii files
    - if fp==0 assume (Fortran case), try to take filename from environment 
  */
  
int   OtrDqPrint(FILE *fp)     { 
  int i;
  int fmt;
  
  if (OtrDqAlgorithm&OTRDQAsciiWrite) {
    fmt=1;
    if (!OtrDqData) {
      printf("OtrDqPrint: OtrDqData is not allocated. Full output not possible\n");
      return OtrIInvalid();
    }  
  }
  else fmt=2;

  if (!fp) {
    printf("OtrDqScan> no open file to write to\n");    
    return OtrIInvalid();
    }
 
  for(i=0; i<OtrIAsdNumberOf();i++) {
    int j;
    for (j=0;j<OtrIAsdNumberOfChannel(); j++) {
      int ci =OtrIFromAsd(i,j);
      int k=ID(ci);
      int chid=OtrIdFromIndex(ci);
      if (OtrIError()) chid=0;
      if (k!=i*16+j) printf("ciii %d %d %d\n",ci,k,i*16+j);
      if (fmt==1) {   
        fprintf(fp,"Index %5d ChID %d Data %hu %hd %d %d %f %hd %hd %hd\n",
	    k, chid,(chid!=0) ? (OtrDqData+k)->Status : 1,  
	    (OtrDqData+k)->Nevents, (OtrDqData+k)->Rms, 
	    (OtrDqData+k)->Tmean, (OtrDqData+k)->Lhood, 
	    (OtrDqData+k)->Nboxa, (OtrDqData+k)->Nboxb, 
	    (OtrDqData+k)->Nboxc ); 
      }
      else {
        fprintf(fp,"%5d %d %hu\n",k, chid,(chid!=0)?*(OtrDqStat+k):1);
      }
    }
  }
  return OtrIValid();  
}


/** ---------------------------------------------------------------------
  Tasks:
    - free allocate memory 
    - clean up to be in a state that initialization is necessary 
  */
  
int   OtrDqClose()     { 
  if (OtrDqAlgorithm==OTRIInvalid) return OtrIInvalid();
   
  if (OtrDqData) free(OtrDqData);
  if (OtrDqStat) free(OtrDqStat);
  numasd=0;
  numcha=0;
  OtrDqData=0;
  OtrDqStat=0;
  OtrDqEvents=0;
  return OtrIValid();
}


/*--------------------------------------------------------------------*/
/**
  Tasks:
    - test bench to be sucessfully executed before any release of a new version
    - it is requested, that the author carefully thinks about testing
    - in addition it should be a collection how to test bugs, which happened
      at least once 
  */

int   OtrDqTestMe()    {
  if (OtrDqAlgorithm==OTRIInvalid) return OTRIInvalid;

/*-----Set enviroment to open DB files --------------------*/
//  putenv("OtrDqCfgFile=/afs/desy.de/group/hera-b/OTR/dev/Tables/OtrDq/OtrDq.cfg");

  putenv("OtrDqInput=/hb/otr2/run2000/DQ/OtrDq14154.stat");
  OtrDbUserDefine("/OTR_TEST","OTR_Status",OtrDqUpdate,0x7FFFFFFF,0x7FFFFFFF,2,1);

  OtrDqSwitch(OTRDQSelectOccupOr);
  
  OtrDqInitialize();
  OtrDqAsciiDef(OTRDQAsciiWrite,"dq.full");
  OtrDqWriteToDb();
  OtrDqAsciiDef(OTRDQDbaseUpdate,"dq.in");
  OtrDqWriteToDb();
  
  OtrDqClose();
  putenv("OtrDqInput=dq.in");
  putenv("OtrDqOutput=");
  OtrDqInitialize();
  OtrDqWriteToDb();
  putenv("OtrDqInput=");
  OtrDqUpdate();
  putenv("OtrDqOutput=dq.out");
  OtrDqWriteToDb();
  return OTRIInvalid;
  }
 
 

