#ifndef __OtrDq__H__
#define __OtrDq__H__

/* return values  */ 
#define OTRDQValid 0
#define OTRDQInvalid -1

/* define debug level */
#define OTRDQNodebug 0
#define OTRDQNoinit  0
#define OTRDQDebug   1
#define OTRDQDdebug  2
#define OTRDQDddebug 3


/* to steer the algorithems and the  initilization */

/* switch: bit 0 -  7 database/calculation
               8 - 15 active DQ algo's
              16 - 23 status word eval: algo's in logic Or
              24 - 32 status word eval: algo's in logic And
	      */

#define OTRDQNoInit      0x0
#define OTRDQDbaseRead   0x1 
#define OTRDQDbaseUpdate 0x2 
#define OTRDQAsciiRead   0x4 
#define OTRDQAsciiWrite  0x8 
#define OTRDQCalculate   0x10


#define OTRDQDefault 0xff 
#define OTRDQOccup 0x1
#define OTRDQLhood 0x2
#define OTRDQRms   0x4
#define OTRDQTmean 0x8

#define OTRDQAlgoDefault (OTRDQDefault<<8) 
#define OTRDQAlgoOccup   (OTRDQOccup<<8)
#define OTRDQAlgoLhood   (OTRDQLhood<<8)
#define OTRDQAlgoRms     (OTRDQRms<<8)
#define OTRDQAlgoTmean   (OTRDQTmean<<8)

#define OTRDQSelectOccupOr (OTRDQOccup<<16)
#define OTRDQSelectLhoodOr (OTRDQLhood<<16)
#define OTRDQSelectRmsOr   (OTRDQRms<<16)
#define OTRDQSelectTmeanOr (OTRDQTmean<<16)

#define OTRDQSelectOccupAnd (OTRDQOccup<<24)
#define OTRDQSelectLhoodAnd (OTRDQLhood<<24)
#define OTRDQSelectRmsAnd   (OTRDQRms<<24)
#define OTRDQSelectTmeanAnd (OTRDQTmean<<24)



#ifdef __cplusplus
extern "C" {
#endif
void OtrDqSwitchDebug(int i) ;
void  OtrDqSwitch(int i);
void  OtrDqAsciiDef(int flag, char *name);
int   OtrDqAlgorithmGet();

void OtrDqDefine();
void OtrDqUserDefine(int major,int minor,int match,int version);

int   OtrDqInitialize();

int   OtrDqClose();

int   OtrDqUpdate();

int   OtrDqWriteToDb();

int   OtrDqScan(FILE *fp);

int   OtrDqPrint(FILE *fp);

int   OtrDqTestMe();

unsigned short OtrDqStatus(int ci);
int OtrDqGood(int ci);
int OtrDqDead(int ci);

#ifdef __cplusplus
}
#endif

inline int OtrDqBad(int ci) { return !OtrDqGood(ci); }

#define OTRDQGood    0
#define OTRDQBad     1
#define OTRDQUnknown 2
#define OTRDQDead    3

inline int OtrDqStatusOccupancy(int ci) {
  int stat=OtrDqStatus(ci);
  int bit0=(stat&OTRDQOccup);
  int bit1=((~stat)&(OTRDQOccup<<4))>>3;
  return bit1|bit0;
  }

inline int OtrDqStatusLikelyhood(int ci) {
  int stat=OtrDqStatus(ci);
  int bit0=(stat&OTRDQLhood);
  int bit1=((~stat)&(OTRDQLhood<<4))>>3;
  return (bit1|bit0)>>(OTRDQLhood-1);
  }
  

inline int OtrDqStatusRms(int ci) {
  int stat=OtrDqStatus(ci);
  int bit0=(stat&OTRDQRms);
  int bit1=((~stat)&(OTRDQRms<<4))>>3;
  return (bit1|bit0)>>(OTRDQRms-1);
  }

inline int OtrDqStatusTmean(int ci) {
  int stat=OtrDqStatus(ci);
  int bit0=(stat&OTRDQTmean);
  int bit1=((~stat)&(OTRDQTmean<<4))>>3;
  return (bit1|bit0)>>(OTRDQTmean-1);
  }
  
#endif






