#include "Otr/OtrDqAlgo.h"
#include "Otr/OtrEvent.h"

static int OtrDqInitilization=0;
extern int OtrDqEvents;


static int OtrDqCycle=5000;


/* algorithms' parametres from configuration*/
static FILE *OtrDqCfg=0;
static float OtrDqLhoodLCut;
static float OtrDqLhoodUCut;
static float OtrDqOccupLCut;
static float OtrDqOccupUCut;
static short OtrDqRmsLCut[2];
static short OtrDqRmsUCut[2];



/* -------------  Explaination of the status word:
   convention: flag=OTRDQGood/OTRDQBad/OTRDQUnknown/Dead (0/1/2/3)
   --> Bit i    = 0  /   1  /  0  / 1
   --> Bit i+4  = 1  /   1  /  0  / 0                
   ---------------------------------------------------------------*/     


void OtrDqSetBit(int ci, int algo, int flag) {
  OtrDqAlgoStruct* d=(OtrDqAlgoStruct*)OtrDqDataGet(ci);
  if (d){
    d->Status |= ((flag%2)*algo) | (((((~(flag/2))%2)&0x1)*algo)<<4);
  }
}



/* --------------------------------------------------------------------------

               Likelyhood method 

     ----------------------------------------------------------------------*/

/* define the OtrDqLhood inner data structure */
typedef struct {
  char *type;
  float spectrum[64];
} OtrRefSpStruct;


OtrRefSpStruct *OtrRefSp;

int OtrDqLhoodInit(){
  char myname[32] = "OtrDqLhoodInit";
  char auxstring[16];
  short i, type;
  FILE *RefSp;

  /* -- read parameters for the algorithm -- */
  if(!getenv("OtrDqCfgFile")){
    printf("%s> Error: DQ Config file not determined\n", myname);
    return 0;
  }
  OtrDqCfg = fopen(getenv("OtrDqCfgFile"), "r");
  if (!OtrDqCfg) {
    printf("%s> Error: Unable to read DQ configuration\n", myname);
    return 0;
  }    
  while(fgets(auxstring, 32, OtrDqCfg) != NULL){
    if(!strncmp("OtrDqLhoodCut", auxstring, 12)){
      OtrDqLhoodLCut = (float) atoi(auxstring+13);
      OtrDqLhoodUCut = (float) atoi(auxstring+16);
    }
  }
  fclose(OtrDqCfg);
  OtrDqCfg = NULL;
  

  /*--- Put reference spectra in memory ---*/
  if(!getenv("OtrDqRefSpectra")){
    printf("%s> Error: Reference spectra file not determined\n", 
           myname);
    return 0;
  }

  RefSp = fopen(getenv("OtrDqRefSpectra"),"r");
  if(RefSp == NULL){
    printf("%s> Error: Unable to open file %s\n", myname,
           getenv("OtrDqRefSpectra"));
    return 0;
  }

#ifdef __cplusplus
  OtrRefSp = new OtrRefSpStruct[3];
#else
  OtrRefSp = malloc(sizeof(OtrRefSpStruct)*3);
#endif
  if (!OtrRefSp) {
    printf("%s> Error: Not enough memory to alloc RefSpectra\n", myname);
    return 0;
  }  
  memset(OtrRefSp, 0, sizeof(OtrRefSpStruct)*3);
    
  while(fgets(auxstring, 50, RefSp) != NULL){
    if(auxstring[0] == '#'){
      i = 0; type = atoi(auxstring+3)>>3;
      OtrRefSp[type].type = auxstring+1;
    }
    if(auxstring[0] == '0'){
      OtrRefSp[type].spectrum[i++] = atof(auxstring);
      if( OtrRefSp[type].spectrum[i-1] <= 0 &&
          OtrRefSp[type].spectrum[i-1] >= 1){
        printf("%s> Error: RefSpectrum DB file corrupted\n", myname);
        return 0;
      }
    }
  }
  fclose(RefSp);
  RefSp = NULL;
  
  return 1;
}

int  OtrDqLhoodEvent( int ci, unsigned char tdc){
  short sl   = OtrISuperLayer(ci);
  short type = int(OtrIWirePitch(ci));

  if((sl<9) || (sl>16 && sl<25)) type = 1+!type;

  float lhood = -log(OtrRefSp[type].spectrum[int(tdc/4)]);
  OtrDqAddLhood(ci, lhood);

  return 1;
}

int OtrDqLhoodUpdate( int ci){

  if (OtrDqGetEvents(ci) < 99){
    OtrDqSetBit(ci, OTRDQLhood, OTRDQUnknown);
    return 1;
  }

  if (OtrDqGetLhood(ci) > OtrDqGetEvents(ci)*OtrDqLhoodLCut &&
      OtrDqGetLhood(ci) < OtrDqGetEvents(ci)*OtrDqLhoodUCut)
    OtrDqSetBit(ci, OTRDQLhood, OTRDQGood);
  else OtrDqSetBit(ci, OTRDQLhood, OTRDQBad);

  return 1;
}


/* --------------------------------------------------------------------------

               Rms method 

     ----------------------------------------------------------------------*/


int  OtrDqRmsInit() {
  char myname[32] = "OtrDqRmsInit";
  char auxstring[16];
  
  /* -- read parameters for the algorithm -- */
  if(!getenv("OtrDqCfgFile")){
    printf("%s> Error: DQ Config file not determined\n", myname);
    return 0;
  }
  OtrDqCfg = fopen(getenv("OtrDqCfgFile"), "r");
  if (!OtrDqCfg) {
    printf("%s> Error: Unable to read DQ configuration\n", myname);
    return 0;
  }    
  while(fgets(auxstring, 32, OtrDqCfg) != NULL){
    if(!strncmp("OtrDqRmsUCut", auxstring, 12)){
      OtrDqRmsUCut[0] = atoi(auxstring+12);
      OtrDqRmsUCut[1] = atoi(auxstring+15);
    }
    if(!strncmp("OtrDqRmsLCut", auxstring, 12)){
      OtrDqRmsLCut[0] = atoi(auxstring+12);
      OtrDqRmsLCut[1] = atoi(auxstring+15);
    }
  }
  fclose(OtrDqCfg);
  OtrDqCfg = NULL;

  return 1;
}

int  OtrDqRmsEvent(int ci, unsigned char tdc) {
  int flag = (OtrDqAlgorithmGet()&0xff00)>>8;
  
  if(!(flag&OTRDQTmean))
    OtrDqAddTmean(ci, tdc);
  OtrDqAddRms(ci, tdc);

  return 1;
}

int  OtrDqRmsUpdate(int ci) {
  short type = int(OtrIWirePitch(ci));
  
  if(OtrDqGetEvents(ci) < 20){
    OtrDqSetBit(ci, OTRDQRms, OTRDQUnknown);
    return 1;
  }
  float rms = sqrt(OtrDqGetRms(ci));
  if((rms > OtrDqRmsLCut[type]) && (rms < OtrDqRmsUCut[type]))
    OtrDqSetBit(ci, OTRDQRms, OTRDQGood);
  else
    OtrDqSetBit(ci, OTRDQRms, OTRDQBad);

  return 1;
}


/* --------------------------------------------------------------------------

               Box method / Chi2  

     ----------------------------------------------------------------------*/

/* Box method/Chi2  */
int  OtrDqBoxInit() {
  return 0;

}
int  OtrDqBoxEvent(int ci, unsigned char tdc) {
  ci=0;
  tdc=0;
  return 0;

}
int  OtrDqBoxUpdate(int ci) {
  ci=0;
  return 0;

} 


/* --------------------------------------------------------------------------

               Occupancy method 

     ----------------------------------------------------------------------*/



typedef struct {
  float sectavrg[7][13];
} OtrOccupStruct;


OtrOccupStruct *OtrOccupAvrg;

int  OtrDqOccupInit() {
  char myname[32] = "OtrDqOccupInit";
  char auxstring[16];
  short nSl = OtrISuperLayerNumberOf();

  /*--- Put reference spectra in memory ---*/
#ifdef __cplusplus
  OtrOccupAvrg = new OtrOccupStruct[nSl];
#else
  OtrOccupAvrg = malloc(sizeof(OtrOccupStruct)*nSl);
#endif
  if (!OtrOccupAvrg) {
    printf("%s> Error: Not enough memory to alloc OccupAvrg\n", myname);
    return 0;
  }
  memset(OtrOccupAvrg, 0, sizeof(OtrOccupStruct)*nSl);

  /* -- read parameters for the algorithm -- */
  if(!getenv("OtrDqCfgFile")){
    printf("%s> Error: DQ Config file not determined\n", myname);
    return 0;
  }
  OtrDqCfg = fopen(getenv("OtrDqCfgFile"), "r");
  if (!OtrDqCfg) {
    printf("%s> Error: Unable to read DQ configuration\n", myname);
    return 0;
  }    
  while(fgets(auxstring, 32, OtrDqCfg) != NULL){
    if(!strncmp("OtrDqOccupCut", auxstring, 13)){
      OtrDqOccupLCut = atof(auxstring+13);
      OtrDqOccupUCut = atof(auxstring+18);
    }
  }
  fclose(OtrDqCfg);
  OtrDqCfg = NULL;

  return 1;
}

int  OtrDqOccupEvent(int ci) {
  short sl = OtrISuperLayer(ci);
  short lay = OtrILayer(ci);
  short sec = OtrISector(ci);
  
  OtrOccupAvrg[sl].sectavrg[lay][sec]++;

  return 1;
}

int  OtrDqOccupUpdate(int ci) {
  short sl = OtrISuperLayer(ci);
  short lay = OtrILayer(ci);
  short sec = OtrISector(ci);
  float error, average;
  
  average = OtrOccupAvrg[sl].sectavrg[lay][sec] /
    (float) OtrIAsdNumberPerSector(sl,lay,sec) /
    (float) OtrIAsdNumberOfChannel();
  error = (float) OtrDqGetEvents(ci)/average - 1;
  
  if(error < -OtrDqOccupLCut){
    OtrDqSetBit(ci, OTRDQOccup, OTRDQDead);
    return 1;
  } else if(error > OtrDqOccupUCut){
    OtrDqSetBit(ci, OTRDQOccup, OTRDQBad);
  } else OtrDqSetBit(ci, OTRDQOccup, OTRDQGood);
  
  return 1;
}

/* Tmean method  */
int  OtrDqTmeanInit() {

  return 1;
}
int  OtrDqTmeanEvent(int ci, unsigned char tdc) {

  OtrDqAddTmean(ci, tdc);

  return 1;
}
int  OtrDqTmeanUpdate(int ci) {
  float mean;
  
  if(OtrDqGetEvents(ci) < 30){
    OtrDqSetBit(ci, OTRDQTmean, OTRDQUnknown);
    return 1;
  }

  mean = OtrDqGetTmean(ci)/OtrDqGetEvents(ci);
  if((mean > 230) && (mean < 50))
    OtrDqSetBit(ci, OTRDQTmean, OTRDQBad);
  else
    OtrDqSetBit(ci, OTRDQTmean, OTRDQGood);

  return 1;
}


/* --------------------------------------------------------------------------

               G e n e r a l   A l g o r i t h m i c    P a r t 

     ----------------------------------------------------------------------*/


int OtrDqAlgoInit() {
  int rc=0;
  int flag= (OtrDqAlgorithmGet()&0xff00)>>8;
  
  OtrIValid();
  if (flag&OTRDQOccup  && OtrDqOccupInit()) rc=rc|OTRDQAlgoOccup;
  if (flag&OTRDQLhood  && OtrDqLhoodInit()) rc=rc|OTRDQAlgoLhood;
  if (flag&OTRDQRms   && OtrDqRmsInit())    rc=rc|OTRDQAlgoRms;
  if (flag&OTRDQTmean && OtrDqTmeanInit())  rc=rc|OTRDQAlgoTmean;

  OtrDqInitilization=rc;

  if(OtrDqDebug(OTRDQDebug))
    printf("OtrDqAlgoInit> RetC of Occup|Lhood|Rms|Tmean: %1d%1d%1d%1d\n",
	   (rc&OTRDQAlgoOccup)?1:0, (rc&OTRDQAlgoLhood)?1:0, 
	   (rc&OTRDQAlgoRms)?1:0, (rc&OTRDQAlgoTmean)?1:0);
	
  if ((rc>>8)&flag!=flag) return OtrIInvalid();   
  return rc;
  }

void OtrDqAlgoSetCycle(int cycle) {
  OtrDqCycle=cycle;
}

int OtrDqAlgoUpdate() {
  int ci, n, i, rc=0;
  int flag= (OtrDqAlgorithmGet()&0xff00)>>8;
  int init= (OtrDqInitilization&0xff00)>>8;

  if (!OtrDqInitilization&OTRDQCalculate) {
    printf("OtrDqAlgoEvent> package not initilized. Return.\n");
    return OtrIInvalid();
  }
  
  for(n=0; n<OtrIAsdNumberOf(); n++) {
    for(i=0; i<OtrIAsdNumberOfChannel(); i++) {
      ci = (n&0xffff)+(i<<16);
      if (flag&OTRDQOccup && init&OTRDQOccup) rc=rc|OtrDqOccupUpdate(ci)<<0;
      if (flag&OTRDQLhood && init&OTRDQLhood) rc=rc|OtrDqLhoodUpdate(ci)<<1;
      if (flag&OTRDQRms   && init&OTRDQRms)   rc=rc|OtrDqRmsUpdate(ci)<<2;
      if (flag&OTRDQTmean && init&OTRDQTmean) rc=rc|OtrDqTmeanUpdate(ci)<<3;
    }
  }
  if (rc&flag==flag){
    OtrIValid();
    OtrDqUpdate();
    return OTRDQAlgoUpdate;
    }
  return OtrIInvalid();
}


int OtrDqAlgoEvent() {
  int i; 
  int rc=0;
  int flag= (OtrDqAlgorithmGet()&0xff00)>>8;
  int init= (OtrDqInitilization&0xff00)>>8;

  if (!OtrDqInitilization&OTRDQCalculate) {
    printf("OtrDqAlgoEvent> package not initilized. Return.\n");
    return OtrIInvalid();
  }

  OtrDqEvents++;

  for (i=0; i<OtrEventHitNumberOf(); i++) {
    unsigned char tdc = OtrEventHitTdcValue(i);
    int ci = OtrEventHitChannelIndex(i);
    OtrDqAddEvent(ci);
    if (flag&OTRDQOccup && init&OTRDQOccup) rc=rc|OtrDqOccupEvent(ci)<<0;
    if (flag&OTRDQLhood && init&OTRDQLhood) rc=rc|OtrDqLhoodEvent(ci, tdc)<<1;
    if (flag&OTRDQRms   && init&OTRDQRms)   rc=rc|OtrDqRmsEvent(ci, tdc)<<2;
    if (flag&OTRDQTmean && init&OTRDQTmean) rc=rc|OtrDqTmeanEvent(ci, tdc)<<3;
    if ((flag&rc)!=flag) {
      printf("OtrDqAlgoEvents> Error Occup|Lhood|Rms|Tmean: %1d%1d%1d%1d\n",
	     (rc&OTRDQOccup)?1:0, (rc&OTRDQLhood)?1:0, 
	     (rc&OTRDQRms)?1:0, (rc&OTRDQTmean)?1:0);
      return OtrIInvalid();
    }
  }
  if(OtrDqDebug(OTRDQDebug))
    printf("OtrDqAlgoEvents> RetC of Occup|Lhood|Rms|Tmean: %1d%1d%1d%1d\n",
	   (rc&OTRDQOccup)?1:0, (rc&OTRDQLhood)?1:0, 
	   (rc&OTRDQRms)?1:0, (rc&OTRDQTmean)?1:0);

  if (rc&flag!=flag) return OtrIInvalid();
  rc = OtrIValid();
  if (OtrDqEvents%OtrDqCycle==0) rc = OtrDqAlgoUpdate(); 
  return rc; 
}








