#ifndef __OtrDqAlgo__H__
#define __OtrDqAlgo__H__

#include <stdio.h>
#include <Otr/OtrI.h>

#define OTRDQGood    0
#define OTRDQBad     1
#define OTRDQUnknown 2
#define OTRDQDead    3

#define OTRDQAlgoUpdate    0x01

/* ------------------------------------------------------------------- */

typedef struct {
  int Chid;
  unsigned short Status;
  unsigned int  Nevents;
  int Rms;
  int Tmean;
  float Lhood;
  short Nboxa;
  short Nboxb;
  short Nboxc;
  } OtrDqAlgoStruct;

/* ------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#endif

const OtrDqAlgoStruct *OtrDqDataGet(int ci);
int OtrDqAlgoInit();
int OtrDqAlgoEvent();
int OtrDqAlgoUpdate();
int OtrDqDebug(int level);

#ifdef __cplusplus
}
#endif

/* utility functions to set the DQ structure contents */

inline void OtrDqSetChid(int ci, int chid) {
   OtrDqAlgoStruct *d=(OtrDqAlgoStruct*)OtrDqDataGet(ci);
   if (d) d->Chid = chid;
   else printf("OtrDqSetChid> Error\n");
}

inline void OtrDqAddLhood(int ci, float lhood) {
   OtrDqAlgoStruct *d=(OtrDqAlgoStruct*)OtrDqDataGet(ci);
   if (d) d->Lhood += lhood;
}

inline void OtrDqAddEvent(int ci) {
   OtrDqAlgoStruct *d=(OtrDqAlgoStruct*)OtrDqDataGet(ci);
   if (d) d->Nevents++;
}

inline void OtrDqAddRms(int ci, unsigned char tdc) {
   OtrDqAlgoStruct *d=(OtrDqAlgoStruct*)OtrDqDataGet(ci);
   if (d) d->Rms += tdc*tdc;
}

inline void OtrDqAddTmean(int ci, unsigned char tdc) {
   OtrDqAlgoStruct *d=(OtrDqAlgoStruct*)OtrDqDataGet(ci);
   if (d) d->Tmean += tdc;
}

/* utility to access DQ structures */

inline int OtrDqGetChid(int ci) { 
  const OtrDqAlgoStruct *d=OtrDqDataGet(ci);
  return d->Chid;
}

inline float OtrDqGetLhood(int ci) {
  const OtrDqAlgoStruct *d=OtrDqDataGet(ci);
  return d->Lhood;
}

inline int OtrDqGetEvents(int ci) {
  const OtrDqAlgoStruct *d=OtrDqDataGet(ci);
  return d->Nevents;
}

inline float OtrDqGetTmean(int ci) {
  const OtrDqAlgoStruct *d=OtrDqDataGet(ci);
  if(d->Nevents>0 ) return ((float)d->Tmean)/((float)d->Nevents);
  else return -1;
}

inline float OtrDqGetRms(int ci) {
  const OtrDqAlgoStruct *d=OtrDqDataGet(ci);
  if(d->Nevents>0)
    return ((float)d->Rms)/((float)d->Nevents)
           - OtrDqGetTmean(ci)*OtrDqGetTmean(ci);
  else return -1;
}

inline int OtrDqGetTzero(int ci) {
  float tmean=OtrDqGetTmean(ci);
  int sec=OtrISector(ci);
  /* correction from Tmean to TZero (72/36) taken from R.Zimmermann */
  if (sec>0 && sec<7) {
    return (int) tmean+36;
  } else if (sec<=12) {
    return (int) tmean+72;
  }
  return 0;
}


#endif
