#include "Otr/OtrI.h"
#include "Otr/OtrEvent.h"

#ifdef _cplusplus
extern "C" {
#endif

int                      OtrEventFirstHit=0;
int                      OtrEventNumHit=0;
int                      maxhit=0;

#include "event.h"
#include "tgts_raw.h"

#ifdef USE_ARTE
#  include "farm/farm.hh"
#else
  static OtrEventHitStruct *hits=0;
#endif

typedef struct{
  char stat;
  char bx;
} fedstat;

static fedstat        *OtrEventFedStat=0;
static int            OtrEventFedNum=0;
static int            OtrEvent_Switch=3;

void OtrEventStatisticsClear();
int OtrEventStatisticsFill(int ci, unsigned char tdc);
static int            OtrEventSlayerHits[32];
static int            OtrEventSlayerThits[32];
static int            OtrEventSlayerTsum[32];


static TGTS_RAW OtrEventTargetInfo;
static EVT_DATA OtrEventEventInfo;

void OtrEventInfoTargetGet(EVENT_HDR *event);


/* -------------------------------------------------------------- */ 

int            OtrEventFedNumberOf() {
  return OtrEventFedNum;
  }

#ifndef USE_ARTE 
const OtrEventHitStruct* OtrEventHitGet() {
  return hits;
  }
#endif  
  
/* 
unsigned short OtrEventFedId   (int hitindex) {
  }
*/  
 
int            OtrEventFedStatus(int hitindex) {
  return (OtrEventFedStat) ? OtrEventFedStat[hitindex].stat : OtrIInvalid();
  }

int            OtrEventFedBx(int hitindex) {
  return (OtrEventFedStat) ? OtrEventFedStat[hitindex].bx : OtrIInvalid();
  }

/* -------------------------------------------------------------- */
void  OtrEventSwitch(int i) {
  OtrEvent_Switch=i;
  }
  

inline int   OtrEventChannelValid(int ci) {
  switch ((OtrEvent_Switch&0xff)) {
    case 0: return (ci!=-1) ? 1:0;
    case 1: return (ci!=-1) ? 1:0;
    case 2: return (ci!=-1) ? 1:0;
    case 3: return (ci!=-1&&!OtrIIsBad(ci)) ? 1:0;
    }
  return 0;
  }
 
inline int   OtrEventDataValid(unsigned char tdc) {
  switch ((OtrEvent_Switch&0xff)) {
    case 0: return (tdc!=0xff) ? 1 : 0;
    case 1: return (tdc<0xfe && tdc>0) ? 1:0;
    case 2: return (tdc<0xfe && tdc>16) ? 1:0;
    case 3: return (tdc<0xfe && tdc>16) ? 1:0;
    }
  return 0;
  }
 

/* -------------------------------------------------------------- */
/**
  Tasks:
    - free allocate memory
    - clean up to be in a state that initialization is necessary 
  */
  
void   OtrEventClose()     { 
  OtrEventFirstHit=0;
  OtrEventNumHit=0;
  OtrEventFedNum=0;
  maxhit=0;
  if (OtrEventFedStat) {
    free(OtrEventFedStat);
    OtrEventFedStat=0;
    }

#ifndef USE_ARTE 
  if (hits) {
    free(hits);
    hits=0;
  }
#endif   
  
}


/* ---------  Statistical Part of OtrEvent       ------------- */
void OtrEventStatisticsClear(){
  int i;
  for(i=0; i<32; i++) OtrEventSlayerThits[i]=0;
  for(i=0; i<32; i++) OtrEventSlayerTsum[i]=0;
  for(i=0; i<32; i++) OtrEventSlayerHits[i]=0;
}

int OtrEventStatisticsFill(int ci, unsigned char tdc) {
  if (ci<0) return OTRIInvalid; 
  int sl=OtrISuperLayer(ci);
  int sec=OtrISector(ci);
  if (sl<=0 || sl>=32) return OTRIInvalid;

  /* for timing use only the 5mm sectors */ 
  if (sec>0 && sec<7) {
    OtrEventSlayerThits[sl]++;
    OtrEventSlayerTsum[sl]+=tdc;
    /* in case of PC/TC chambers */
    if ((sl&0xf)>8) {
      OtrEventSlayerThits[0]++;
      OtrEventSlayerTsum[0]+=tdc;
    }
  }   
  /* hits per superlayer */
  OtrEventSlayerHits[sl]++;
  OtrEventSlayerHits[0]++;
  
  return OTRIValid;
}

float OtrEventStatisticsTmean(int sl) {
  if (OtrEventSlayerThits[sl]==0) return -1.;
  return (float) OtrEventSlayerTsum[sl]/ (float) OtrEventSlayerThits[sl]; 
}
int OtrEventStatisticsTsum(int sl) {
  return OtrEventSlayerTsum[sl];
}
int OtrEventStatisticsThits(int sl) {
  return OtrEventSlayerThits[sl];
}
int OtrEventStatisticsHits(int sl) {
  return OtrEventSlayerHits[sl];
}
int OtrEventStatisticsOffbx() {
  float tm = OtrEventStatisticsTmean(0);
  return (tm>150 && tm<190) ? 0:1; 
}


/* --------------------------------------------------- */


/* ..................................................................... */
#ifndef USE_ARTE

#define SetChannel(v)    h->ci=v
#define SetGede(v)       
#define SetTdcValue(v)   h->tdcvalue=v
#define SetR(v)          h->r=v
#define SetRErr(v)       h->rerror=v

#else

#define SetChannel(v)    h->chid=OtrIdFromIndex(ci);
#define SetGede(v)       h->gede=ArtePointer<GEDE>(OtrGeoGet(OtrIGeoIndex(ci)).index())
#define SetTdcValue(v)   h->y=v
#define SetR(v)          h->q=v
#define SetRErr(v)       h->r=v

#include <set>

void OtrEventUpdateWax();
void OtrEventSetGedePointers();


template<> inline bool less<HITB>::operator ()
        (const struct HITB & h1, const struct HITB & h2) const {
  if (h1.gede.index()<h2.gede.index()) return 1;
  if (h1.gede.index()>h2.gede.index()) return 0;
  if (h1.z<h2.z) return 1;
  if (h1.z>h2.z) return 0;  
  return (h1.x<h2.x);
  //  if (h1.x>h2.x) return 0;
  /*  
  if (h1.y<h2.y) return 1;
  if (h1.y>h2.y) return 0;  
  return  (h1.z<h2.z);
  */
  }

int WaxWasUsed() {
  for (ArteTable<HITB>::iterator hitb = ArteTable<HITB>::begin();
       hitb != ArteTable<HITB>::end(); ++hitb ) {
    ArtePointer<GEDE> gede = hitb->gede;
    
    if (gede->cmp == 2) {
      return 1;
      }
    }
  return 0;
  }

#endif
/* ..................................................................... */


/* =====================================================================
 *
 *              OtrEventUpdate Function
 *
 * =====================================================================*/

#ifdef USE_ARTE
void OtrEventUpdate(EVENT_HDR* event);

void OtrEventUpdate() {
  if (WaxWasUsed()) {
    OtrEventUpdateWax();
    return;
    }
  EVENT_HDR *event=ARTE_get_event();
  if(!event) {
    OtrIErrorMsg(1,"OtrEventUpdate","EVENT_HDR undefined","");
    return;
    }
  OtrEventUpdate(event);
  }
#endif
  
void OtrEventUpdate(EVENT_HDR* event) {
  static ETB_fed *fed=new ETB_fed[OtrITdcNumberOf()];
  const OtrDecode* decode=OtrDecodeGet();
  unsigned char* d;
  int i, k;

#ifdef USE_ARTE
  set< HITB >hitbmap;
#endif

  ETB_get_good_feds(event,COMP_ID_OTR, OtrITdcNumberOf(), fed, &OtrEventFedNum);
  OtrEventInfoTargetGet(event);
  OtrEventInfoEventGet(event);
            
  if (OtrEvent_Switch&OTREVENTDoStatistics) OtrEventStatisticsClear(); 
  if (!OtrEventFedStat) {
    OtrEventFedStat=(fedstat*) malloc(sizeof(fedstat)*OtrITdcNumberOf());
    }
  memset(OtrEventFedStat,0,sizeof(fedstat)*OtrITdcNumberOf());
  OtrEventNumHit=0;


  for (i=0;i<OtrEventFedNum;i++) {
    unsigned short fedid=fed[i].hdr.fedid;
    int fi=OtrIFedIndexFromId(fedid);
    if (OtrIError()) continue;
    OtrEventFedStat[fi].stat=1;
    OtrEventFedStat[fi].bx=fed[i].hdr.bxid;

    d= (unsigned char*)  fed[i].data-1;
    for (k=0;k<OtrITdcNumberOfChannel();k++) {
      d++;
      if (*d!=0xff&&OtrEventDataValid(*d)) {
        int asdindex=decode->TDCs[fi].ASD[k>>4];
        int asdchannel=k&0xf;
        if (asdindex!=(unsigned short)0xFFFF&&
	    decode->ASD8s[asdindex].wire[asdchannel].ChID) {
          int ci=OtrIFromAsd(asdindex,asdchannel);
	  if (!OtrIError()) {
            if (OtrEventChannelValid(ci)) {

// .....................................................................
#ifdef USE_ARTE
              HITB hh,*h=&hh;
#else
              OtrEventHitStruct* h;
	      if (OtrEventNumHit==maxhit){
	        maxhit+=1000;
	        hits=(OtrEventHitStruct*)realloc(hits,sizeof(OtrEventHitStruct)*maxhit);
	        }
              h=&hits[OtrEventNumHit];
#endif
// .....................................................................

              SetChannel(ci);	  
              if (OtrIError()) continue;
	      SetGede(ci);
              if (OtrIError())  continue;
              SetTdcValue(*d);
              OtrEventNumHit++;
	      if (!(OtrEvent_Switch&OTREVENTNoHitPreparation)) {
  	        int   ml=OtrIMonoLayer(ci);
	        float t=*d-OtrITdcT0(ci);
#ifdef USE_ARTE
                h->x      = h->gede->xfw[ml] + OtrIWire(ci)*OtrIWirePitch(ci); 
                h->z      = h->gede->zfw[ml];
#else
                int   gi=OtrIGeoIndex(ci);
                h->x      = OtrGeoGet(gi)->xfw[ml] + OtrIWire(ci)*OtrIWirePitch(ci); 
                h->z      = OtrGeoGet(gi)->zfw[ml];
#endif
                SetR(OtrIRtRadius(ci,t));
                SetRErr(OtrIRtRadiusSigma(ci,t));
	        }
	      if (OtrEvent_Switch&OTREVENTDoStatistics) 
	        OtrEventStatisticsFill(ci, *d);

#ifdef USE_ARTE
	      hitbmap.insert(hh) ;
#endif

	      } /* OtrEventChannelValid */
	    } /* !OtrIError */
	  } /* chid is ok */
        } /* data is valid */
      } /* fedchannel loop */
    } /* fed loop */

#ifdef USE_ARTE
  OtrEventFirstHit=ArtePointer<HITB>::nrow()+1; 
  if (OtrEventNumHit>0) {
    for(set< HITB >::const_iterator hitbidIt = hitbmap.begin(); 
        hitbidIt != hitbmap.end() ; hitbidIt++) {
      ArteTable<HITB>::push_back(*hitbidIt);
      }
    OtrEventSetGedePointers();
    }
#endif

  } 


/* ======================================================================*/

#ifdef USE_ARTE

void OtrEventSetGedePointers() {
  // first reset HITB pointers
  for(ArteTable<GEDE>::iterator gedeIt = ArteTable<GEDE>::begin() ;
      gedeIt != ArteTable<GEDE>::end(); gedeIt++) 
    if(gedeIt->cmp==2) {
      gedeIt->hit1 = ArtePointer<HITB>(0) ;
      gedeIt->hitl = ArtePointer<HITB>(-1) ;
    } 
     
  // now loop over all hitbs and set pointers 
  int first=1 ;
  ArtePointer<GEDE> currentGEDE ;
  ArtePointer<HITB> prevhitbIt ;
  for(ArteTable<HITB>::iterator hitbIt = ArteTable<HITB>::begin() ;
      hitbIt != ArteTable<HITB>::end() ; hitbIt++)
    if(hitbIt->gede->cmp==2) {
      if(first) {
        currentGEDE=hitbIt->gede ;
        currentGEDE->hit1 = hitbIt ;
        first=0 ;
      } else {
        if(hitbIt->gede != currentGEDE) {
          currentGEDE->hitl = prevhitbIt ;
          currentGEDE = hitbIt->gede ;
          currentGEDE->hit1 = hitbIt ;
        }
      }
      prevhitbIt = hitbIt ;
    }
  currentGEDE->hitl = prevhitbIt ;

 }


/* ARTE and usage of WAX - temporary solution: only redifinition of channel ID */
/* tdcvalues and many other things are not set !!! */

void OtrEventUpdateWax() {
  OtrEventNumHit=0;
  for (ArteTable<HITB>::iterator hitb = ArteTable<HITB>::begin();
       hitb != ArteTable<HITB>::end(); ++hitb ) {
    ArtePointer<GEDE> gede = hitb->gede;
    
    if (gede->cmp == 2) {
      char hwc[13];

      if (!OtrEventFirstHit) OtrEventFirstHit=hitb.index();
      OtrEventNumHit++;
 
      strncpy(hwc,gede->hwc,12);
      hwc[12] = '\0';
      //printf("hwc=\"%s\"\n",hwc);

      int mod = atoi(hwc+6);
      int sl = OtrISuperLayerFromName(hwc);

      // In Wax hitb->chid = wireId + 128*(chId + 16*asdId), where wireId = wire+32*mono
      if (hitb->chid>15000000 || hitb->chid<0) {
        printf("TransformHitWax2OtrI: hitb->chid = %x is illegal as Wax hit chid\n",
               hitb->chid);
      }
      int waxid=hitb->chid;
      int mono = (waxid&0x7f)>>5;
      int wire = (waxid&0x7f)&0x1f;
 //     int asdi = (waxid>>7)>>4;  do not use this one for OtrDecode
//      int asdc = (waxid>>7)&0xf;
      //      printf("sl=%d, mod=%d, mono=%d, wire=%d\n",sl,mod,mono,wire);
      int ci = OtrIFromModule(sl,gede->lay,mod,gede->sec,mono,wire);
      if (ci==OTRIInvalid) {
  /*      
        printf("TransformHitWax2OtrI: OtrIFromModule(sl=%2d, lay=%1d, mod=%2d, sec=%2d, mono=%1d, wire=%2d) returned error for hwc=\"%s\"\
n",
               sl, gede->lay, mod, gede->sec, mono, wire, hwc);
	       */
        hitb->chid =OTRIInvalid;
        continue;
        }
/*
      hitb->x    = hitb->gede->xfw[mono] + OtrIWire(ci)*OtrIWirePitch(ci); 
      hitb->z    = hitb->gede->zfw[mono];
      unsigned char d=(unsigned char)(int)hitb->y;
      float t=d-OtrITdcT0(ci);
      hitb->q    = OtrRtRadius(ci,t);
      hitb->r    = OtrRtRadiusSigma(ci,t);
*/
      hitb->chid = OtrIdFromIndex(ci);
      if (hitb->chid==0) {
        printf("Channel Id is zero for channel ci=%d\n",ci);
	}
    }
  }
}


#endif



/*--------------------------------------------------------*/
/*    Interface to the DAQ and Target information in      */
/*  event record.         from F.Sanchez                  */
/*--------------------------------------------------------*/

void OtrEventInfoTargetGet(EVENT_HDR *event) {
  int rc;
  TGTS_RAW targ_info;
  ETB_fed fed_t;
  int nTg=0;  
  rc = ETB_get_feds(event,COMP_ID_TARG,1,&fed_t,&nTg);
  ESP_tran_from_int( (int *) fed_t.data, (int *) &targ_info,
		     sizeof(TGTS_RAW)/sizeof(int), (ESP_machine) (event->format) ); 
  memcpy(&OtrEventTargetInfo, &targ_info, sizeof(TGTS_RAW));
}

float OtrEventInfoIaRate() {
  return (float) OtrEventTargetInfo.ia;
}

float OtrEventInfoBgRate() {
  return (float) OtrEventTargetInfo.ia; 
}

void OtrEventInfoEventGet(EVENT_HDR *event) {
 EVT_DATA_RECORD all_daqs;
 int nb_daqs;
 ETB_get_daqs(event, 1, &all_daqs, &nb_daqs);
 memcpy(&OtrEventEventInfo, &(all_daqs.data), sizeof(EVT_DATA));
}

int OtrEventInfoPhysBx() {
  return (int) OtrEventEventInfo.physBX;
}
int OtrEventInfoFltBx() {
  return (int) OtrEventEventInfo.FLT_BX;
}
int OtrEventInfoRunNumb() {
  return OtrEventEventInfo.run_number;
}
int OtrEventInfoTime() {
  return OtrEventEventInfo.event_time;
}
int OtrEventInfoDbKey() {
  return OtrEventEventInfo.dbKey;
}
int OtrEventInfoFltStatus() {
  return OtrEventEventInfo.flt_status;
}
int OtrEventInfoFltEvent() {
  return (int) OtrEventEventInfo.fltevt;
}
int OtrEventInfoSltStatus() {
  return OtrEventEventInfo.slt_status;
}
int OtrEventInfoFcsTcode() {
  return (int) OtrEventEventInfo.FCS_tcode;
}


#ifdef _cplusplus
  }
#endif





