#ifndef __OtrEvent__H__
#define __OtrEvent__H__

#include "Otr/OtrI.h"
#include "etb.h"

#ifdef __cplusplus
extern "C" { 
#endif


/*  Event Switch settings */
#define OTREVENTValidAll     0
#define OTREVENTValidTdc     1
#define OTREVENTValidNoshort 2
#define OTREVENTDoStatistics  (1<<8)
#define OTREVENTNoHitPreparation (1<<9)


extern int OtrEventNumHit;


#ifdef USE_ARTE


extern int OtrEventFirstHit;
void OtrEventUpdate();
void  OtrEventSwitch(int i);

inline int OtrEventHitNumberOf() {return OtrEventNumHit;}


inline int OtrEventHitChannelId(int hitindex) { 
  return ArtePointer<HITB>(OtrEventFirstHit+hitindex)->chid; 
  }

inline int OtrEventHitChannelIndex(int hitindex) { 
  int ci,chid=OtrEventHitChannelId(hitindex);
  if (chid==0||chid==OTRIInvalid) {
    OtrIInvalid();
    ci=OTRIInvalid;
    }
  else ci=OtrIFromId(chid); 
  return ci;
  }

inline float OtrEventHitX(int hitindex) { 
  return ArtePointer<HITB>(OtrEventFirstHit+hitindex)->x; 
  }

inline float OtrEventHitZ(int hitindex) { 
  return ArtePointer<HITB>(OtrEventFirstHit+hitindex)->z; 
  }

inline float OtrEventHitR(int hitindex) { 
  return ArtePointer<HITB>(OtrEventFirstHit+hitindex)->q; 
  }

inline float OtrEventHitRError(int hitindex) { 
  return ArtePointer<HITB>(OtrEventFirstHit+hitindex)->r; 
  }

inline unsigned char  OtrEventHitTdcValue(int hitindex) { 
  return (unsigned char)(int)(ArtePointer<HITB>(OtrEventFirstHit+hitindex)->y+0.5); 
  }
  
inline int OtrEventHitGeoIndex(int hitindex) { 
  int ci=OtrEventHitChannelIndex(hitindex);
  return OtrIGeoIndex(ci); 
  }

void  OtrEventClose();

#else


typedef struct {
  int ci;
  float x,z,r,rerror;
  unsigned char tdcvalue;
  } OtrEventHitStruct;
  
  
/** @name OtrEvent 

    The OtrEvent interface provide access to the hits in the current
    event in the OTR. Using the OtrEventUpdate(EVENT_HDR*) function 
    (resp. OtrEventUpdate() inside ARTE) all hits are searched and the
    hitpreparation is done. 
    */
/*@{*/

/** Steer what TDC-values are valid.

 @param i   i=0:  all with valid channel id, i=1: (tdc>0 && tdc<0xfe) 
            default is i=1.       
 */

void  OtrEventSwitch(int i);

/** Hit preparation for current event (in ARTE replaced by OtrEventUpdate()).

  @param event  pointer to EVENT_HDR* structure (no argument in case of ARTE)
 */
void                     OtrEventUpdate(EVENT_HDR* event);
const OtrEventHitStruct* OtrEventHitGet();

/** Free all allocated memory.       
 */

void  OtrEventClose();


/** Access number of hits in the current event.


 @return Number of hits in current event
 */

inline int OtrEventHitNumberOf() { return OtrEventNumHit; }

/** Access the channel index of the i'th hit.

 @param   i  hit index (0 .. OtrEventHitNumberOf()-1)
 @return  channel index of the i'th hit
 */
inline int OtrEventHitChannelIndex(int i) { 
  return OtrEventHitGet()[i].ci; 
  }

/** Access the channel index of the i'th hit.

 @param   i  hit index (0 .. OtrEventHitNumberOf()-1)
 @return  channel index of the i'th hit
 */
inline int OtrEventHitChannelId(int i) { 
  return OtrIdFromIndex(OtrEventHitGet()[i].ci); 
  }

/** Access the x of the middle of the wire for the i'th hit.

 @param   i  hit index (0 .. OtrEventHitNumberOf()-1)
 @return  x of the middle of the wire for the i'th hit
 */
inline float OtrEventHitX(int i) { 
  return OtrEventHitGet()[i].x; 
  }

/** Access the z of the middle of the wire for the i'th hit.

 @param   i  hit index (0 .. OtrEventHitNumberOf()-1)
 @return  z of the middle of the wire for the i'th hit
 */
inline float OtrEventHitZ(int i) { 
  return OtrEventHitGet()[i].z; 
  }

/** Access the drift distance [cm] for the i'th hit.

 @param   i  hit index (0 .. OtrEventHitNumberOf()-1)
 @return  drift distance for the i'th hit
 */
inline float OtrEventHitR(int i) { 
  return OtrEventHitGet()[i].r; 
  }

/** Access the error on the drift distance [cm] for the i'th hit.

 @param   i  hit index (0 .. OtrEventHitNumberOf()-1)
 @return  error on the drift distance for the i'th hit
 */
inline float OtrEventHitRError(int i) { 
  return OtrEventHitGet()[i].rerror; 
  }


/** Access the geometry index for the i'th hit.

 @param   i  hit index (0 .. OtrEventHitNumberOf()-1)
 @return  index of volume (nede-volume) which is hit  for the i'th hit
 */
inline int OtrEventHitGeoIndex(int i) { 
  return OtrIGeoIndex(OtrEventHitGet()[i].ci); 
  }
  
/** Access TDC-value of the i'th hit (not available if wax was used).

 @param   i  hit index (0 .. OtrEventHitNumberOf()-1)
 @return  tdc value of the i'th hit
 */
inline unsigned char  OtrEventHitTdcValue(int i) { 
  return OtrEventHitGet()[i].tdcvalue; 
  }
  

#endif

/** Access event statistics: Tmean of all hits in superlayer.

 @param   sl superlayer identifier
 @return  Tmean of all hits of superlayer sl in event
 */
  float OtrEventStatisticsTmean(int sl);

/** Access event statistics: sum of TDC values of all hits in superlayer.

 @param   sl superlayer identifier
 @return  sum of TDC values of all hits of superlayer sl in event
 */
  int OtrEventStatisticsTsum(int sl);

/** Access event statistics: number of hits in superlayer used for Tsum.

 @param   sl superlayer identifier
 @return  number of hits in superlayer sl used for Tsum
 */
  int OtrEventStatisticsThits(int sl);

/** Access event statistics: Number of hits in superlayer

 @param   sl superlayer identifier
 @return  number of hits in superlayer sl 
 */
  int OtrEventStatisticsHits(int sl);

/** Access event statistics: check if off-BX event  

 @return 1 if off-BX event, else 0
 */
  int OtrEventStatisticsOffbx();



/*--------------------------------------------------------------------
        F E D  information in event
  ---------------------------------------------------------------------*/
/** Access FED Info: Number of FEDs in Event

 @return FedNumber;
 */
int            OtrEventFedNumberOf(); 

/** Access FED Info: Status of ith FED

  @param i
  @return FedStatus;
 */
int            OtrEventFedStatus(int hitindex);

/** Access FED Info: BX number of ith FED

  @param i
  @return FedBx;
  */
int            OtrEventFedBx(int hitindex);


/*--------------------------------------------------------*/
/*    Interface to the DAQ and Target information in      */
/*  event record.                                         */
/*--------------------------------------------------------*/

/** initialization of info structure. 

 @param Event structure;
 */

void OtrEventInfoEventGet(EVENT_HDR *event);

/** Access event info: Target IA Rate

 @return Ia_Rate;
 */
float OtrEventInfoIaRate();


/** Access event info: Target Bg Rate

 @return BgRate;
 */
float OtrEventInfoBgRate();


/** Access event info: RunNumb

 @return run_number;
 */
int OtrEventInfoRunNumb();


/** Access event info: RunNumber

 @return run_number;
 */
inline int OtrEventInfoRunNumber() { return OtrEventInfoRunNumb(); }

/** Access event info: Event Time

 @return event_time;
 */
int OtrEventInfoTime();

/** Access event info: Event Database Key used in online processing

 @return database key;
 */
int OtrEventInfoDbKey();

/** Access event info: FLT Status

 @return flt_status;
 */
int OtrEventInfoFltStatus();

/** Access event info: SLT Status

 @return stl_status;
 */
int OtrEventInfoSltStatus();

/** Access event info: FCS trigger Code

 @return fcs_tcode;
 */
int OtrEventInfoFcsTcode();


/** Access event info: Physical BX

 @return phys_BX;
 */
int OtrEventInfoPhysBx();

/** Access event info: FLT BX

 @return flt_bx;
 */
int OtrEventInfoFltBx();

/** Access event info: FLT Event

 @return flt_event;
 */
int OtrEventInfoFltEvent();


/*@}*/
#ifdef __cplusplus
}
#endif

#endif








