#include "Otr/OtrDb.h"
#include "Otr/OtrI.h"

#include <stdlib.h>
#include <stdio.h>

//extern "C" void OtrDbPrint(FILE* fp,OtrDbTable* tt);

/* Fortran Interface */
extern "C" {
  void  OtrGeoSetAlgorithm_(int* i)  { OtrGeoSetAlgorithm(*i); }
  int   OtrGeoInitialize_()          { return OtrGeoInitialize(); }
  int   OtrGeoUpdate_()              { return OtrGeoUpdate(); }
  int   OtrGeoWriteToDb_()           { return OtrGeoWriteToDb(); }
  int   OtrGeoClose_()               { return OtrGeoClose(); }
  int   OtrGeoTestMe_()              { return OtrGeoTestMe(); }
  }
  
static int OtrGeoAlgorithm=OTRIInvalid;
static int numberofvolumes=0;
static NEDE*    nede=0;
static int IsInit=0;
static unsigned char module2plate[40][7][31];

/* ----------------------------------------------------------------------*/
OtrDbField structfmt[]={
  {"int",   "cmp",   OtrDbOffset(NEDE,cmp)},
  {"int",   "lay",   OtrDbOffset(NEDE,lay)},
  {"int",   "sec",   OtrDbOffset(NEDE,sec)},
  {"int",   "shp",   OtrDbOffset(NEDE,shp)},
  {"float", "xmin",  OtrDbOffset(NEDE,xmin)},
  {"float", "xmax",  OtrDbOffset(NEDE,xmax)},
  {"float", "ymin",  OtrDbOffset(NEDE,ymin)},
  {"float", "ymax",  OtrDbOffset(NEDE,ymax)},
  {"float", "zmin",  OtrDbOffset(NEDE,zmin)},
  {"float", "zmax",  OtrDbOffset(NEDE,zmax)},
  {"float", "sxmn",  OtrDbOffset(NEDE,sxmn)},
  {"float", "sxmx",  OtrDbOffset(NEDE,sxmx)},
  {"float", "symn",  OtrDbOffset(NEDE,symn)},
  {"float", "symx",  OtrDbOffset(NEDE,symx)},
  {"float", "cos",   OtrDbOffset(NEDE,cos)},
  {"float", "sin",   OtrDbOffset(NEDE,sin)},

  {"float", "rot0",  OtrDbOffset(NEDE,rot[0])},
  {"float", "rot1",  OtrDbOffset(NEDE,rot[1])},
  {"float", "rot2",  OtrDbOffset(NEDE,rot[2])},
  {"float", "rot3",  OtrDbOffset(NEDE,rot[3])},
  {"float", "rot4",  OtrDbOffset(NEDE,rot[4])},
  {"float", "rot5",  OtrDbOffset(NEDE,rot[5])},
  {"float", "rot6",  OtrDbOffset(NEDE,rot[6])},
  {"float", "rot7",  OtrDbOffset(NEDE,rot[7])},
  {"float", "rot8",  OtrDbOffset(NEDE,rot[8])},

  {"float", "acc",   OtrDbOffset(NEDE,acc)},
  {"float", "eff",   OtrDbOffset(NEDE,eff)},
  {"float", "dod",   OtrDbOffset(NEDE,dod)},
  {"float", "pit",   OtrDbOffset(NEDE,pit)},
  {"float", "ms",    OtrDbOffset(NEDE,ms)},

  {"float", "xfw0",  OtrDbOffset(NEDE,xfw[0])},
  {"float", "xfw1",  OtrDbOffset(NEDE,xfw[1])},
  {"float", "xfw2",  OtrDbOffset(NEDE,xfw[2])},
  {"float", "xfw3",  OtrDbOffset(NEDE,xfw[3])},

  {"float", "zfw0",  OtrDbOffset(NEDE,zfw[0])},
  {"float", "zfw1",  OtrDbOffset(NEDE,zfw[1])},
  {"float", "zfw2",  OtrDbOffset(NEDE,zfw[2])},
  {"float", "zfw3",  OtrDbOffset(NEDE,zfw[3])},

  {"int",   "nwi0",  OtrDbOffset(NEDE,nwi[0])},
  {"int",   "nwi1",  OtrDbOffset(NEDE,nwi[1])},
  {"int",   "nwi2",  OtrDbOffset(NEDE,nwi[2])},
  {"int",   "nwi3",  OtrDbOffset(NEDE,nwi[3])},

  {"int",   "nsl",   OtrDbOffset(NEDE,nsl)},

                               // direct declarations give error in ARTE
  {"int",   "onesl", OtrDbOffset(NEDE,nsl)+4}, // OtrDbOffset(NEDE,onsl)
  {"int",   "ogmat", OtrDbOffset(NEDE,nsl)+8}, // OtrDbOffset(NEDE,ogmat)

  {"int",   "nwl",   OtrDbOffset(NEDE,nwl)},
  {"int",   "fwl",   OtrDbOffset(NEDE,fwl)},
  {"float", "xmad",  OtrDbOffset(NEDE,xmad)},
  {"float", "ymad",  OtrDbOffset(NEDE,ymad)},
  {"float", "zmad",  OtrDbOffset(NEDE,zmad)},

  {"float", "y1w0",  OtrDbOffset(NEDE,y1w[0])},
  {"float", "y1w1",  OtrDbOffset(NEDE,y1w[1])},
  {"float", "y1w2",  OtrDbOffset(NEDE,y1w[2])},
  {"float", "y1w3",  OtrDbOffset(NEDE,y1w[3])},

  {"float", "y2w0",  OtrDbOffset(NEDE,y2w[0])},
  {"float", "y2w1",  OtrDbOffset(NEDE,y2w[1])},
  {"float", "y2w2",  OtrDbOffset(NEDE,y2w[2])},
  {"float", "y2w3",  OtrDbOffset(NEDE,y2w[3])},

  {"char[]","gvol",  OtrDbOffset(NEDE,gvol)},
  {"char[]","hwc",   OtrDbOffset(NEDE,hwc)},
  {"float", "dedx",  OtrDbOffset(NEDE,dedx)},
  {"int",   "gcop",  OtrDbOffset(NEDE,gcop)},
  {"float", "noise", OtrDbOffset(NEDE,noise)},
  {"float", "ares",  OtrDbOffset(NEDE,ares)},
};

static OtrDbTable NEDETableDef = {"OTR_NEDE",1,"NEDE table",sizeof(NEDE),
			  structfmt,sizeof(structfmt)/sizeof(OtrDbField)};


/* ----------------------------------------------------------------------*/

void OtrGeoDefine() {
  OtrDbDefine("/OTR_CONFIG",NEDETableDef.name,OtrGeoUpdate);
  }
  
/* ----------------------------------------------------------------------*/

#ifdef USE_ARTE
  static char     (*nesl)[4]=0;
  static char     (*gmat)[4]=0;
  static int* inede=0;


/* ----------------------------------------------------------------------*/

static int FillArteNede(NEDE *nede,int *inede) {//,char nesl[4], char gmat[4]) {
  unsigned int i;
  char t[13];
    
  ArtePointer<NEDE> n;
  
  printf("FillArteNede[debug}: Enter function. NEDE-tablesize: %d\n",
                 ArtePointer<NEDE>::nrow());
    
  memset(inede,0xff,sizeof(int)*numberofvolumes);

  numberofvolumes=0;  
  int j=0;
  for (i=1;i<=ArtePointer<NEDE>::nrow();i++) {
    int k;
    n=ArtePointer<NEDE>(i);
    if (n->cmp!=2) continue;
    inede[numberofvolumes++]=i;
    
    continue;
    NEDE *m=&nede[j++];
    if (!isalpha(n->hwc[11])) n->hwc[11]=' ';
    memcpy(t,n->hwc,12);
    t[12]='\0';
    char s[13];
    memcpy(s,m->hwc,12);
    s[12]='\0';
    printf("nede: %d arte=%s db=%s\n",j,t,s);
    if (strcmp(s,t)) printf("xede: %d arte=%s db=%s\n",j,t,s);


    if (!memcmp(n->hwc,m->hwc,sizeof(m->hwc))) {
      inede[j-1]=i;
      }
    else {
      for (k=0;k<numberofvolumes;k++) {
        NEDE *m=&nede[k];
        if (fabs(n->xfw[0]-m->xfw[0])>1||fabs(n->zfw[0]-m->zfw[0])>1) continue;
        printf("found %d\n",k);
        inede[k]=i;
        break;
      
/*
      if ((fabs(n->xfw[0]-m->xfw[0])<1&&fabs(n->zfw[0]-m->zfw[0])<1) ||
          ((!memcmp(n->hwc,m->hwc,sizeof(m->hwc)))&&strcmp(t,"none"))
	  ) {
*/
	break;
      /*
        n->cmp=m->cmp;
        n->lay=m->lay;
        n->sec=m->sec;
        n->shp=m->shp;
        n->xmin=m->xmin;
        n->xmax=m->xmax;
        n->ymin=m->ymin;
        n->ymax=m->ymax;
        n->zmin=m->zmin;
        n->zmax=m->zmax;
        n->sxmn=m->sxmn;
        n->sxmx=m->sxmx;
        n->symn=m->symn;
        n->symx=m->symx;
        n->cos=m->cos;
        n->sin=m->sin;
        memcpy(n->rot,m->rot,sizeof(m->rot));
        n->acc=m->acc;
        n->eff=m->eff;
        n->dod=m->dod;
        n->pit=m->pit;
        n->ms=m->ms;
        memcpy(n->xfw,m->xfw,sizeof(m->xfw));
        memcpy(n->zfw,m->zfw,sizeof(m->zfw));
        memcpy(n->nwi,m->nwi,sizeof(m->nwi));

        n->nsl=m->nsl;
        n->nwl=m->nwl;
        n->fwl=m->fwl;
        n->xmad=m->xmad; 
        n->ymad=m->ymad;
        n->zmad=m->zmad;
        memcpy(n->y1w,m->y1w,sizeof(m->y2w));
        memcpy(n->y2w,m->y2w,sizeof(m->y2w));
        memcpy(n->gvol,m->gvol,sizeof(m->gvol));
        memcpy(n->hwc,m->hwc,sizeof(m->hwc));
        n->dedx=m->dedx;
        n->gcop=m->gcop;
        n->noise=m->noise;
        n->ares=m->ares;
        */
	}
      }
    if (k==numberofvolumes) {
      printf("FillArteNede[debug]: ARTE-NEDE not found in db table: %s\n",t);fflush(stdout);
      }
    }


  for (int i=0;i<numberofvolumes;i++) {
    if (inede[i]==-1) {
      char s[13];
      memcpy(s,nede[i].hwc,12);
      s[12]='\0';
      printf("FillArteNede[debug]: inede not assigned for %3d %s\n",i,s);
      }
    }
/*
  if ((i-1)==ArtePointer<NEDE>::nrow()) n=ArteTable<NEDE>::push_back(*m);
   
  for (i=1;i<=ArtePointer<NESL>::nrow();i++) {
    if (!memcmp(nesl,ArtePointer<NESL>(i)->name,sizeof(nesl))) {
      n->nesl=ArtePointer<NESL>(i);
      }
    }
  for (i=1;i<=ArtePointer<GMAT>::nrow();i++) {
    if (!memcmp(gmat,ArtePointer<GMAT>(i)->name,sizeof(gmat))) {
      n->gmat=ArtePointer<GMAT>(i);
      }
    }
*/    
  return OtrIValid();
  }
   
/* ----------------------------------------------------------------------*/

static void FillLocalNede(int inede,NEDE *n,char nesl[4], char gmat[4]) {
   ArtePointer<NEDE> m(inede);
   n->cmp=m->cmp;
   n->lay=m->lay;
   n->sec=m->sec;
   n->shp=m->shp;
   n->xmin=m->xmin;
   n->xmax=m->xmax;
   n->ymin=m->ymin;
   n->ymax=m->ymax;
   n->zmin=m->zmin;
   n->zmax=m->zmax;
   n->sxmn=m->sxmn;
   n->sxmx=m->sxmx;
   n->symn=m->symn;
   n->symx=m->symx;
   n->cos=m->cos;
   n->sin=m->sin;
   memcpy(n->rot,m->rot,sizeof(m->rot));
   n->acc=m->acc;
   n->eff=m->eff;
   n->dod=m->dod;
   n->pit=m->pit;
   n->ms=m->ms;
   memcpy(n->xfw,m->xfw,sizeof(m->xfw));
   memcpy(n->zfw,m->zfw,sizeof(m->zfw));
   memcpy(n->nwi,m->nwi,sizeof(m->nwi));
   n->nsl=m->nsl;

   n->nwl=m->nwl;
   n->fwl=m->fwl;
   n->xmad=m->xmad;
   n->ymad=m->ymad;
   n->zmad=m->zmad;
   memcpy(n->y1w,m->y1w,sizeof(m->y2w));
   memcpy(n->y2w,m->y2w,sizeof(m->y2w));
   memcpy(n->gvol,m->gvol,sizeof(m->gvol));
   memcpy(n->hwc,m->hwc,sizeof(m->hwc));
   n->dedx=m->dedx;
   n->gcop=m->gcop;
   n->noise=m->noise;
   n->ares=m->ares;
       
   memcpy(gmat,m->gmat->name,sizeof(m->gmat->name));
   memcpy(nesl,m->nesl->name,sizeof(m->nesl->name));
   }
   
#endif

#ifdef USE_ARTE
#define NEDE(i) ArtePointer<NEDE> n(inede[i]);
#else
#define NEDE(i) NEDE* n=&nede[i];
#endif

/* ----------------------------------------------------------------------*/
int OtrGeoCmp() { 
  NEDE(0)
  return n->cmp;
  }

int OtrGeoSuperLayer(int gi) {
  NEDE(gi)
  return OtrISuperLayerFromName(n->hwc);
  }

int OtrGeoLayer(int gi) {
  NEDE(gi)
  return n->lay;
  }

int OtrGeoSector(int gi) {
  NEDE(gi)
  return n->sec;
  }

int OtrGeoModule(int gi) {
  return atoi((OtrGeoGet(gi)->hwc)+6);
}

void OtrGeoId(int gi,char hwc[13]) {
  memcpy(hwc,OtrGeoGet(gi)->hwc,12);
  hwc[12]='\0';
}

/* ----------------------------------------------------------------------*/
/**
  Tasks:
    - switch between different solutions of the task of the package
  */

void  OtrGeoSetAlgorithm(int i) {
  OtrGeoAlgorithm=i;
  }
  
/* ----------------------------------------------------------------------*/

/**
  Tasks:
    - Memory allocation 
    - prepare for reinitialization
    - after this call reinitalization
    
  All function of the package (beside OtrGeoSetAlgorithm) have to check, 
  if the initialization was done properly.
  */

int   OtrGeoInitialize() {
  int rc;
  if (IsInit) return OtrIValid();
  if (OtrGeoAlgorithm==OTRIInvalid) OtrGeoAlgorithm=0;
  rc=OtrGeoUpdate();
  IsInit=1;
  return rc;
}
  
int   OtrGeoScan(FILE *fp) {
  char *s,line[1024];
  NEDE *n;
  int len;
  
  if (OtrGeoAlgorithm==OTRIInvalid) {
    return OtrIErrorMsg(OTRIInvalid,"OtrGeoUpdate","OtrGeo is not initialized","");
    }
  while (fgets(line,1024,fp)) {
    int k;
    if (strlen(line)==1023) {
      return OtrIErrorMsg(OTRIInvalid,"OtrGeoScan","Line exceed buffer lenght","");
      }
    nede=(NEDE*)realloc(nede,(numberofvolumes+=1)*sizeof(*nede));
    n=&nede[numberofvolumes-1];
    s=line;
    sscanf(s,"%d %d %d %d %n",&n->cmp,&n->lay,&n->sec,&n->shp,&len);s+=len;
    sscanf(s,"%e %e %e %e %e %e %n",&n->xmin,&n->xmax,&n->ymin,&n->ymax,&n->zmin,&n->zmax,&len);s+=len;
    sscanf(s,"%e %e %e %e %n",&n->sxmn,&n->sxmx,&n->symn,&n->symx,&len);s+=len;
    sscanf(s,"%e %e %n",&n->cos,&n->sin,&len);s+=len;
    for (k=0;k<9;k++) {
      sscanf(s,"%e %n",&n->rot[k],&len);s+=len;
      }
    sscanf(s,"%e %e %e %e %e %n",&n->acc,&n->eff,&n->dod,&n->pit,&n->ms,&len);s+=len;
    for (k=0;k<4;k++) {
      sscanf(s,"%e %n",&n->xfw[k],&len);s+=len;
      }
    for (k=0;k<4;k++) {
      sscanf(s,"%e %n",&n->zfw[k],&len);s+=len;
      }
    for (k=0;k<4;k++) {
      sscanf(s,"%d %n",&n->nwi[k],&len);s+=len;
      }
    sscanf(s,"%d %d %d %n",&n->nsl,&n->nwl,&n->fwl,&len);s+=len;
    sscanf(s,"%e %e %e %n",&n->xmad,&n->ymad,&n->zmad,&len);s+=len;
    for (k=0;k<4;k++) {
      sscanf(s,"%e %n",&n->y1w[k],&len);s+=len;
      }
    for (k=0;k<4;k++) {
      sscanf(s,"%e %n",&n->y2w[k],&len);s+=len;
      }
    for (k=0;k<4;k++) {
      sscanf(s,"%c",&n->gvol[k]);s+=1;
      }
    s+=1;
    for (k=0;k<12;k++) {
      sscanf(s,"%c",&n->hwc[k]);s+=1;
      }
    sscanf(s,"%e %d %e %e %n",&n->dedx,&n->gcop,&n->noise,&n->ares,&len);s+=len;
	
    }

//printf("numberofvolumes: %d\n",numberofvolumes);

#ifdef USE_ARTE  
  {
    int i;
    inede=(int*)malloc(numberofvolumes*sizeof(*inede));
    FillArteNede(nede,inede);
    for (i=0;i<numberofvolumes;i++) {
      char s[512];//,t[512];
      memcpy(s,nede[i].hwc,12);
      s[12]='\0';
      if (((unsigned int)inede[i])==(unsigned int) ArtePointer<NEDE>::nrow()+1) {
        if (s[11]!='n') printf("nede not found: %s\n",s);
	}
      }
    }
#endif    

  return OtrIValid();
  }
  
  
/* ----------------------------------------------------------------------*/
/* Prepare lookup table for plate No. on which the module resides.
 */

void   OtrGeoUpdatePlate() {
#ifdef DEBUG
  printf("Calling OtrGeoUpdatePlate\n");
#endif
  for (int sl=0; sl<OtrISuperLayerNumberOf(); sl++) {
    if (!OtrISuperLayerExist(sl)) continue;
    for (int lay=0; lay<=OtrILayerNumberOf(sl); lay++) {
      if (!OtrILayerExist(sl,lay)) continue;
      
      float zarr[40];
      int nplate = 0;
      for (int mod=0; mod<OtrIModuleNumberOf(sl,lay); mod++) {
	if (!OtrIModuleExist(sl,lay,mod)) continue;
	float zmin= 10000;
	float ymax=-10000;
	float ymin= 10000;
	float xmax=-10000;
	float xmin= 10000;
	int gi=0;
	for (int sec=0; sec<OtrISectorNumberOf(); sec++) {
	  gi=OtrIGeoIndexFromModule(sl,lay,mod,sec);
	  if (OtrIError()) continue;
	  if (zmin > OtrGeoGet(gi)->zmin) zmin = OtrGeoGet(gi)->zmin;
	  if (ymin > OtrGeoGet(gi)->ymin) ymin = OtrGeoGet(gi)->ymin;
	  if (ymax < OtrGeoGet(gi)->ymax) ymax = OtrGeoGet(gi)->ymax;
	  if (xmin > OtrGeoGet(gi)->xmin) xmin = OtrGeoGet(gi)->xmin;
	  if (xmax < OtrGeoGet(gi)->xmax) xmax = OtrGeoGet(gi)->xmax;
	}
	float x = 0.5*(xmin+xmax);
	float y1 = -nede->rot[1]*x + nede->rot[0]*ymin;
	float y2 = -nede->rot[1]*x + nede->rot[0]*ymax;
	
	int iplate;
	if (y1*y2<0) {
	  for (iplate=0; iplate<40; iplate++) {
	    if (fabs(zarr[iplate]-zmin)<1.e-4) {
	      module2plate[mod][lay][sl] = iplate;
	      goto endmod;
	    }
	  }
	}
	iplate = nplate;
	module2plate[mod][lay][sl] = nplate;
	zarr[nplate] = zmin;
	nplate++;
	if (nplate>=40) printf("Error in OtrGeoModulePlate\n");
	
      endmod: ;
	//	printf("sl %2d, lay %d, mod %2d: plate = %2d, zmin = %f, x = %7.2f, y1 = %7.2f, y2 = %7.2f\n",
	//sl,lay,mod,iplate,zmin,x,y1,y2);
      }
    }
  }
}

/* ----------------------------------------------------------------------*/
/**
  Tasks:
    - update memory structures, arrays ... from database or ascii files 
  */

int   OtrGeoUpdate() {
  char* fname=getenv("OtrGeoInput");
  int   version;
  if (OtrGeoAlgorithm==OTRIInvalid) {
    OtrGeoAlgorithm=0;
    }
  
  OtrGeoClose();
  IsInit=1;
  if (fname&&!fname[0]=='\0') {
    int ret;
    FILE *fp=fopen(fname,"r");
    if (!fp) {
      return OtrIErrorMsg(OTRIInvalid,"OtrGeoUpdate",
	                    "Cannot open file",fname);
    }
    //    printf("OtrGeoUpdate: scanning in file %s\n",fname);fflush(stdout);
    printf("OtrGeoUpdate[debug]: Update from file %s\n",fname);

    ret=OtrGeoScan(fp);
    fclose(fp);
    return ret;
  }
  else if ((version=OtrDbGetTableVersion(NEDETableDef.name))!=-1){
#ifndef USE_ARTE
    printf("OtrGeoUpdate[debug]: Update from db\n");
    nede = (NEDE*) OtrDbGet(&NEDETableDef,nede,&numberofvolumes);
#endif
#ifdef USE_ARTE
    /*
    inede=(int*)malloc(numberofvolumes*sizeof(*inede));
    FillArteNede(nede,inede);
    */
    printf("OtrGeoUpdate[debug]: ARTE Geometry will be used\n");
    inede=0;
    for (unsigned int i=1;i<=ArtePointer<NEDE>::nrow();i++) {
      ArtePointer<NEDE> n=ArtePointer<NEDE>(i);
      if (n->cmp==2) {
        inede=(int*)realloc(inede,(numberofvolumes+=1)*sizeof(*inede));
	inede[numberofvolumes-1]=i;
        }
      }
    nede=(NEDE*)malloc(numberofvolumes*sizeof(*nede));
    gmat=(char(*)[4])malloc(numberofvolumes*sizeof(*gmat));
    nesl=(char(*)[4])malloc(numberofvolumes*sizeof(*nesl));
    for (unsigned int i=0;i<(unsigned int)numberofvolumes;i++) {
      FillLocalNede(inede[i],&nede[i],nesl[i],gmat[i]);
      }
    }
  else if (ArtePointer<NEDE>::nrow()!=0) {
    unsigned int i;
    
    printf("OtrGeoUpdate[debug]: ARTE Geometry will be used\n");
    for (i=1;i<=ArtePointer<NEDE>::nrow();i++) {
      ArtePointer<NEDE> n=ArtePointer<NEDE>(i);
      if (n->cmp==2) {
        inede=(int*)realloc(inede,(numberofvolumes+=1)*sizeof(*inede));
	inede[numberofvolumes-1]=i;
        }
      }
    //	printf("number of volumes: %d\n",numberofvolumes);
    nede=(NEDE*)malloc(numberofvolumes*sizeof(*nede));
    gmat=(char(*)[4])malloc(numberofvolumes*sizeof(*gmat));
    nesl=(char(*)[4])malloc(numberofvolumes*sizeof(*nesl));
    for (i=0;i<(unsigned int)numberofvolumes;i++) {
      FillLocalNede(inede[i],&nede[i],nesl[i],gmat[i]);
    }
#endif
  }
  else  {
    return OtrIErrorMsg(OTRIInvalid,"OtrGeoUpdate",
			"Nether is Db open nor enviroment variable OtrGeoInput is set nor NEDE-Table is initialized.",
			"Cannot update values");
    
  }
  return OtrIValid();
}


int OtrGeoPrint(FILE *fp) {
  if (OtrGeoAlgorithm==OTRIInvalid) {
    return OtrIErrorMsg(OTRIInvalid,"OtrGeoUpdate","OtrGeo is not initialized","");
    }
  int i;
  for (i=0;i<numberofvolumes;i++) {
    int k;
    
    NEDE* n=&nede[i];

    fprintf(fp,"%d %d %d %d ",n->cmp,n->lay,n->sec,n->shp);
    fprintf(fp,"%e %e %e %e %e %e ",n->xmin,n->xmax,n->ymin,n->ymax,n->zmin,n->zmax);
    fprintf(fp,"%e %e %e %e ",n->sxmn,n->sxmx,n->symn,n->symx);
    fprintf(fp,"%e %e ",n->cos,n->sin);
    for (k=0;k<9;k++) fprintf(fp,"%e ",n->rot[k]);
    fprintf(fp,"%e %e %e %e %e ",n->acc,n->eff,n->dod,n->pit,n->ms);
    for (k=0;k<4;k++) fprintf(fp,"%e ",n->xfw[k]);
    for (k=0;k<4;k++) fprintf(fp,"%e ",n->zfw[k]);
    for (k=0;k<4;k++) fprintf(fp,"%d ",n->nwi[k]);
    fprintf(fp,"%d %d %d ",n->nsl,n->nwl,n->fwl);
    fprintf(fp,"%e %e %e ",n->xmad,n->ymad,n->zmad);
    for (k=0;k<4;k++) fprintf(fp,"%e ",n->y1w[k]);
    for (k=0;k<4;k++) fprintf(fp,"%e ",n->y2w[k]);
    for (k=0;k<4;k++) fprintf(fp,"%c",n->gvol[k]);
    fprintf(fp," ");
    for (k=0;k<12;k++) fprintf(fp,"%c",n->hwc[k]);
    fprintf(fp," ");
    fprintf(fp,"%e %d %e %e ",n->dedx,n->gcop,n->noise,n->ares);
    fprintf(fp,"\n");
    }
  return OtrIValid();
  };
  
/* ----------------------------------------------------------------------*/
/**
  Tasks:
    - provide output memory structures, arrays ... to database or ascii files 
  */
  
int   OtrGeoWriteToDb() {
  if (OtrGeoAlgorithm==OTRIInvalid) {
    return OtrIErrorMsg(OTRIInvalid,"OtrGeoUpdate","OtrGeo is not initialized","");
    }
  /*  
  printf("numberofvolumes=%d\n",numberofvolumes);
  printf("sizeof(NEDE)=%d\n",sizeof(NEDE));
  printf("sizeof(structfmt)=%d\n",sizeof(structfmt));
  printf("sizeof(OtrDbField)=%d\n",sizeof(OtrDbField));
  */
  printf("OtrGeoWriteToDb[info]: Write %d volumes to db\n",numberofvolumes);
  OtrDbPut(&NEDETableDef,nede,numberofvolumes);
  /*
  printf("OtrGeoWriteToDb: OtrDbPut has been called.\n");
  printf("OtrGeoWriteToDb: database written.\n");
  */ 
  return OtrIValid();
  }
 
/* ----------------------------------------------------------------------*/
/**
  Tasks:
    - free allocate memory
    - clean up to be in a state that initialization is necessary 
  */

int   OtrGeoClose()     { 
  if (OtrGeoAlgorithm==OTRIInvalid) {
    return OtrIErrorMsg(OTRIInvalid,"OtrGeoUpdate","OtrGeo is not initialized","");
    }

  numberofvolumes=0;
  if (nede) {
    free(nede);
    nede=0;
    }
  IsInit=0;
  return OtrIValid();
  }
  
/* ----------------------------------------------------------------------*/

/** plate No. on which the module resides.

  @return No. of plate
 */

int OtrGeoModulePlate(int sl,int lay,int mod) {
  if (!OtrIModuleExist(sl,lay,mod)) return -1;
  static int first=1;
  if (first) {
    OtrGeoUpdatePlate();
    first=0;
  }
  return module2plate[mod][lay][sl];
}

/* ----------------------------------------------------------------------*/
/**
  Tasks:
    - test bench to be sucessfully executed before any release of a new version
    - it is requested, that the author carefully thinks about testing
    - in addition it should be a collection how to test bugs, which happened
      at least once 
  */
  
int   OtrGeoTestMe()    {
/*  char cmd[512];*/

  OtrIValid();

  printf("OtrGeoTestMe[info]: Printf tables\n");fflush(stdout);
  putenv("OtrGeoOutput=Geo.test");
  OtrGeoInitialize();
  OtrGeoDefine();
  OtrIValid();

#ifdef DEBUG
  fprintf(stdout,"Calling OtrDbPrint...\n");
  OtrDbPrint(stdout,&NEDETableDef);
#endif
  if (OtrGeoWriteToDb()==OTRIInvalid) {
    printf("OtrGeoTestMe: Error during write to Database\n");
    return OTRIInvalid;
    }

  /*
  printf("check\n");fflush(stdout);
  sprintf(cmd,"diff %s %s | more\n",getenv("OtrGeoInput"),getenv("OtrGeoOutput"));
  printf("exec %s\n",cmd);fflush(stdout);
  system(cmd);
  */

  OtrGeoClose();
  putenv("OtrGeoInput=");
  OtrGeoUpdate();
  FILE *fp=fopen("Geo.db","w");
  OtrGeoPrint(fp);
  fclose(fp);

  return OtrIValid();
  }
   
/* ----------------------------------------------------------------------*/
/** Test bench for OtrGeoModulePlate.
  */
  
int   OtrGeoPlateTestMe()    {
  for (int sl=0; sl<OtrISuperLayerNumberOf(); sl++) {
    if (!OtrISuperLayerExist(sl)) continue;
    for (int lay=0; lay<OtrILayerNumberOf(sl); lay++) {
      if (!OtrILayerExist(sl,lay)) continue;
      for (int mod=0; mod<OtrIModuleNumberOf(sl,lay); mod++) {
	if (!OtrIModuleExist(sl,lay,mod)) continue;
	int iplate = OtrGeoModulePlate(sl,lay,mod);
	printf(" sl %2d, lay %d, mod %2d ==> plate %2d\n",sl,lay,mod,iplate);
      }
      printf("\n");
    }
  }
  
  return OtrIValid();
}

/* ----------------------------------------------------------------------*/
extern "C" {
int OtrGeoNumberOfVolumes() {
  return numberofvolumes;
  }
}
/* ----------------------------------------------------------------------*/

#ifdef USE_ARTE
ArtePointer<NEDE> OtrGeoGet(int gi) {
  if (gi<0||gi>=numberofvolumes) {
    return OtrIErrorMsg(OTRIOutOfRange,"OtrGeoGet","Index out of range","");
    }  
  return ArtePointer<NEDE>(inede[gi]);
  }

int OtrGeoArteIndex(int ci) {
  int gi=OtrIGeoIndex(ci);
  if (gi<0||gi>=numberofvolumes) {
    char s[256];
    sprintf(s,"gi=%d  nvol=%d",gi,numberofvolumes);
    return OtrIErrorMsg(OTRIOutOfRange,"OtrGeoArteIndex","Index out of range",s);
    }  
 return inede[gi]; 
 }
#else
const NEDE* OtrGeoGet(int gi) {
  NEDE* n=&nede[gi];
  if (gi<0||gi>=numberofvolumes) {
    OtrIErrorMsg(OTRIOutOfRange,"OtrGeoGet","Index out of range","");
    n=0;
    }
  return n;
  }
#endif
    
/* ----------------------------------------------------------------------*/

void OtrGeoVolumeNominal(int gi,float c1[3], float c2[3]) {
  NEDE(gi)
  if (gi<0||gi>=numberofvolumes) {
    OtrIErrorMsg(OTRIOutOfRange,"OtrGeoVolume","Index out of range","");
    return;
    }

  // Extract value in rotated system of NEDE plane
  float r1[3],r2[3];
  if (n->sec == 0) {
    r1[0]=n->xmin;
    r2[0]=n->xmax;
    r1[1]=n->ymin;
    r2[1]=n->ymax;
  }
  else {
    r1[0]=n->sxmn;
    r2[0]=n->sxmx;
    r1[1]=n->symn;
    r2[1]=n->symx;
  }
  r1[2]=n->zmin;
  r2[2]=n->zmax;

  // Transform to laboratory system
  for (int i=0; i<3; i++) {
    c1[i] = 0;
    c2[i] = 0;
    for (int k=0; k<3; k++) {
      c1[i] += n->rot[3*i+k] * r1[k];
      c2[i] += n->rot[3*i+k] * r2[k];
    }
  }
}

/* ----------------------------------------------------------------------*/

#define min(a,b) (((a)<(b)? (a):(b)))
#define max(a,b) (((a)>(b)? (a):(b)))

int OtrGeoMatchInitialize(int sl,int lay,int flag,float ylimit,
			  OtrGeoMatchStruct *s) {
  s->sl = sl;
  s->lay = lay;
  s->flag = flag;
  s->gifirst = 0;
  s->gilast = 0;
  s->ylimit = ylimit;
#ifdef DEBUG
  printf("OtrGeoMatchInitialize was initialized for superlayer %d, stereolayer %d\n",
	 sl,lay);
  printf("Parameters are flag=%d, Y limit = %7.2f\n",flag,ylimit);
#endif
  return 0;
}

int OtrGeoMatch(OtrGeoMatchStruct *s) {
#ifdef DEBUG
  static int dgi=0;
  static float dxmax=0.;
#endif
  int same=1;
  
  if (s->gilast==0) {
    // Find out first and last module of this stereolayer
    s->gifirst = numberofvolumes;
    s->gilast = 0;
    for (int gi=0; gi<numberofvolumes; gi++) {
#ifdef USE_ARTE
      ArtePointer<NEDE> n(inede[gi]);
#else
      NEDE* n=&nede[gi];
#endif
      int sl=OtrISuperLayerFromName(n->hwc);
      int lay=n->lay;
      if (sl==s->sl && lay==s->lay) {
	if (gi<s->gifirst) s->gifirst = gi;
	if (gi>s->gilast ) s->gilast  = gi;
      }
    }
    if (s->gifirst>s->gilast) {
      printf("OtrGeoMatch Error: sl=%d, lay=%d not found in OTR\n",s->sl,s->lay);
      return 0;
    }
    s->gi1 = s->gifirst;
#ifdef DEBUG
    printf("OtrGeoMatch: gifirst=%d, gilast=%d\n", s->gifirst, s->gilast);
#endif
    same = 0;
  }
  
  if (s->gi1 == -1) {
    printf("OtrGeoMatch called after all pairs have been found\n");
    return 0;
  }
  
  for (int gi1=s->gi1; gi1<s->gilast; gi1++) {
#ifdef USE_ARTE
    ArtePointer<NEDE> n1(inede[gi1]);
#else
    NEDE* n1=&nede[gi1];
#endif
    if (n1->sec==0) continue;
    if (gi1 != s->gi1) same=0;

    for (int gi2= same? s->gi2 : gi1+1; gi2 <= min(gi1+30,s->gilast); gi2++) {
#ifdef USE_ARTE
      ArtePointer<NEDE> n2(inede[gi2]);
#else
      NEDE* n2=&nede[gi2];
#endif
      if (n2->sec==0) continue;
      if (gi2 != s->gi2) same=0;

      //printf("gi1=%d, gi2=%d\n",gi1,gi2);

      static float xmin=999,xmax=999,ymin=999,ymax=999;
      if (same==0) {
	xmin=max(n1->sxmn,n2->sxmn);
	xmax=min(n1->sxmx,n2->sxmx);
	if (xmin>xmax) continue;

	ymin=max(n1->symn,n2->symn);
	ymax=min(n1->symx,n2->symx);
	if (ymin - ymax + s->ylimit > 0.) continue;

	s->dxmax = 0.2499*min(n1->pit,n2->pit);
#ifdef DEBUG
	if (xmax-xmin > dxmax) {
	  dxmax = xmax-xmin;
	  if (dxmax>10.) printf(" *************** dxmax = %f\n",dxmax);
	}
	printf("gi1=%4d, gi2=%4d, xmin=%7.2f, xmax=%7.2f, ymin=%7.2f, ymax=%7.2f, dxmax=%5.2f\n",
	       gi1, gi2, xmin, xmax, ymin, ymax, s->dxmax);
#endif
      }

      for (int mono1= same? s->mono1 : 0; mono1 < n1->nsl; mono1++) {
	//printf("    mono1=%d\n",mono1);
	if (mono1 != s->mono1) same=0;
	for (int mono2= same? s->mono2 : 0; mono2 < n2->nsl; mono2++) {
	  //printf("      mono2=%d\n",mono2);
	  if (mono2 != s->mono2) same=0;

	  int w1f = max((int) floor((xmin - n1->xfw[mono1])/n1->pit), 0);
	  int w1l = min((int) floor((xmax - n1->xfw[mono1])/n1->pit), n1->nwi[mono1]-1);

	  for (int w1= same? s->wire1+1 : w1f; w1 <= w1l; w1++) {
	    //printf("        w1=%d\n",w1);
	    if (mono1 != s->mono1) same=0;
	    float x1 = n1->xfw[mono1] + w1*n1->pit;

	    int w2 = (int) floor((x1 - n2->xfw[mono2]) / n2->pit + 0.5);
	    if (w2<0) continue;
	    if (w2 > n2->nwi[mono2]-1) continue;
	    float x2 = n2->xfw[mono2] + w2*n2->pit;

	    int sl1=OtrGeoSuperLayer(gi1);
	    int mod1=OtrGeoModule(gi1);
	    int ci1 = OtrIFromModule(sl1,n1->lay,mod1,n1->sec,mono1,w1);

	    int sl2=OtrGeoSuperLayer(gi2);
	    int mod2=OtrGeoModule(gi2);
	    int ci2 = OtrIFromModule(sl2,n2->lay,mod2,n2->sec,mono2,w2);

	    if (ci1==OTRIInvalid) continue;
	    if (ci2==OTRIInvalid) continue;

	    if (fabs(x1-x2)<s->dxmax) {
	      s->gi1 = gi1;
	      s->gi2 = gi2;

	      s->mod1 = mod1;
	      s->ci1 = ci1;
	      s->mod2 = mod2;
	      s->ci2 = ci2;

	      s->mono1 = mono1;
	      s->mono2 = mono2;
	      s->wire1 = w1;
	      s->wire2 = w2;
	      s->x1 = x1;
	      s->x2 = x2;
	      s->dx = x1-x2;
	      s->dz = n1->zfw[mono1] - n2->zfw[mono2];

#ifdef DEBUG
	      if (gi2-gi1 > dgi) {
		dgi = gi2-gi1;
		if (dgi>3) printf(" *************** dgi = %d\n",dgi);
	      }
#endif
	      return 1;
	    }
	  }
	}
      }
    }
  }

#ifdef DEBUG
  printf("OtrGeoMatch: dgi = %d\n",dgi);
  printf("OtrGeoMatch: dxmax = %f\n",dxmax);
#endif
  s->gi1 = -1;
  s->gi2 = -1;
  s->ci1 = -1;
  s->ci2 = -1;
  s->mod1 = -1;
  s->mod2 = -1;
  s->mono1 = -1;
  s->mono2 = -1;
  s->wire1 = -1;
  s->wire2 = -1;
  s->x1 = 0.;
  s->x2 = 0.;
  s->dx = 0.;
  s->dz = 0.;
  return 0;
}

int OtrGeoMatchTestMe() {
  int total=0;
  for (int sl=1; sl<32; sl++) {
    if (!OtrISuperLayerExist(sl)) continue;

    for (int lay=1; lay<=6; lay++) {
      if (!OtrILayerExist(sl,lay)) continue;
      
      printf("\n========================================================\n\n");

      OtrGeoMatchStruct m;
      OtrGeoMatchInitialize(sl, lay, 0, 7., &m);
      int i=0;
      while (OtrGeoMatch(&m)) {
	i++;
	total++;

	float x1 = OtrGeoGet(m.gi1)->xfw[m.mono1] + m.wire1 * OtrGeoGet(m.gi1)->pit;
	float x2 = OtrGeoGet(m.gi2)->xfw[m.mono2] + m.wire2 * OtrGeoGet(m.gi2)->pit;

	if (fabs(x1 - m.x1)>1.e-5) printf("x1 = %f, m.x1 = %f\n",x1,m.x1);
	if (fabs(x2 - m.x2)>1.e-5) printf("x2 = %f, m.x2 = %f\n",x2,m.x2);

	if (m.gi1 != OtrIGeoIndex(m.ci1))
	  printf("m.gi1 = %d, OtrIGeoIndex(m.ci1) = %d\n",
		 m.gi1,OtrIGeoIndex(m.ci1));

	printf("%3d: gi1 %4d \"%12.12s\", mono1 %d, w1 %2d, x=%7.2f <==> "
                    "gi2 %4d \"%12.12s\", mono2 %d, w2 %2d, x=%7.2f\n",
	       i, m.gi1, OtrGeoGet(m.gi1)->hwc, m.mono1, m.wire1, m.x1,
	          m.gi2, OtrGeoGet(m.gi2)->hwc, m.mono2, m.wire2, m.x2);

	//printf("mod1=%d, ci1=%d, mod2=%d, ci2=%d\n",m.mod1,m.ci1,m.mod2,m.ci2);
      }
      printf("For Superlayer %d, layer %d there are %d matches\n",sl,lay,i);
    }
  }
  printf("\nFor OTR there are total %d matches\n",total);

  return 0;
}

/* ----------------------------------------------------------------------*/
/* ---------------------- KUIP routines ---------------------------------*/

#ifdef USE_ARTE
#ifdef DEBUG
extern "C" {
  extern char* ku_gets();
  extern int ku_geti();
  /* Test of writing to the database */
  void OtrGeoDb_kuip() {
    //    int i = ku_geti();
    //    printf("Setting OtrAlignAlgorithm to %d.\n",i);
    
    
    char DBname[100];
    strcpy(DBname, ku_gets());
    //    int maj = ku_geti();
    //    int min = ku_geti();
    
    //OtrDbUserDefine("/OTR_CONFIG","OTR_NEDE",OtrGeoUpdate,
    //		    TABLEmaxrev,TABLEmaxrev,TABLElt);
    char env[100];
    strcpy(env,"OTR_TEST_SERVER=");
    strcat(env,DBname);
    printf("Putting environment string: %s\n",env);
    putenv(env);
    
    OtrGeoUpdate();         
    OtrGeoWriteToDb();
  }
  
  /* Write an ASCII dump of geometry */
  void OtrGeoDump_kuip() {
    
    char dumpfile[100];
    strcpy(dumpfile, ku_gets());
    printf("Dumping OTR geometry to ASCII file %s\n",dumpfile);
    FILE *fp=fopen(dumpfile,"w");
    OtrGeoPrint(fp);
    fclose(fp);
    printf("Dumping finished\n");
  }
}
#endif
#endif

/*@}*/
