#include "Otr/OtrI.h"
#include "Otr/OtrDq.h"
#include "Otr/OtrRt.h"
#include "Otr/OtrT0.h"
#include "Otr/OtrEvent.h"

#include <stdarg.h>

#ifndef USE_ARTE
#include "grr.h"
#endif

/* ------------------------------------------------------------------------ */

int otrierrorflag;

static int otridebugswitch=OTRIDebugToStd|(!OTRIDebug&!OTRIError);
int otridebugtostd =1;

static int mode=OTRIUseEnvironment|OTRIUseDefaults;

/* ------------------------------------------------------------------------ */

void OtrISwitch(int i) {
  mode=i;
  }

/* ------------------------------------------------------------------------ */

void OtrIDebugSwitch( int i ) {
  otridebugswitch=i;
  otridebugtostd=(i&OTRIDebugToStd)!=0;
  }     

/* ------------------------------------------------------------------------ */
  
int OtrIInitialize() {
  char *align=getenv("OtrAlignInput");
  char *geo=getenv("OtrGeoInput");
  char *fe=getenv("OTRDECODEUseCables");
  char *fed=getenv("OTRDECODEUseSharc");
  char *dq=getenv("OtrDqInput");
  char *rt=getenv("OtrRtInput");
  char *t0=getenv("OtrT0Input");
  
  int ialign=0;
  int igeo=0;
  int idecode=0;
  int idq=0;
  int irt=0;
  int it0=0;
  int decodemode=0;

  printf("OtrIInitialize[info]: settings: %x  debug: %x\n",
                                           mode,otridebugswitch);
  if ((mode&OTRIUseDb)) {
    OtrDbOpen();
    if (!(geo&&(mode&OTRIUseEnvironment))) OtrGeoDefine();
    if (!(fe&&fed&&(mode&OTRIUseEnvironment))) OtrDecodeDefine();
    if (!(align&&(mode&OTRIUseEnvironment))) OtrAlignDefine();
    if (!(dq&&(mode&OTRIUseEnvironment))) OtrDqDefine();
    else printf("OtrIInitialize[info]: Take masks from file %s\n",dq);
    if (!(rt&&(mode&OTRIUseEnvironment))) OtrRtDefine();
    if (!(t0&&(mode&OTRIUseEnvironment))) OtrT0Define();
    else printf("OtrIInitialize[info]: Take T0 from file %s\n",t0);
    decodemode=SHARCTAB|CABTAB;
    OtrDecodeSetAlgorithm(decodemode);
    ialign=1;
    igeo=1;
    idecode=1;
    idq=1;
    irt=1;
    it0=1;    
    return OtrIValid();
    }

  if (!(mode&OTRIUseEnvironment)) {
    putenv("OtrAlignInput=");
    putenv("OtrGeoInput=");
    putenv("OTRDECODEUseCables=");
    putenv("OTRDECODEUseSharc=");
    putenv("OtrDqInput=");
    putenv("OtrRtInput=");
    putenv("OtrT0Input=");
    putenv("OtrDbServerMode=");
    }
  else {
    if (align) ialign=1;
    if (geo) igeo=1;
    if(fe)  { idecode=1;decodemode|=CABTABFILE;}
    if(fed) { idecode=1;decodemode|=SHARCTABFILE;}
    if(dq)  idq=1;
    if (rt) irt=1;
    if (t0) it0=1;

    if ((mode&OTRIUseDefaults)) {
      if (!align) {
//        putenv("OtrAlignInput=/afs/desy.de/group/hera-b/OTR/dev/Tables/OtrAlign/align.tab");
//        ialign=1;
        }
      if (!geo) {
#ifndef USE_ARTE
        putenv("OtrGeoInput=/afs/desy.de/group/hera-b/OTR/dev/Tables/OtrGeo/nede.tab");
#endif
        igeo=1;
        }
      if (!fe&&((decodemode&CABTABFILE==0)||((decodemode&CABTABFILE)!=CABTAB))) {
        putenv("OtrDecodeCables=/afs/desy.de/group/hera-b/OTR/dev/Tables/OtrDecode/cabling/list_of_tables_hallw");
        putenv("OTRDECODEUseCables=/afs/desy.de/group/hera-b/OTR/dev/Tables/OtrDecode/cabling/list_of_tables_hallw");
        idecode=1;
        decodemode|=CABTABFILE;
        }
      if (!fed&&((decodemode&SHARCTABFILE==0)||((decodemode&SHARCTABFILE)!=SHARCTAB))) {
        putenv("OtrDecodeSharc=/afs/desy.de/group/hera-b/OTR/dev/Tables/OtrDecode/fed/tdc_fed_hallw");
        putenv("OTRDECODEUseSharc=/afs/desy.de/group/hera-b/OTR/dev/Tables/OtrDecode/fed/tdc_fed_hallw");
        decodemode|=SHARCTABFILE;
        idecode=1;
        }
      if (!rt) {
        irt=1;
        putenv("OtrRtInput=/afs/desy.de/group/hera-b/OTR/dev/Tables/OtrRt/OtrRt.tab");
        }
      if (!t0) {
        it0=1;
        putenv("OtrT0Input=/afs/desy.de/group/hera-b/OTR/db/tzero/run14154.tzero");
        }
      }
    }
      
  OtrIValid();
  if (igeo) OtrGeoInitialize();
  if (OtrIError()) return OTRIInvalid;
  if (idecode) {
    OtrDecodeSetAlgorithm(decodemode);
    OtrDecodeInitialize();
    }
  if (OtrIError()) return OTRIInvalid;
  if (ialign) OtrAlignInitialize();
  if (OtrIError()) return OTRIInvalid;
  if (irt) OtrRtInitialize();
  if (OtrIError()) return OTRIInvalid;
  if (it0) OtrT0Initialize();
  if (OtrIError()) return OTRIInvalid;
  if (idq) {
    printf("OtrInitialize: Now init dq\n");fflush(stdout);
    OtrDqSwitch(OTRDQSelectOccupOr);
    if (OtrDqInitialize()==OTRIInvalid) return OtrIInvalid();
    }
  else {
    OtrEventSwitch(2);
    }
  return OtrIValid();
  }
  
/* ------------------------------------------------------------------------ */
 
int OtrIGetTables(int runno,int time,const OtrDbTableStruct **t,int *n) {
  OtrISwitch(OTRIUseDb);
  OtrIInitialize();
  return OtrDbGetTableList(runno,time,t,n);
  }



/* ------------------------------------------------------------------------ */
          
int otri_errormsg(const char *fcnname,int severity, int errorflag,char *msg) {
  char level[20];
  int  outfl=0;
  
  char *fname;
  size_t n=1024;
  char *s=(char*)malloc(n*sizeof(char));
  int i,l,rc;
  
  level[0]='\0';
  
  if (fcnname) fname=(char*) fcnname;
  else {
    l=strlen(msg);
    for (i=0;i<l;i++) {
      if (msg[i]==':' || msg[i]=='[') break;
      }
    if (i!=l) {
      fname=(char*)malloc((i+1)*sizeof(char));
      strncpy(fname,msg,i);
      fname[i]='\0';
      if (msg[i]=='[') {
	int k;
	for (k=i+1;k<l;k++) 
	  if (msg[k]==']') break;
        if (k!=l) strncpy(level,&msg[i+1],k-i-1);
	level[k-i-1]='\0';
	i=k;
	
	if (!strcmp(level,"debug")) severity=OTRIDebug;
	else if (!strcmp(level,"info")) severity=OTRIInfo;
	else if (!strcmp(level,"warning")) severity=OTRIWarning;
	else if (!strcmp(level,"error")) severity=OTRIError;
	else if (!strcmp(level,"SEVERE")) severity=OTRISevere;
	else if (!strcmp(level,"FATAL")) severity=OTRIFatal;
	}

      strcpy(msg,&msg[i+2]);
      }
    else fname=strdup("Unknown");
    }

  switch (severity) {
    case OTRIDebug:   strcpy(level,"debug"); 
                      outfl|=(otridebugswitch&OTRIDebug);
		      break;
    case OTRIInfo:    strcpy(level,"info"); 
                      outfl|=(otridebugswitch&OTRIInfo);
                      break;
    case OTRIWarning: strcpy(level,"warning"); 
                      outfl|=(otridebugswitch&OTRIWarning);
                      break;
    case OTRIError:   strcpy(level,"error"); 
                      outfl|=(otridebugswitch&OTRIError);
                      break;
    case OTRISevere:  strcpy(level,"SEVERE"); 
                      outfl|=(otridebugswitch&OTRISevere);
                      break;
    case OTRIFatal:   strcpy(level,"FATAL"); 
                      outfl|=(otridebugswitch&OTRIFatal);
                      break;
    default:          strcpy(level,"Unknown");
		      
    }

  otrierrorflag=errorflag;
  
  if (outfl) {
    char error[80];

    switch (errorflag) {
      case -3: strcpy(error,"Out of range"); break;
      case -2: strcpy(error,"Initialization missing"); break;
      case -1: strcpy(error,"Fishy");   break;
      case  0: strcpy(error,"None");    break;
      }

    l=strlen(msg);
    for (i=l-1;i>0;i--) {
      if (msg[i]!=' ') break;
      }
    if (msg[i]=='\n') msg[i]='\0';

    if (errorflag!=0) {  
      while (snprintf(s,n,"Otr[%s] %s: %s: error %s",level,fname,msg,error)<0) {
        n+=1024;
        s=(char*)realloc(s,n*sizeof(char));
        }
      }
    else {
      while (snprintf(s,n,"Otr[%s] %s: %s",level,fname,msg)<0) {
        n+=1024;
        s=(char*)realloc(s,n*sizeof(char));
        }
      }
    
    if (otridebugswitch&OTRIDebugToStdout) {
      rc=puts(s);
      fflush(stdout);
      }
    
    if (otridebugswitch&OTRIDebugToStderr) {
      rc=fputs(s,stderr);
      }
    
#ifndef USE_ARTE
    if (otridebugswitch&OTRIDebugToGrr) {
      int sev=0;
      int err=0;
      int l=strlen(s);
      char* t=(char*)malloc(l*sizeof(char));
    
      while (snprintf(t,l,"%s: %s",fname,msg)<0) {
        l+=256;
        t=(char*)realloc(t,l*sizeof(char));
        }
    
      switch (severity) {
        case OTRIDebug:   sev=GRR_SDEBUG;   err=GRR_OK;  break;
        case OTRIInfo:    sev=GRR_SINFO;    err=GRR_OK;  break;
        case OTRIWarning: sev=GRR_SWARNING; err=GRR_FISHY; break;
        case OTRIError:   sev=GRR_SERROR;   err=GRR_FISHY; break;
        case OTRISevere:  sev=GRR_SEVERE;   err=GRR_FISHY; break;
        case OTRIFatal:   sev=GRR_SFATAL;   err=GRR_FISHY; break;
        }  
      rc=grr_error(sev,t,err);
      free(t);
      }
#endif
    if (fname!=fcnname) free(fname);
    return rc;
    } /* end of if (outfl)  */
  return -1;    
  }

/* ------------------------------------------------------------------------ */
  
int otri_printf(char *format, ...) {
  int rc=0;
  if (otridebugtostd) {
    int n=1024;
    char* s=(char*)malloc(n*sizeof(char));
  
    va_list ap;
    va_start(ap,format);
    while ((rc=vsnprintf(s,n,format,ap))<0) {
      n+=1024;
      s=(char*)realloc(s,n*sizeof(char));
      }
    va_end(ap);

    if (rc>=0) rc=otri_errormsg(0,OTRIDebug,otrierrorflag,s);
    free(s);
    }
  return rc;
  }

int otri_fprintf(FILE* fp,char *format, ...) {
  if ((fp!=stdout&&fp!=stderr)||(otridebugtostd)) {
    int n=1024;
    char* s=(char*)malloc(n*sizeof(char));
    int err,lev=-1;
    int rc=0;
  
    va_list ap;
    va_start(ap,format);
    while (vsnprintf(s,n,format,ap)<0) {
      n+=1024;
      s=(char*)realloc(s,n*sizeof(char));
      }
    va_end(ap);
  
    if (fp==stdout)   { lev=OTRIDebug; err=OtrIValid(); }
    else if (fp==stderr) { lev=OTRIError; err=OtrIInvalid(); }

    if (lev!=-1) rc=otri_errormsg(0,lev,err,s);
    else rc=fputs(s,fp);
   
    free(s);
    return rc;
    }
  return -1;
  }

/* ------------------------------------------------------------------------ */
  
  

typedef struct {
  unsigned int val,cmp,min,max;
  char *name;
  } TestId;

static int err;

void CheckId(TestId* id) {
  if (id->val!=id->cmp)   { 
    err=1;
    printf("%s decoding of ChannelId is wrong %d %d\n",id->name,id->val,id->cmp);
//    getchar();
    }
  else {   
    if (id->val>id->max) id->max=id->val;
    if (id->val<id->min) id->min=id->val;
    }
  }
  
  
int OtrDecodeTest() {
  int i,j;
  int notusedtdc=0;
  int numchannels=0;
  int nwires=0;
  char hwc[12];
  
  
  TestId sl ={0,0,0xFFFFFFFF,0,"SuperLayer"};
  TestId l  ={0,0,0xFFFFFFFF,0,"Layer"};
  TestId sec={0,0,0xFFFFFFFF,0,"Sector"};
  TestId mod={0,0,0xFFFFFFFF,0,"Module"};
  TestId ml ={0,0,0xFFFFFFFF,0,"MonoLayer"};
  TestId w  ={0,0,0xFFFFFFFF,0,"Wire"};
  TestId asd={0,0,0xFFFFFFFF,0,"Asd"};
  
  FILE *fp;
  printf("start debug decode\n");fflush(stdout);

  fp=fopen("OtrDecode.notused","w");
  if (!fp) {
    printf("Cannot open file:  OtrDecode.notused\n");
    exit(1);
    }
  printf("debug decode: %d %d\n",OtrITdcNumberOf(),OtrITdcNumberOfChannel());fflush(stdout);
  for (i=0;i<OtrITdcNumberOf();i++) {
    for (j=0;j<OtrITdcNumberOfChannel();j++) {
      int ci=OtrIFromTdcIndex(i,j);      
      unsigned short tdcid;
      int tdcch,c1,chid;
      if (OtrIError()) {
        unsigned short tdcid=OtrITdcIdFromIndex(i);

        fprintf(fp,"Invalid Channel: crate: %d  slot: %d tdcid: %04X tdc-connector: %d asd-channel: %2d\n",
	         OtrITdcCrateByTdcId(tdcid),OtrITdcSlotByTdcId(tdcid),tdcid,j/16,j%16);

        notusedtdc++;
	continue;
        }
      tdcid=OtrITdcId(ci);
      tdcch=OtrITdcChannel(ci);
      c1=OtrIFromTdc(tdcid,tdcch);
      
     
      if (c1!=ci) {
        printf("Channel inconsistent i=%d j=%d  f1 %04X c0=%08X c1=%08X\n",
	                 i,j,tdcid,ci,c1);
//	getchar();
	}
      sl.val=OtrISuperLayer(ci);
      l.val=OtrILayer(ci);
      sec.val=OtrISector(ci);
      ml.val=OtrIMonoLayer(ci);
      w.val=OtrIWire(ci);
      mod.val=OtrIModule(ci);
      asd.val=OtrIAsd(ci);
      
      chid=OtrIdFromIndex(ci);
      sl.cmp=OtrISuperLayerFromId(chid);
      l.cmp=OtrILayerFromId(chid);
      sec.cmp=OtrISectorFromId(chid);
      ml.cmp=OtrIMonoLayerFromId(chid);
      w.cmp=OtrIWireFromId(chid);
      asd.cmp=OtrIAsdFromId(chid);
      mod.cmp=OtrIModuleFromId(chid);
     
      err=0;
      CheckId(&sl);
      CheckId(&l);
      CheckId(&sec);
      CheckId(&mod);
      CheckId(&ml);
      CheckId(&w);
      CheckId(&asd);
      if (err) {
        printf("sl %d %d l %d %d s %d %d m %d %d w %d %d a %d %d\n",
	   sl.val,sl.cmp,l.val,l.cmp,sec.val,sec.cmp,ml.val,ml.cmp,w.val,w.cmp,asd.val,asd.cmp);
//        getchar();
	}
      c1=OtrIFromId(chid); 
      if (ci!=c1) {
        printf("channel id not ok %08X %08X\n",ci,c1);
        printf("sl %d l %d  s %d mod %d ml %d w %d a %d\n",
	   sl.val,l.val,sec.val,mod.val,ml.val,w.val,asd.val);
//	getchar();
        }
	
      {
        char sln[4];  
	int s,ll,m,se,fl;
        OtrIGeoId(ci,hwc);
	sscanf(hwc,"%3s%*c%d%*c%d%*c%d",&sln[0],&ll,&m,&se);
	if (OtrISide(ci)>0) sln[0]='+';
	else sln[0]='-';
	
	s=OtrISuperLayerFromName(sln);
	sl.cmp=s;
	l.cmp=ll;
	mod.cmp=m;
	sec.cmp=se;

        CheckId(&sl);
        CheckId(&l);
	/* sector check cannot be passed by wire interconnecting
	   modules, since the ASD-sector is different from the module sector*/

	fl=1;
	if (!strcmp(&sln[0],"-p1")||!strcmp(&sln[0],"-p4")) {
	  if (ll>=4) {
	    if (OtrIAsdChannel(ci)==0xC||OtrIAsdChannel(ci)==0xE) fl=0;
	    }
	  }
	else if (!strcmp(&sln[0],"+p1")||!strcmp(&sln[0],"+p4")) {
	  if (ll>=4) { 
	    if (OtrIAsdChannel(ci)==0x1||OtrIAsdChannel(ci)==0x3) fl=0;
	    }
	  }
	else if (!strcmp(&sln[0],"-t1")||!strcmp(&sln[0],"-t2")) {
	  if (OtrIAsdChannel(ci)==0xC||OtrIAsdChannel(ci)==0xE) fl=0;
	  }
	else if (!strcmp(&sln[0],"+t1")||!strcmp(&sln[0],"+t2")) {
	  if (OtrIAsdChannel(ci)==0x1||OtrIAsdChannel(ci)==0x3) fl=0;
	  }
	  
	  
        if (fl) CheckId(&sec);
        CheckId(&mod);
	
        if (err) {
	  printf("sn %s hwc %s sl %s s %d l %d m %d sec %d ci:%08X\n",
	     OtrISuperLayerName(ci),hwc,sln,s,ll,m,se,ci);
	  getchar();
	  }
	}
   
      numchannels++;
      }
    }
  fclose(fp);
   
  nwires=0; 
  for (i=0;i<OtrGeoNumberOfVolumes();i++) {
    for (j=0;j<OtrIMonoLayerNumberOf(i);j++) {
      nwires+=OtrIMonoLayerNumberOfWires(i,j);
      }
    }
    
  printf("range sl  %d %d\n",sl.min,sl.max);
  printf("range lay %d %d\n",l.min,l.max);
  printf("range sec %d %d\n",sec.min,sec.max);
  printf("range mod %d %d\n",mod.min,mod.max);
  printf("range asd %d %d\n",asd.min,asd.max);
  printf("range ml  %d %d\n",ml.min,ml.max);
  printf("range w   %d %d\n",w.min,w.max);
  if (notusedtdc!=6736) {
    printf("OtrDecode: Number of not used tdc-channel is not correct: %d instead of 6736\n",
             notusedtdc);
    }
  printf("number of channels: %d\n",numchannels);
  printf("number of wires:    %d\n",nwires);


  {
    int sl,l,mod,sec;
    
    printf("geo loop test\n");fflush(stdout);
    for (sl=0;sl<32;sl++) {
      if (OtrDecodeGet()->SLname[sl]) {
        printf("superlayer name: %d %s %d\n",sl,OtrDecodeGet()->SLname[sl],
	OtrILayerNumberOf(sl));fflush(stdout);
	}
      }
      
    for (sl=0;sl<OtrISuperLayerNumberOf();sl++) {
      if (!OtrISuperLayerExist(sl)) continue;
      for (l=0;l<OtrILayerNumberOf(sl);l++) {
        if (!OtrILayerExist(sl,l)) continue;
        for (mod=0;mod<OtrIModuleNumberOf(sl,l);mod++) {
          if (!OtrIModuleExist(sl,l,mod)) {
	    continue;
	    }
          for (sec=0;sec<OtrISectorNumberOf();sec++) {
	    int ci,i,j;
	    char id[12],id1[12],id2[13];
	    float x[3],n[3];
            int gi=OtrIGeoIndexFromModule(sl,l,mod,sec); 
	    if (OtrIError()) continue;
            OtrIGeoIdFromModule(sl,l,mod,sec,id);
            OtrIGeoIdFromIndex(gi,id1);
	    memcpy(id2,OtrGeoGet(gi)->hwc,12);
	    id2[12]='\0';
	    for (i=0;i<OtrIMonoLayerNumberOf(gi);i++) {
	      for (j=0;j<OtrIMonoLayerNumberOfWires(gi,i);j++) {
  	        ci=OtrIFromModule(sl,l,mod,sec,i,j);
		if (!OtrIError()) break;
		}
              }
	    if (OtrIError()) {
	      printf("module has no wires: %s \n",id2);
	    printf("layer %d %d %d %d %d\n",OtrIMonoLayerNumberOf(gi),
	      OtrIMonoLayerNumberOfWires(gi,0),
	      OtrIMonoLayerNumberOfWires(gi,1),
	      OtrIMonoLayerNumberOfWires(gi,2),
	      OtrIMonoLayerNumberOfWires(gi,3));
	      continue;
	      }
	    OtrIWireMiddleNominal(ci,x);
	    OtrIWireDirectionNominal(ci,n);
	  
            printf("GeoTest: %s %s %s %f %f %f %f %f %f %f %f\n",
              id,id1,id2,OtrIWireStereoAngle(ci),OtrIWirePitch(ci),
	      x[0],x[1],x[2],n[0],n[1],n[2]);
	    fflush(stdout);
	    }
	  }
        }
      }
    }
  printf(" done\n");fflush(stdout);
  return OtrIValid();
  }
  
  

int OtrIFromWax(char *hwc,int waxid) {
  int sl,lay,sec,mod,mono,wire;

// In Wax hitb->chid = wireId + 128*(chId + 16*asdId), where wireId = wire+32*mono
  if (waxid>15000000 || waxid<0) {
    printf("OtrIdFromWax: waxid = %x is illegal\n",waxid);
    }

  sl = OtrISuperLayerFromName(hwc);
  sscanf(hwc,"%*c%*c%*c%*c%d%*c%d%*c%d",&lay,&mod,&sec);
  mono = ((waxid)%128) /32;
  wire = (waxid) %32;
  
  OtrIValid();
  int ci = OtrIFromModule(sl,lay,mod,sec,mono,wire);
  if (OtrIError()) {
    char s[15];
    memcpy(s,hwc,12);
    s[12]='\0';
    printf("OtrIFromWax: wax channel not existing in decoding %s %d %d %d %d %d %d %8x\n",
                 s,sl,lay,mod,sec,mono,wire,waxid);getchar();
    }
  return ci;
  }


int OtrITestMe() {
  OtrGeoTestMe();
  OtrDecodeTest();
  return 0;
  }
  


