#ifndef __OtrMon__H__
#define __OtrMon__H__

#include "OtrI.h"
#include "dq.h"
#include "etb.h"

/* return values  */ 
#define OTRMONValid 0
#define OTRMONInvalid -1

/* define debug level */
#define OTRMONDebug_0  0
#define OTRMONDebug_1  1
#define OTRMONMsgInfo  0
#define OTRMONMsgWarn  1
#define OTRMONMsgError 2



/* to steer the algorithems and the  initilization */
/* switch: bit 0 -  7 database/calculation
               8 - 15 active DQ algo's
              16 - 23 status word eval: algo's in logic Or
              24 - 32 status word eval: algo's in logic And
	      */


#define OTRMONAlgoDefault 0xff 
#define OTRMONAlgoFed      0x1
#define OTRMONAlgoStatus   0x2
#define OTRMONAlgoOccup    0x4
#define OTRMONAlgoSynchr   0x8
#define OTRMONAlgoTime    0x10

#define OTRMONAlgoDefaultDbug (0xff<<8) 
#define OTRMONAlgoFedDbug      (0x1<<8)
#define OTRMONAlgoStatusDbug   (0x2<<8)
#define OTRMONAlgoOccupDbug    (0x4<<8)
#define OTRMONAlgoSynchrDbug   (0x8<<8)
#define OTRMONAlgoTimeDbug    (0x10<<8)

#define OTRMONNoInit      0x0

#define OTRMONDbaseRead   (0x1<<16) 
#define OTRMONDbaseUpdate (0x2<<16) 
#define OTRMONAsciiRead   (0x4<<16)
#define OTRMONAsciiWrite  (0x8<<16)
#define OTRMONRhpRead    (0x10<<16)
#define OTRMONRhpWrite   (0x20<<16)
#define OTRMONOnline     (1<<30)
#define OTRMONDebug      (1<<31)

#define OTRMONAllSl 0

#define OTRMONGood    0
#define OTRMONUsable  1
#define OTRMONBad     2
#define OTRMONUnknown 4

#define OTRMONAllLayer 0
#define OTRMONHistSum  0
#define OTRMONHistBins 32

#ifdef __cplusplus
extern "C" {
#endif
void  OtrMonSwitch(int i);
int   OtrMonInitialize();
int   OtrMonUpdate();
int   OtrMonWriteToDb();
int   OtrMonClose();
int   OtrMonScan(FILE *fp);
int   OtrMonPrint(FILE *fp);
int   OtrMonTestMe();
int   OtrMonHelp(int,const char **);
int   OtrMonBegin();
int   OtrMonEnd(); 
int   OtrMonEvent();
int   OtrMonHalt();
#ifdef __cplusplus
}
#endif

/*
  HISTOGRAM ID (32 bits)
  
    4 bits              4 bits               4 bits              20 bits
    free (maybe          component-id         reconstruction       free for
    classification       (as defined before)  level (as defined    user
    of DQ/Alignment/...                       before) 

    20 user bits:  1 ....  7  Histo ID   (Max ID = 127 (Decimal))
                   8 .... 12  SLayer id  (SL ID = 0 -> general Histograms)
                  13 .... 20  Algorithm 
		  */

inline int OtrMonHistId(int level, int algo, int slid, int hid) { 
  return (COMP_OTR<<24)+(level<<20)+(algo<<12)+(slid<<7)+hid; 
}

#endif



