/*--------------------------------------------------------------------------*
 *                                                                          *
 *    dmon-user  user routines for the data monitor program.                *
 *                                                                          *
 *    dmonuser_help     user's program help                                 *
 *    dmonuser_init     program initialization                              *
 *    dmonuser_begin    begin-run processing                                *
 *    dmonuser_event    event processing                                    *
 *    dmonuser_loop     user's loop                                         *
 *    dmonuser_end      end-run processing                                  *
 *    dmonuser_halt     program termination                                 *
 *                                                                          *
 *--------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <math.h>
#include <string.h>
#include "component.h"
#include "grr.h"
#include "etb.h"
#include "dmon.h"
#include "par.h"
#include "smc.h"
#include "rhp.h"
#include "Otr/OtrMon.h"
#include "Otr/OtrDq.h" 
#include "Otr/OtrDqAlgo.h"
#include "Otr/OtrEvent.h"

static int debug=0;
static int events=0;

int    majort0=0;
int    majordq=0;

 
/* ----- some HBOOK stuff ------ */
#ifdef __cplusplus
extern "C" {
#endif

#ifdef __cplusplus
}
#endif

/*  <--------------------------  */
/*--------------------------------------------------------------------------*
 *                                                                          *
 * dmonuser_help()  user's help invoked if '-h' online parameter is present *
 *                                                                          *
 *--------------------------------------------------------------------------*/
int dmonuser_help(int argc , const char **argv){
  OtrMonHelp(argc, argv);
  return GRR_OK;
}


/*--------------------------------------------------------------------------*
 *                                                                          *
 *     dmonuser_begin()  monitor program begin run routine                  *
 *                                                                          *
 *--------------------------------------------------------------------------*/
int dmonuser_begin(){
  OtrMonBegin();
  return GRR_OK;
}

/*--------------------------------------------------------------------------*
 *                                                                          *
 *     dmonuser_end()    monitor program end run routine                    *
 *                                                                          *
 *--------------------------------------------------------------------------*/
int dmonuser_end(){

  OtrMonEnd();
  
  if (events>5000) {
    OtrDbSetKey(0,majordq,0);  
    OtrDqAlgoUpdate();
    OtrDqWriteToDb();
  }

  if (debug) {
  }
  if (debug) {
    char name[1024];
    int ci, asd, cha, dead=0, noisy=0;
    
    for (asd=0; asd<OtrIAsdNumberOf(); asd++){
      for (cha=0; cha<OtrIAsdNumberOfChannel(); cha++) {
	ci = OtrIFromAsd(asd, cha);
	if (OtrDqDead(ci)) dead++; 
	else if (OtrDqBad(ci)) noisy++; }}
    printf("\n DATA QUALITY> Noisy/Dead Channels = %d %d \n\n", noisy, dead);
 
    sprintf(name,"OtrDq%d.dat",OtrEventInfoRunNumber());
    OtrDqAsciiDef(OTRDQAsciiWrite,name);
    OtrDqWriteToDb();
    sprintf(name,"OtrDq%d.stat",OtrEventInfoRunNumber());
    OtrDqAsciiDef(OTRDQDbaseUpdate,name);
    OtrDqWriteToDb();
    }
    
  if (events>5000&&events<50000) {
    OtrT0Calculate();
    OtrDbSetKey(0,majort0,0);  
    OtrT0WriteToDb();
    }
  
  return GRR_OK;
}

/*--------------------------------------------------------------------------*
 *                                                                          *
 *     dmonuser_event()   monitor program event processing routine          *
 *                                                                          *
 *--------------------------------------------------------------------------*/
int dmonuser_event(EVENT_HDR *event, int length){
  
  rhp_wait_for_one_msg_and_dispatch(NO_BLOCK);
  if (length>0) {

    OtrEventUpdate(event);
   
    if (debug&&events%100==0) printf("events: %d\n",events);
    events++;    
    if (events==1) {
      majordq=majort0=OtrEventInfoTime();
      }  
    
    OtrDqAlgoEvent();
   /*   
    if ((rc=OtrDqAlgoEvent())==OTRDQAlgoUpdate) {
      OtrDbSetKey(0,majordq,0);  
      OtrDqUpdate();
      OtrDqWriteToDb();
      majordq=OtrEventInfoTime();
      }
    if (OtrIError()) printf("Error in Dq %d\n",rc); 
*/
    OtrMonEvent();

    if (events<50000) {
      int i;
      for(i=0;i<OtrEventHitNumberOf();i++){  // for all hits
        int ci=OtrEventHitChannelIndex(i);
        OtrT0Event(ci,OtrEventHitTdcValue(i));
        }
      }
    else if (events==50000) {
      OtrT0Calculate();
      OtrDbSetKey(0,majort0,0);
      OtrT0WriteToDb();
      }
      
    if (debug==2) {
      if (events%10==0) {
        printf("dmonuser_event> Processed %d Events: ", events);
        printf(" IA = %f", OtrEventInfoIaRate());
        printf(" PhysBx = %d", OtrEventInfoPhysBx());
        printf(" Run = %d\n", OtrEventInfoRunNumb());
	}
      }
    }
  return GRR_OK;
}

/*--------------------------------------------------------------------------*
 *                                                                          *
 *     dmonuser_loop()   monitor program idle loop routine                  *
 *                                                                          *
 *--------------------------------------------------------------------------*/
int dmonuser_loop(){
 return GRR_OK;
}


/*--------------------------------------------------------------------------*
 *                                                                          *
 *     dmonuser_halt()   terminate a monitor program                        *
 *                                                                          *
 *--------------------------------------------------------------------------*/
int dmonuser_halt(){
  OtrMonHalt();
  return GRR_OK;
}

/*--------------------------------------------------------------------------*
 *                                                                          *
 *     dmonuser_init()   initialize a monitor program                       *
 *                                                                          *
 *--------------------------------------------------------------------------*/
int dmonuser_init(int argc, const char **argv){
  int i, flag=0,test=1,db=0;
  int ci, asd, cha, dead=0, noisy=0;

  /*-----Set enviroment to open DB files --------------------*/
  /*------------------------------------------------------------*/
  
  for(i=1;i<argc;i++){
    if(      strcmp(argv[i],"-debug") == 0)  debug=2;
    else if (!strncmp(argv[i],"-debug:",strlen("-debug:"))) {
      sscanf(argv[i]+strlen("-debug:"),"%x",&debug);
      }
    else if (strcmp(argv[i],"-online") == 0) flag=1;
    else if (strcmp(argv[i],"-notest") == 0) test=0;
    else if (strcmp(argv[i],"-db") == 0) db=1;
  }
  printf("debug: %d %d\n",debug,debug&0x2);
  /* rhp_init(RHP_FLG_RESET);
     rhps_init();*/
  /* pub_sinit();
     pub_sbind(); */
  
  rhp_msinitf(RHP_FLG_RESET);
  rhp_msbind();
   
  if (!test) {
    putenv("OtrDbServerMode=notest");
    OtrDbSwitch(0);
    }
  if (db) {
    OtrGeoDefine();
    OtrDecodeSwitch(SHARCTAB|CABTAB);
    OtrDecodeDefine();
    }
  else {
    OtrDecodeSwitch(SHARCTABFILE|CABTABFILE);
    }
 
  putenv("OtrGeoInput=/afs/desy.de/group/hera-b/OTR/pro/Tables/OtrGeo/nede.tab");

  OtrGeoInitialize();
  OtrDecodeInitialize();

  OtrEventSwitch(OTREVENTValidTdc|
                 OTREVENTDoStatistics|
		 OTREVENTNoHitPreparation);  

  if (debug&0x2) OtrDqSwitchDebug(1);
  
  OtrDqSwitch(OTRDQCalculate|OTRDQAlgoOccup|OTRDQAlgoLhood|OTRDQAlgoRms|
	      OTRDQSelectOccupOr);
  OtrDqInitialize();
  OtrDqAlgoInit(); 
  if (OtrIError()) {
    printf("OtrDqAlgoInit() not correctly initialized !. Abort \n");
    exit(1);
    }
  OtrDqDefine();

  if (debug) {
    for (asd=0; asd<OtrIAsdNumberOf(); asd++){
      for (cha=0; cha<OtrIAsdNumberOfChannel(); cha++) {
	ci = OtrIFromAsd(asd, cha);
	if (OtrDqDead(ci)) dead++; 
	else if (OtrDqBad(ci)) noisy++; 
      }
    } 
    printf("\n DATA QUALITY> Noisy/Dead Channels = %d %d \n\n", noisy, dead);
  }
   
  OtrT0Switch(OTRT0Calculate);
  OtrT0Initialize();
  OtrT0Define();
    
  OtrMonSwitch((OTRMONDebug)|(flag*OTRMONOnline)| 
	       OTRMONAlgoFed|OTRMONAlgoStatus|OTRMONAlgoOccup|OTRMONAlgoSynchr);
  OtrMonInitialize();

  return GRR_OK;
}



























