#include "Otr/OtrRt.h"
#include "Otr/OtrI.h"
#include "Otr/OtrDb.h"


#define NUM_RT (2*16)

/* --------------------------------------------------------------------*/
typedef struct {
  int   sl;
  int   rt_sl;
  } RtMapInfo;
 
OtrDbField RtMapFmt[]={
  {"int"          ,"sl"        ,OtrDbOffset(RtMapInfo,sl)},
  {"int"          ,"rtsl"      ,OtrDbOffset(RtMapInfo,rt_sl)}
  };

OtrDbTable RtTableMap={"OTR_RtMap",1,"RtMap",sizeof(RtMapInfo),
                         RtMapFmt,sizeof(RtMapFmt)/sizeof(OtrDbField)};     


/* --------------------------------------------------------------------*/
typedef struct {
  int   sl;
  int   ptype;
  int   i0;
  int   numt;
  float tmin,tmax,t0;
  } RtTableInfo;
 
OtrDbField RtHeader[]={
  {"int"          ,"sl"         ,OtrDbOffset(RtTableInfo,sl)},
  {"int"          ,"ptype"      ,OtrDbOffset(RtTableInfo,ptype)},
  {"int"          ,"i0"         ,OtrDbOffset(RtTableInfo,i0)},
  {"int"          ,"numt"       ,OtrDbOffset(RtTableInfo,numt)},
  {"float"        ,"tmin"       ,OtrDbOffset(RtTableInfo,tmin)},
  {"float"        ,"tmax"       ,OtrDbOffset(RtTableInfo,tmax)}
  };

OtrDbTable RtTableHeader={"OTR_RtHeader",1,"RtHeader",sizeof(RtTableInfo),
                         RtHeader,sizeof(RtHeader)/sizeof(OtrDbField)};     


/* --------------------------------------------------------------------*/

typedef struct {
  float val,sig;
  } RtDataPoint;
  
OtrDbField RtField[]={
  {"float"        ,"r"     ,OtrDbOffset(RtDataPoint,val)},
  {"float"        ,"rsig"  ,OtrDbOffset(RtDataPoint,sig)}
  };

OtrDbTable RtTableData={"OTR_RtData",1,"RtData",sizeof(RtDataPoint),
                         RtField,sizeof(RtField)/sizeof(OtrDbField)};     


RtMapInfo*   mapinfo=0;
int          nummap=0;
RtTableInfo* tableinfo=0;
int          numtable=0;
RtDataPoint* rtdata=0;
int          numdata=0;

/* --------------------------------------------------------------------*/

void OtrRtDefine() {
  OtrDbDefine("/OTR_STATUS",RtTableMap.name,OtrRtUpdate);
  OtrDbDefine("/OTR_STATUS",RtTableHeader.name,OtrRtUpdate);
  OtrDbDefine("/OTR_STATUS",RtTableData.name,OtrRtUpdate);
  }

/* --------------------------------------------------------------------*/

class RtRelation  {
  RtTableInfo* hdr;
  RtDataPoint* r;
 
  float  tbin;

  float *t,*drdt;
  

public:

  void Set(int sl,int ptype, float t0,float tmin,float tmax,int numt, 
           float *r,float* rsig);  
  

  inline RtRelation() {
    tbin=0;
    drdt=t=0;
    r=0;
    hdr=0;
    }

  RtRelation(RtTableInfo *);

  inline ~RtRelation() {
    if (hdr->i0==-1) {
      delete r;
      delete hdr;
      }
    if (t) delete t;
    if (drdt) delete drdt;
    }

  inline int PitchType()  { return (hdr) ? hdr->ptype : 0; }
  inline int SuperLayer() { return (hdr) ? hdr->sl : 0; }
  void Scan(FILE *fp);
  void Print(FILE *fp);
  
  float Time(float r) const;

  inline float Radius(float tt) const {
    float ret=0;
    if (hdr) {
      if (tt<=hdr->tmin) ret=r[0].val;
      else if (tt>=hdr->tmax) ret=r[hdr->numt-1].val;
      else {
        int it=(int)((tt-hdr->tmin)/tbin);
        ret=r[it].val+(tt-t[it])*drdt[it];
        }
      }
    return ret;
    }
    
  inline float RadiusSigma(float tt) const {
    float ret=0;
    if (hdr) {
      int it=(int)((tt-hdr->tmin)/tbin);
      if (tt<hdr->tmin) ret=r[0].sig;
      else if (tt>=hdr->tmax) ret=r[hdr->numt-1].sig;
      else ret=r[it].sig;
      }
    return ret;
    }

  };


/* -------------------------------------------------------------------- */

RtRelation::RtRelation(RtTableInfo* ti) { 
  int i;
  hdr=ti;
  tbin=(hdr->tmax-hdr->tmin)/(float)(hdr->numt-1);

  r=&rtdata[hdr->i0];
  t   =new float[hdr->numt];
  drdt=new float[hdr->numt];
  
  for (i=0;i<hdr->numt;i++) t[i]=hdr->tmin+i*tbin;
  for (i=0;i<hdr->numt-1;i++) drdt[i]=(r[i+1].val-r[i].val)/(t[i+1]-t[i]);
  drdt[hdr->numt-1]=0;
  }


/* -------------------------------------------------------------------- */

void RtRelation::Set(int ssl,int pptype,float t0,
                      float ttmin,float ttmax,int numtt, 
                      float *rr,float* rrsig) {
  if (t)    delete t;
  if (drdt) delete drdt;
    
  t   =new float[numtt];
  drdt=new float[numtt];
  
  hdr=(RtTableInfo*)malloc(sizeof(RtTableInfo));
  r=(RtDataPoint*)malloc(numtt*sizeof(RtDataPoint));

  hdr->tmin=ttmin;
  hdr->tmax=ttmax;
  hdr->i0  =-1;
  hdr->numt=numtt;
  hdr->t0=t0;
  hdr->sl  =ssl;
  hdr->ptype=pptype;
  tbin=(hdr->tmax-hdr->tmin)/(hdr->numt-1);
  
  int i;
  for (i=0;i<hdr->numt;i++) {
    t[i]=hdr->tmin+i*tbin;
    r[i].val=rr[i];
    r[i].sig=rrsig[i];
    }
  for (i=0;i<hdr->numt-1;i++) drdt[i]=(r[i+1].val-r[i].val)/(t[i+1]-t[i]);
  drdt[hdr->numt-1]=0;

  }


void RtRelation::Print(FILE *fp) {
  int i;
  fprintf(fp,"RtRelation\n");
  fprintf(fp,"tmin %f tmax %f numt %d  type %d %d t0 %f\n",
            hdr->tmin,hdr->tmax,hdr->numt,hdr->sl,(hdr->ptype==0)?5:10,
	    hdr->t0);
	    
  for (i=0;i<hdr->numt;i++) {
    fprintf(fp,"%12e %12e %12e %12e\n",t[i],r[i].val,r[i].sig,drdt[i]);
    }
  fprintf(fp,"\n");
  }


float RtRelation::Time(float rr) const {
  int it=0;
  if (rr<r[0].val) it=0;
  else if (rr>r[hdr->numt-1].val) it=hdr->numt-1;
  else {
    for (it=0;it<hdr->numt-1;it++) {
      if (r[it].val<rr&&rr<r[it+1].val) break;
      }
    }
  return t[it]+(rr-r[it].val)/drdt[it];
  }



/* -------------------------------------------------------------------- */

int    numrt=NUM_RT;
static RtRelation **rt=0;



void OtrRtSwitch(int i) {
  i=0;
  }


RtRelation* OtrRtGet(int ci) {
  int p   =(OtrIWirePitch(ci)>1);
  int sl  =OtrISuperLayer(ci)&0xF;
  return rt[sl*2+p];
  }

float OtrRtRadius(int ci,float t) {
  return (rtdata) ? OtrRtGet(ci)->Radius(t) : 0;
  }


float OtrRtRadiusSigma(int ci,float t) {
  return (rtdata) ? OtrRtGet(ci)->RadiusSigma(t) :0;
  }


float OtrRtTime(int ci,float r) {
  return (rtdata) ? OtrRtGet(ci)->Time(r) : 0;
  }


/* -------------------------------------------------------------------- */


int OtrRtInitialize() {
  if (rtdata!=0) return OtrIValid();

  OtrRtUpdate();
  return 0;
  }

void OtrFreeRt() {
  int i,n;
  RtRelation** rrt=new RtRelation*[numrt];
  for (i=0;i<numrt;i++) rrt[i]=0;
  n=0;
  for (i=0;i<numrt;i++) { 
    int k,fl=0;
    for (k=0;k<n;k++) {
      if (rrt[k]==rt[i]) {
        fl=1;
	break;
        }
      }
    if (!fl&&rt[i]) {
      delete rt[i];
      rrt[n++]=rt[i];
      }
    rt[i]=0;
    } 
  delete rrt;
  }

int OtrRtClose() {
  
  OtrFreeRt();
  
  if (rt) delete rt;
  rt=0;
  
  if (tableinfo) free(tableinfo);
  if (rtdata) free(rtdata);
  if (mapinfo) free(mapinfo);

  rtdata=0;
  tableinfo=0;
  mapinfo=0;
  numdata=0;
  numtable=0;
  nummap=0;
  return 0;
  }

/* --------------------------------------------------------------------*/

int OtrRtWriteToDb() {
  OtrDbPut(&RtTableMap,mapinfo,nummap);
  OtrDbPut(&RtTableHeader,tableinfo,numtable);
  OtrDbPut(&RtTableData,rtdata,numdata);
  return OtrIValid();
  }

/* --------------------------------------------------------------------*/

int OtrRtUpdate() {
  char* fname=getenv("OtrRtInput");
  int version;

  if (!rt) {
    rt=new RtRelation*[numrt];
    memset(rt,0,sizeof(RtRelation*)*numrt);
    }
  else OtrFreeRt();

  if (fname&&fname[0]!='\0') {
    int ret;
      
    FILE *fp=fopen(fname,"r");
    printf("OtrRtUpdate: from file %s\n",fname);
    if (!fp) {
      return OtrIErrorMsg(OTRIInvalid,"OtrRtUpdate",
                            "Cannot open file: %s",fname);
      }
    OtrRtScan(fp);
    return ret;
    }
  else if ((version=OtrDbGetTableVersion(RtTableHeader.name))!=-1) {
    int i;
    version=OtrDbGetTableVersion(RtTableData.name);
    version=OtrDbGetTableVersion(RtTableMap.name);
    
    tableinfo=(RtTableInfo*)OtrDbGet(&RtTableHeader,tableinfo,&numtable);
    rtdata=(RtDataPoint*)OtrDbGet(&RtTableData,rtdata,&numdata);
    mapinfo=(RtMapInfo*)OtrDbGet(&RtTableMap,mapinfo,&nummap);
    
    for (i=0;i<numtable;i++) {
      int sl=tableinfo[i].sl;
      int p=tableinfo[i].ptype;
      int j=sl*2+p;
      rt[j]=new RtRelation(&tableinfo[i]);
      }
      
    for (i=0;i<nummap;i++) {
      int sl =mapinfo[i].sl*2;
      int rsl=mapinfo[i].rt_sl*2;
      rt[sl+0]=rt[rsl+0];
      rt[sl+1]=rt[rsl+1];
      }
    }
  return 0;
  };


/* --------------------------------------------------------------------*/


int OtrRtScan(FILE *fp) {
  int i;
  char line[256],tag[256];
      
  OtrIValid();
  while (fgets(line,256,fp)) {
    tag[0]='\0';
    sscanf(line,"%s",tag);
    if (!strcmp(tag,"RtRelation")) {
      RtTableInfo* ti;
      RtDataPoint* d;
      
      numtable+=1;
      tableinfo=(RtTableInfo*)realloc(tableinfo,numtable*sizeof(RtTableInfo));
      ti=&tableinfo[numtable-1];

      while (fgets(line,256,fp)) {
        if (line[0]=='#') continue;
	ti->t0=0;
        sscanf(line,"%*s %f %*s %f %*s %d %*s %d %d %*s %f",
	            &ti->tmin,&ti->tmax,&ti->numt,&ti->sl,&ti->ptype,&ti->t0);
        if (ti->ptype==5) ti->ptype=0;
        else ti->ptype=1;
        break;
        }

      ti->i0=numdata;
      numdata+=ti->numt;
      rtdata=(RtDataPoint*)realloc(rtdata,numdata*sizeof(RtDataPoint));
      d=&rtdata[ti->i0];
      for (i=0;i<ti->numt;i++) {
        float t,drdt;
        if (fscanf(fp,"%e %e %e %e",&t,&d[i].val,&d[i].sig,&drdt)!=4) {
          OtrIErrorMsg(OTRIInvalid,"Otr: RtRelation::scan",
                            "Wrong Format","");
          }
        }  
      }
    else if (!strcmp(tag,"RtMap")) {
      int i;
      fscanf(fp,"%*s %d",&nummap);
      mapinfo=(RtMapInfo*)malloc(nummap*sizeof(RtMapInfo));
      for (i=0;i<nummap;i++) {
        fscanf(fp,"%*s %d %*s %d\n",&mapinfo[i].sl,&mapinfo[i].rt_sl);
        }
      }
    }
  if (OtrIError()) return OTRIInvalid;

  
  for (i=0;i<numtable;i++) {
    RtRelation* rrt=new RtRelation(&tableinfo[i]);
    int sl=rrt->SuperLayer();
    int p=rrt->PitchType();
    rt[sl*2+p]=rrt;
    }
    
  for (i=0;i<nummap;i++) {
    int sl =mapinfo[i].sl*2;
    int rsl=mapinfo[i].rt_sl*2;
    rt[sl+0]=rt[rsl+0];
    rt[sl+1]=rt[rsl+1];
    }
  return OtrIValid();
  }
  
  
  
int OtrRtPrint(FILE *fp) {
  int i,n;
  RtRelation** rrt=new RtRelation*[numrt];
  for (i=0;i<numrt;i++) rrt[i]=0;
  n=0;
  for (i=0;i<numrt;i++) { 
    int k,fl=0;
    for (k=0;k<n;k++) {
      if (rrt[k]==rt[i]) {
        fl=1;
	break;
        }
      }
    if (!fl&&rt[i]) {
      rt[i]->Print(fp);
      rrt[n++]=rt[i];
      }
    } 

  fprintf(fp,"RtMap\n");
  n=0;
  for (i=0;i<numrt;i+=2) {
    if (rt[i]!=0) n++;
    }
  fprintf(fp,"numsl: %d\n",n);
  for (i=0;i<numrt;i+=2) {    
    if (rt[i]) fprintf(fp,"sl: %2d  rt: %2d\n",i/2,rt[i]->SuperLayer());
    }
  delete rrt;
  return OtrIValid();
  }
  
  
  
void OtrRtSet(int sl,int cell,float t0,float tmin,float tmax,int numt, 
                     float *r,float* rsig) {
  int i,p;
  if (!rt) {
    numrt=16*2;
    rt=new RtRelation*[numrt];
    memset(rt,0,sizeof(RtRelation*)*numrt);
    }
  if (cell>3) p=(cell==5)?0:1;
  else p=(cell==0)?0:1;
  i=sl*2+p;
  rt[i]=new RtRelation;
  rt[i]->Set(sl,p,t0,tmin,tmax,numt,r,rsig);
  }



void OtrRtSetEqual(int new_sl,int existing_sl) {
  rt[new_sl*2]=rt[existing_sl*2];
  rt[new_sl*2+1]=rt[existing_sl*2+1];
  }
  
  
int OtrRtTestMe() {
  float tmin=0;
  float tmax=256;
  int   numt=10;
  float r[10],rsig[10];
  
  r[9] = 0.0;   rsig[9]= 0.01;  
  r[8] = 1.0;   rsig[8]= 0.02;  
  r[7] = 1.5;   rsig[7]= 0.03;  
  r[6] = 2.0;   rsig[6]= 0.04;  
  r[5] = 3.0;   rsig[5]= 0.02;  
  r[4] = 4.5;   rsig[4]= 0.03;  
  r[3] = 5.3;   rsig[3]= 0.04;  
  r[2] = 6.2;   rsig[2]= 0.02;  
  r[1] = 8.2;   rsig[1]= 0.03;  
  r[0] = 9.0;   rsig[0]= 0.01; 
  
  RtRelation rt1;rt1.Set(5,0,0,tmin,tmax,numt,r,rsig);
  RtRelation rt2;rt2.Set(5,1,0,tmin,tmax,numt,r,rsig); 
  RtRelation rt3;rt3.Set(14,0,0,tmin,tmax,numt,r,rsig);
  RtRelation rt4;rt4.Set(14,1,0,tmin,tmax,numt,r,rsig);
  
  FILE *fp=fopen("rt.in","w");
  if (!fp) {
    printf("OtrRtTestMe: Cannot open file rt.in.\n");
    exit(1);
    }
  rt1.Print(fp);  
  rt2.Print(fp);  
  rt3.Print(fp);  
  rt4.Print(fp);  
  fprintf(fp,"RtMap\n");
  fprintf(fp,"numsl: 13\n");
  fprintf(fp,"sl:  1  rt:  5\n");
  fprintf(fp,"sl:  2  rt:  5\n");
  fprintf(fp,"sl:  3  rt:  5\n");
  fprintf(fp,"sl:  4  rt:  5\n");
  fprintf(fp,"sl:  5  rt:  5\n");
  fprintf(fp,"sl:  6  rt:  5\n");
  fprintf(fp,"sl:  8  rt:  5\n");
  fprintf(fp,"sl:  9  rt: 14\n");
  fprintf(fp,"sl: 10  rt: 14\n");
  fprintf(fp,"sl: 11  rt: 14\n");
  fprintf(fp,"sl: 12  rt: 14\n");
  fprintf(fp,"sl: 13  rt: 14\n");
  fprintf(fp,"sl: 14  rt: 14\n");
  fclose(fp);
  
  putenv("OtrRtInput=rt.in");
  
  OtrRtInitialize();
  
  fp=fopen("rt.in1","w");
  OtrRtPrint(fp);
  fclose(fp);
  
  OtrRtClose();  
  putenv("OtrRtInput=rt.in1");
  OtrRtUpdate();
  
  fp=fopen("rt.out","w");
  OtrRtPrint(fp);
  fclose(fp);
  
  OtrDbUserDefine("/OTR_TEST","OTR_RtMap",OtrRtUpdate,
                    0x7FFFFFFF,0x7FFFFFFF,2,1);
  OtrDbUserDefine("/OTR_TEST","OTR_RtHeader",OtrRtUpdate,
                    0x7FFFFFFF,0x7FFFFFFF,2,1);
  OtrDbUserDefine("/OTR_TEST","OTR_RtDP",OtrRtUpdate,
                    0x7FFFFFFF,0x7FFFFFFF,2,1);
  OtrRtWriteToDb();
  OtrRtClose();
  putenv("OtrRtInput=");
  OtrRtUpdate();
  fp=fopen("rt.db","w");
  if (!fp) {
    printf("OtrT0TestMe: Cannot open file: rt.db\n");
    exit(1);
    }
  OtrRtPrint(fp);
  fclose(fp);
  
  
  int sl=9;
  for (float tt=tmin;tt<tmax;tt+=10) {
    printf("tt: %f %f %f %f %f\n",
         tt,rt[sl*2+0]->Radius(tt),rt[sl*2+0]->RadiusSigma(tt),
	    rt[sl*2+1]->Radius(tt),rt[sl*2+1]->RadiusSigma(tt));
    }  
     
  return 0;
  }
