#include "Otr/OtrI.h"
#include "Otr/OtrT0.h"
#include "Otr/OtrDb.h"

/****************************************************************

OtrT0Switch(1)              T0 calculation switched on
OtrT0Initialize()           Memory alloc, init, ...
OtrT0Reinitialize()         inits the t0i=0 and the array in OtrT0Event
OtrT0Update()               calls OtrT0Calculate, updating the t0's
OtrT0WriteToDb()            at the moment into t0_run[run#].dat in /db/tzero/
OtrT0ReadFromDb()           at the moment from t0_run[run#].dat in /db/tzero/
OtrT0Close()                gives memory free
OtrT0ReadFromFile(rn)       read from t0_run[rn].dat in /db/tzero/
OtrT0WriteToFile(rn)        write into t0_run[rn].dat in /db/tzero/
OtrT0Scan(FILE*)            scans the file and update the t0 array
OtrT0Print(FILE *)           prints the t0 array to the file
OtrT0Get(int ci)              get the t0 of the ChannelId from the t0 array
OtrT0Put(int ci)              put the t0 of the ChannelId into the t0 array
OtrT0Event(int ci,unsigned char time) collecting times for t0 calculation

chindex = 16*asdid+asdchannelid        (asdid=0-7200, asdchannelid=0-15)
chid = OtrEventHitChannelId(runvariable in Hitloop)

*****************************************************************/

/* Fortran Interface */
#ifdef __cplusplus
extern "C" {
#endif
  void  OtrT0Switch_(int* i)        { OtrT0Switch(*i); }
  int   OtrT0Initialize_()          { return OtrT0Initialize(); }
  int   OtrT0Update_()              { return OtrT0Update(); }
  int   OtrT0WriteToDb_()           { return OtrT0WriteToDb(); }
  int   OtrT0Close_()               { return OtrT0Close(); }
  int   OtrT0Scan_()                { return OtrT0Scan(0); }
  int   OtrT0Print_()               { return OtrT0Print(0); }
  int   OtrT0TestMe_()              { return OtrT0TestMe(); }
#ifdef __cplusplus
  }
#endif

#include "Otr/OtrEvent.h"
  
static int   OtrT0Algorithm=0;
static int   nasds=0;
static int   ncha=0;
static int   **CollTime=0;
static short *DeltaT0=0;
static unsigned char *T0=0;
static unsigned char *T0e=0;
static int   runn;

#define NOBIN 64
#define LASTBIN 7
#define T0PERCENT 0.05
#define ID(ci) (((ci&0xffff)<<4) + (ci>>16))

/* --------------------------------------------------------------------*/

typedef struct{
  int           chid;
  unsigned char tzero;
  unsigned char tzerosigma;
  }OtrDbT0Struct;

static OtrDbField structfm[]={
  {"int"          ,"ChId"       ,OtrDbOffset(OtrDbT0Struct,chid)},
  {"unsigned char","T0"         ,OtrDbOffset(OtrDbT0Struct,tzero)},
  {"unsigned char","sigmaT0"  ,OtrDbOffset(OtrDbT0Struct,tzerosigma)},
  };

static OtrDbTable T0TableDef={"OTR_T0",1,"T0's",sizeof(OtrDbT0Struct),
                       structfm,sizeof(structfm)/sizeof(OtrDbField)};     

/* ------------------------------------------------------------------- */

void OtrT0Define() {
  OtrDbDefine("/OTR_STATUS",T0TableDef.name,OtrT0Update);
  putenv("OtrT0InputDeltas=/afs/desy.de/group/hera-b/OTR/db/tzero/dt0_run14029.dat");
  }

/* --------------------------------------------------------------------*/

float OtrT0Get(int ci) {
  return 0.5*(((float)T0[ID(ci)])-128.0);
  }


float OtrT0Sigma(int ci) {
  return ((float)T0e[ID(ci)])/16.0;
  }

void OtrT0Put(int ci, unsigned char t0) {
  if (OtrT0Algorithm==OTRIInvalid) OtrIInvalid();
  else T0[ID(ci)] = t0;
  }

/* --------------------------------------------------------------------*/
/**
  Tasks:
    - switch between different solutions of the task of the package
  */

void  OtrT0Switch(int i) {
  OtrT0Algorithm=i;
  }

/**
  Tasks:
    - Memory allocation 
    - prepare for reinitialization
    - after this call reinitalization
    
  All function of the package (beside OtrT0SetAlgorithm) have to check, 
  if the initialization was done properly.
  */
  
int OtrT0Initialize() {
  if (T0!=0) return OtrIValid();
  if (OtrT0Algorithm==OTRIInvalid) OtrT0Algorithm=OTRT0IsInit;
  if (!getenv("OtrT0InputDeltas")) {
    putenv("OtrT0InputDeltas=/afs/desy.de/group/hera-b/OTR/db/tzero/dt0_run14029.dat");
    }    
  OtrT0Update();
  return OtrIValid();
  }


/**
  Tasks:
    - update memory structures, arrays ... from database or ascii files 
  */
  
  
int   OtrT0Update() {
  printf("OtrT0Update[debug]: entering function\n");
  if (nasds!=OtrIAsdNumberOf()&&ncha != OtrIAsdNumberOfChannel()) {
    FILE *t0const;
    int i;
    char *t0constfname=getenv("OtrT0InputDeltas");
    
    if (CollTime) free(CollTime);
    if (DeltaT0) free(DeltaT0);
    if (T0) free(T0);
    if (T0e) free(T0e);
    nasds = OtrIAsdNumberOf();
    ncha  = OtrIAsdNumberOfChannel();
    T0 = (unsigned char*)malloc(nasds*ncha * sizeof(*T0));
    memset(T0,128,OtrIAsdNumberOf()*OtrIAsdNumberOfChannel());
    T0e = (unsigned char*)malloc(nasds*ncha * sizeof(*T0e));
    memset(T0e,255,OtrIAsdNumberOf()*OtrIAsdNumberOfChannel());

    if (OtrT0Algorithm&OTRT0Calculate) {
      CollTime = (int**)malloc(nasds*ncha * sizeof(int));
      for (i=0;i<nasds*ncha;i++) {
        CollTime[i] = (int*)malloc(NOBIN * sizeof(int));
        }
      DeltaT0  = (short*)malloc(nasds*ncha * sizeof(short));
      memset(DeltaT0,0,sizeof(short)*OtrIAsdNumberOf()*OtrIAsdNumberOfChannel());
      
      if (t0constfname) {
        int   IDummy[5];
        float FDummy[2];
	char  asdname[12];

        t0const  = fopen(t0constfname,"r");  
        if (!t0const) {
          printf("OtrT0Update: Cannot open Delta-T0 file: %s\n",t0constfname);
	  return OtrIInvalid();
  	  }
	
        for (int k=0;k<nasds*ncha;k++) {
	  int waxid;
	  int asdi,asdc;
	  int mono,wire;
	  int ci,ci0;

  	  int sl,layer,sec,asd;	      
	    
	  if (fscanf(t0const,"%d %d %s %d %f %d %f %d\n",&IDummy[0],&waxid,
	         asdname,&IDummy[2],&FDummy[0],&IDummy[3],&FDummy[1],&IDummy[4])<0) break;
		 
          mono = (waxid&0x7f)>>5;
          wire = (waxid&0x7f)&0x1f;
          asdi = (waxid>>7)>>4;
          asdc = (waxid>>7)&0xf;

	  sl=OtrISuperLayerFromName(asdname);
	  sscanf(asdname,"%*c%*c%*c%*c%d%*c%d%*c%d",&layer,&sec,&asd);
	  ci=OtrIFromAsdId(sl,layer,sec,asd,asdc);
	      
	  ci0=OtrIFromAsd(asdi,asdc);

          if (!OtrIError()) DeltaT0[ID(ci)] = IDummy[4];
	  
#ifdef DEBUG	      
	  if (ci!=ci0||OtrIError()) {
	    if (OtrIError()) {
	      printf("OtrT0Update: DeltaT0-scan: wax channel not found  ");
              }
	    else {
	      printf("OtrT0Update: DeltaT0-scan: wax decode mismatch    ");
	      }
            printf("sl:%d %d lay: %d %d sec %d %d asd %d %d ",
	        sl,OtrISuperLayer(ci0),
	        layer,OtrILayer(ci0),
	        sec,OtrISector(ci0),
	        asd,OtrIAsd(ci0));

	    printf(
	      "asd: %s  waxid: %x mono: %d %d wire: %d %d asd: %d %d cha: %d %d\n",
	             asdname,waxid,mono,OtrIMonoLayer(ci),
		     wire,OtrIWire(ci),asdi,OtrIAsdIndex(ci),asdc,OtrIAsdChannel(ci));
		     //OtrIdFromIndex(ci));
	    }
#endif		     
	    
 	  }
        fclose(t0const);
        }
      else {
        printf("OtrT0Update: Cannot calculate T0's, "
	  " since environement variable OtrT0InputDeltas is not specified.\n");
	exit(1);
        }
      }
    }
  
  if (OtrT0Algorithm&OTRT0Calculate) {
    int i;
    printf("OtrT0Update[debug]: perparing to calc T0\n");
    for (i=0;i<nasds*ncha;i++) {
      memset(CollTime[i],0,NOBIN * sizeof(CollTime[0][0]));
      }
    }
  else {
    int version;
    char *t0f=getenv("OtrT0Input");
    if (t0f&&t0f[0]!='\0') {
      FILE *fp=fopen(t0f,"r");
      printf("OtrT0Update[debug]: get T0 from file\n");
      if (!fp) {
        printf("OtrT0Initialize: Cannot open file %s.\n",t0f);
        return OtrIInvalid();
        }
      OtrT0Scan(fp);
      fclose(fp);
      }
    else if ((version=OtrDbGetTableVersion(T0TableDef.name))!=-1) {      
      int k,n;
      
      if (version==1) {
        OtrDbT0Struct *db=0;
        if (!(db=(OtrDbT0Struct*)OtrDbGet(&T0TableDef,db,&n))) {
          printf("OtrT0Update: no database access to T0-table.\n");
  	  OtrDbPrint(stdout,&T0TableDef);
          return OtrIInvalid();
	  }
        ncha =OtrIAsdNumberOfChannel();     
        nasds=n/ncha;    
        if (nasds!=OtrIAsdNumberOf()) {
          free(db);
	  OtrIInvalid();
          return OtrDbUpdateAll;
          }
	OtrIValid();
        for (k=0;k<n;k++) {
          int ci;
	  ci=OtrIFromId(db[k].chid);
	  if (ci==-1||OtrIError()) {
	    if (!OtrIError()) printf("OtrT0Update[warning]: Decoding problem for channel %d."
	      "Was valid, is invalid in this run\n",db[k].chid);
	    continue;
	    }
          if (!OtrIError()) {	  
            T0[ID(ci)] =db[k].tzero;
            T0e[ID(ci)]=db[k].tzerosigma;
	    }  
	  else OtrIValid();
	  /*
            T0[k] =db[k].tzero;
            T0e[k]=db[k].tzerosigma;
	    */
  	  }
        free(db);
	}
      }
    else {
      printf("OtrT0Update: Nether data base nor file "
       "OtrT0Input is specified\n");
      return OtrIInvalid();
      }
    }
  return OtrIValid();
  }

/**
  Tasks:
    - provide output memory structures, arrays ... to database 
  */

int   OtrT0WriteToDb() { 
  int i,j;
  OtrDbT0Struct *db=(OtrDbT0Struct*)
     malloc(OtrIAsdNumberOf()*OtrIAsdNumberOfChannel()*sizeof(OtrDbT0Struct));
  
  if (!db) {
    printf("OtrT0WriteToDb: Out of memory\n");
    return OtrIInvalid();
    }
  
  if (OtrT0Algorithm==OTRIInvalid) return OTRIInvalid;

  for (i=0;i<nasds;i++) {
    for (j=0;j<OtrIAsdNumberOfChannel();j++) {
      int k=i*OtrIAsdNumberOfChannel()+j;
      int ci=OtrIFromAsd(i,j);
      int chid=OtrIdFromIndex(ci);
      db[k].chid=chid;
      db[k].tzero=T0[k];
      db[k].tzerosigma=T0e[k];
      }
    }
  OtrDbPut(&T0TableDef,db,nasds*OtrIAsdNumberOfChannel());
  free(db);
  return OTRIValid;
  }

/**
  Tasks:
    - free allocate memory
    - clean up to be in a state that initialization is necessary 
  */
  
int   OtrT0Close()     { 
  int i;
  if (OtrT0Algorithm==OTRIInvalid) return OTRIInvalid;

  if (DeltaT0) free(DeltaT0);
  if (T0) free(T0);
  if (T0e) free(T0e);
  if (CollTime) {
    for (i=0;i<nasds*ncha;i++) {
      if (CollTime[i]) free(CollTime[i]);
      }
    free(CollTime);
    }
  DeltaT0=0;
  T0=0;
  T0e=0;
  CollTime=0;
  nasds=0;
  ncha=0;
  return OTRIValid;
  }


int OtrT0Scan(FILE *fp) {
  int i,j,sl,l,sec,a;
  
  unsigned int t,te;
  char s[256];
  
  if (!T0) {
    T0=(unsigned char*) malloc(OtrIAsdNumberOf()*OtrIAsdNumberOfChannel());
    memset(T0,128,OtrIAsdNumberOf()*OtrIAsdNumberOfChannel());
    }
   
  if (!T0e) {
    T0e=(unsigned char*) malloc(OtrIAsdNumberOf()*OtrIAsdNumberOfChannel());
    memset(T0e,255,OtrIAsdNumberOf()*OtrIAsdNumberOfChannel());
    }
    
  for (i=0;i<OtrIAsdNumberOf();i++) {
    if (fscanf(fp,"%s",s)!=1) {
      printf("OtrT0Scan: Format error in scanning asd %d",i);
      if (feof(fp)) printf(":  End of file reached");
      printf(".\n");
      }
    sscanf(s,"%*c%*c%*c%*c%d%*c%d%*c%d\n",&l,&sec,&a);
    s[3]='\0';
    sl=OtrISuperLayerFromName(s);
    for (j=0;j<OtrIAsdNumberOfChannel();j++) {
      fscanf(fp,"%d %d",&t,&te);
      T0[i*OtrIAsdNumberOfChannel()+j]=(unsigned char)(t);
      T0e[i*OtrIAsdNumberOfChannel()+j]=(unsigned char)te;
      
      if (t!=128) {
        OtrIFromAsdId(sl,l,sec,a,j);
#ifdef DEBUG
	if (OtrIError()) 
	  printf("OtrT0Scan: Channel not connected: %2d %d %2d %2d %2d\n",sl,l,sec,a,j); 
#endif
        }
      }
    }
  return OtrIValid();
  }
  
  
int OtrT0Print(FILE *fp) {
  int i;
  if (!T0) return OtrIInvalid();
  for (i=0;i<OtrIAsdNumberOf();i++) {
    int j;
    for (j=0;j<OtrIAsdNumberOfChannel();j++) {
      int k=i*OtrIAsdNumberOfChannel()+j;
      if (j==0) {
        int ci,chid,l=0;
	do {
	  ci=OtrIFromAsd(i,l++);
	  chid=OtrIdFromIndex(ci);
	  } while (OtrIError()&&(l<OtrIAsdNumberOfChannel()));
	if (!OtrIError()) {
          fprintf(fp,"%s-%d-%02d-%02d ",
            OtrISuperLayerName(ci),OtrILayer(ci),OtrISector(ci),OtrIAsd(ci));
	  }
	else {
	  printf("OtrT0Print: Asd %d cannot be decoded: %d\n",i,chid);
	  j=k;
	  continue;
	  }
	}
      fprintf(fp," %3d %3d",(int)T0[k],(int)T0e[k]);	
      }
    fprintf(fp,"\n");
    }
    
  return OtrIValid();
  }



/**
  Tasks:
    - test bench to be sucessfully executed before any release of a new version
    - it is requested, that the author carefully thinks about testing
    - in addition it should be a collection how to test bugs, which happened
      at least once 
  */

int   OtrT0TestMe() {
   
  if (OtrT0Algorithm==OTRIInvalid) return OTRIInvalid;
  
  FILE *fp=fopen("t0.out","w");
  if (!fp) {
    printf("OtrT0TestMe: Cannot open file: t0.out\n");
    exit(1);
    }
  putenv("OtrT0Input=/afs/desy.de/group/hera-b/OTR/db/tzero/run14154.tzero");
  OtrT0Close();

  OtrT0Initialize();
  OtrT0Print(fp);
  fclose(fp);
    
  OtrDbUserDefine("/OTR_TEST","OTR_tzero",OtrT0Update,
                    0x7FFFFFFF,0x7FFFFFFF,2,1);
  OtrT0WriteToDb();
  OtrT0Close();
  putenv("OtrT0Input=");
  OtrT0Update();
  fp=fopen("t0.db","w");
  if (!fp) {
    printf("OtrT0TestMe: Cannot open file: t0.out\n");
    exit(1);
    }
  OtrT0Print(fp);
  fclose(fp);

  return OTRIValid;
  }


int   OtrT0Event(int ci,unsigned char time) {

  int bin=time/4;
  int id=ID(ci);  /* asdindex*16 + asdchannel */
  
  if (OtrT0Algorithm==OTRIInvalid) return OTRIInvalid;

   
  runn=OtrEventInfoRunNumb();
  if (bin>=0&&bin<NOBIN) CollTime[id][bin]++;
  return OTRIValid;
}
  
int   OtrT0Calculate()    {

  int IntTime[NOBIN],a,j,id;
  float t0=0,binmin=0,binmax=0,timemin=0,timemax=0,Flag=0;
  float t0err=0;

  if (OtrT0Algorithm==OTRIInvalid) return OTRIInvalid;

  printf("OtrT0Calculate(): --> Calculate t0's ... ");
  fflush(stdout);  

 for (a=0; a<nasds; a++) {
    for (j=0; j<ncha; j++) {
      int i;
      int ci=OtrIFromAsd(a,j);
      id=ID(ci);
      for (i=0; i<NOBIN; i++) IntTime[i] = 0;
    
      IntTime[NOBIN-1]=CollTime[id][NOBIN-1];
      for (i=NOBIN-2; i>LASTBIN-1; i--)
        IntTime[i] = CollTime[id][i]+IntTime[i+1];
  
      Flag = 0;
      binmax=binmin=timemin=timemax=0;
      for (i=NOBIN-2;i>LASTBIN-1;i--) {
        if (IntTime[i]>=IntTime[LASTBIN]*T0PERCENT && Flag==0) {
	  binmax  = (i+1)*4;
 	  timemax = IntTime[i+1];
	  timemin = IntTime[i];
	  Flag  = 1;
        }
        if (IntTime[i]>timemin && Flag==1) {
	  binmin  = (i+1)*4;
	  Flag  = 2;
        }
      }
    
      if ((timemin-timemax)!=0) {
        t0    = (int) (binmax-(binmax-binmin)/(timemin-timemax)*\
		     (IntTime[LASTBIN]*T0PERCENT-timemax)+0.5)- DeltaT0[id];
        t0err = (float) (binmax-binmin)*1/sqrt(IntTime[LASTBIN]);

/*  
   RE: the constants 170+22 and 140+52 are choosen 
        to center the t0's around zero.	
	In this way the Rt-Relation is determined (using 170 + 140).
*/

        if (OtrIWirePitch(ci)<1) {
	  T0[id]  =(unsigned char)(int)((t0-170-22)*2+128);
        } else {
	  T0[id]  =(unsigned char)(int)((t0-140-52)*2+128);
	}
	T0e[id]  = (unsigned char)(int)(t0err*16);
      }
      else {
        t0    = 255;
        t0err = 255;
      }
    }
  }
  return OTRIValid;
}


int   OtrT0WriteToFile(int rn) { 

  FILE *t0file;
  char t0fname[100];

  if (OtrT0Algorithm==OTRIInvalid) return OTRIInvalid;

  if (rn) sprintf(t0fname, "/afs/desy.de/group/hera-b/OTR/db/tzero/t0_run%d.dat", rn);
  else sprintf(t0fname, "/afs/desy.de/group/hera-b/OTR/db/tzero/t0_run%d.dat", runn);
  t0file = fopen(t0fname,"w");
  OtrT0Print(t0file);

  return OTRIValid;
  }

